/*
 * imagex.c
 *
 * Create, modify, extract, mount, or unmount a WIM file
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib.h"
#include "config.h"
#include <getopt.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>

#define ARRAY_LEN(array) (sizeof(array) / sizeof(array[0]))

#define swap(a, b) ({ typeof(a) __a = (a); typeof(b) __b = (b); \
				a = __b; b = __a; })

#define for_opt(c, opts) while ((c = getopt_long_only(argc, (char**)argv, "", \
				opts, NULL)) != -1)

enum imagex_op_type {
	APPEND,
	APPLY,
	CAPTURE,
	DELETE,
	DIR,
	EXPORT,
	INFO,
	MOUNT,
	MOUNTRW,
	UNMOUNT,
};

static const char *path_basename(const char *path)
{
	const char *p = path;
	while (*p)
		p++;

	/* Trailing slashes. */
	while ((p != path - 1) && *p == '/')
		p--;

	while ((p != path - 1) && *p != '/')
		p--;

	return p + 1;
}


static const char *usage_strings[] = {
[APPEND] = 
"    imagex append DIRECTORY WIMFILE [\"IMAGE_NAME\"] [\"DESCRIPTION\"] [--boot]\n"
"        [--check] [--config CONFIG_FILE] [--flags EDITIONID] [--verify]\n",
[APPLY] = 
"    imagex apply WIMFILE [IMAGE_NUM | IMAGE_NAME | all] DIRECTORY [--check]\n"
"        [--verify] [--hardlink] [--symlink] [--verbose]\n",
[CAPTURE] = 
"    imagex capture DIRECTORY WIMFILE [\"IMAGE_NAME\"] [\"DESCRIPTION\"]\n"
"        [--boot] [--check] [--compress [TYPE]] [--config CONFIG_FILE]\n"
"        [--flags \"EditionID\"] [--norpfix] [--verify]\n",
[DELETE] = 
"    imagex delete WIMFILE (IMAGE_NUM | IMAGE_NAME | all) [--check]\n",
[DIR] = 
"    imagex dir WIMFILE (IMAGE_NUM | IMAGE_NAME | \"all\")\n",
[EXPORT] = 
"    imagex export SRC_WIMFILE (SRC_IMAGE_NUM | SRC_IMAGE_NAME | all ) \n"
"        DEST_WIMFILE [\"DEST_IMAGE_NAME\"] [\"DEST_IMAGE_DESCRIPTION\"]\n"
"        [--boot] [--check] [--compress [TYPE]]\n",
[INFO] = 
"    imagex info WIMFILE [IMAGE_NUM | IMAGE_NAME] [NEW_NAME]\n"
"        [NEW_DESC] [--boot] [--check] [--header] [--lookup-table]\n"
"        [--xml] [--extract-xml FILE] [--metadata]\n",
[MOUNT] = 
"    imagex mount WIMFILE [IMAGE_NUM | IMAGE_NAME] DIRECTORY\n"
"        [--check] [--debug]\n",
[MOUNTRW] = 
"    imagex mountrw WIMFILE [IMAGE_NUM | IMAGE_NAME] DIRECTORY\n"
"        [--check] [--debug]\n",
[UNMOUNT] = 
"    imagex unmount DIRECTORY [--commit]\n",
};

static const struct option common_options[] = {
	{"help", 0, NULL, 'h'},
	{"version", 0, NULL, 'v'},
	{NULL, 0, NULL, 0},
};

static const struct option append_options[] = {
	{"boot",   no_argument,       NULL, 'b'},
	{"check",  no_argument,       NULL, 'c'},
	{"config", required_argument, NULL, 'C'},
	{"flags",    required_argument, NULL, 'f'},
	{"verify", no_argument,       NULL, 'V'},
	{NULL, 0, NULL, 0},
};
static const struct option apply_options[] = {
	{"check",    no_argument,       NULL, 'c'},
	{"verify",   no_argument,       NULL, 'V'},
	{"hardlink", no_argument,       NULL, 'h'},
	{"symlink",  no_argument,       NULL, 's'},
	{"verbose",  no_argument,       NULL, 'v'},
	{NULL, 0, NULL, 0},
};
static const struct option capture_options[] = {
	{"boot",     no_argument,       NULL, 'b'},
	{"check",    no_argument,       NULL, 'c'},
	{"compress", optional_argument, NULL, 'x'},
	{"config",   required_argument, NULL, 'C'},
	{"flags",    required_argument, NULL, 'f'},
	{"norpfix",  no_argument,       NULL, 'n'},
	{"verify",   no_argument,       NULL, 'V'},
	{NULL, 0, NULL, 0},
};
static const struct option delete_options[] = {
	{"check", no_argument, NULL, 'c'},
	{NULL, 0, NULL, 0},
};

static const struct option export_options[] = {
	{"boot",       no_argument, NULL, 'b'},
	{"check",      no_argument , NULL, 'c'},
	{"compress",   optional_argument, NULL, 'x'},
	{NULL, 0, NULL, 0},
};

static const struct option info_options[] = {
	{"boot",         no_argument, NULL, 'b'},
	{"check",        no_argument, NULL, 'c'},
	{"header",       no_argument, NULL, 'h'},
	{"lookup-table", no_argument, NULL, 'l'},
	{"xml",          no_argument, NULL, 'x'},
	{"extract-xml",  required_argument, NULL, 'X'},
	{"metadata",     no_argument, NULL, 'm'},
	{NULL, 0, NULL, 0},
};
static const struct option mount_options[] = {
	{"check", no_argument, NULL, 'c'},
	{"debug", no_argument, NULL, 'd'},
	{NULL, 0, NULL, 0},
};
/*static const struct option mountrw_options[] = {*/
	/*{"check", no_argument, NULL, 'c'},*/
	/*{NULL, 0, NULL, 0},*/
/*};*/
static const struct option unmount_options[] = {
	{"commit", no_argument, NULL, 'c'},
	{NULL, 0, NULL, 0},
};


/* Print formatted error message to stderr. */
static void imagex_error(const char *format, ...)
{
	va_list va;
	va_start(va, format);
	fputs("IMAGEX: ", stderr);
	vfprintf(stderr, format, va);
	va_end(va);
}


static inline void version()
{
	static const char *s = 
	"imagex (" PACKAGE ") " PACKAGE_VERSION "\n"
	"Copyright (C) 2012 Eric Biggers\n"
	"License GPLv3+; GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>.\n"
	"This is free software: you are free to change and redistribute it.\n"
	"There is NO WARRANTY, to the extent permitted by law.\n"
	"\n"
	"Report bugs to "PACKAGE_BUGREPORT".\n";
	fputs(s, stdout);
}

static inline void usage(int cmd)
{
	puts("IMAGEX: Usage:");
	fputs(usage_strings[cmd], stdout);
}

static void usage_all()
{
	puts("IMAGEX: Usage:");
	for (int i = 0; i < ARRAY_LEN(usage_strings); i++)
		fputs(usage_strings[i], stdout);
	static const char *extra = 
"    imagex --help\n"
"    imagex --version\n"
"\n"
"    The compression TYPE may be \"maximum\", \"fast\", or \"none\".\n"
	;
	fputs(extra, stdout);
}

static bool verify_image_exists(int image)
{
	if (image == WIM_NO_IMAGE) {
		imagex_error("Not a valid image!\n");
		/* XXX */
		return false;
	}
	return true;
}

static bool verify_image_is_single(int image)
{
	if (image == WIM_ALL_IMAGES) {
		imagex_error("Cannot specify all images for this action!\n");
		return false;
	}
	return true;
}

static bool verify_image_exists_and_is_single(int image)
{
	return verify_image_exists(image) && verify_image_is_single(image);
}

static int get_compression_type(const char *optarg)
{
	if (optarg == NULL)
		return WIM_COMPRESSION_TYPE_LZX;
	if (strcmp(optarg, "maximum") == 0)
		return WIM_COMPRESSION_TYPE_LZX;
	else if (strcmp(optarg, "fast") == 0)
		return WIM_COMPRESSION_TYPE_XPRESS;
	else if (strcmp(optarg, "none") == 0)
		return WIM_COMPRESSION_TYPE_NONE;
	else {
		imagex_error("Invalid compression type `%s'! Must be \"maximum\", "
				"\"fast\", or \"none\".\n", optarg);
		return WIM_COMPRESSION_TYPE_INVALID;
	}
}

static int imagex_append(int argc, const char **argv)
{
	int c;
	bool boot = false;
	bool check = false;
	const char *config = NULL;
	bool verify = false;
	const char *flags = NULL;
	for_opt(c, append_options) {
		switch (c) {
		case 'b':
			boot = true;
			break;
		case 'c':
			check = true;
			break;
		case 'C':
			config = optarg;
			break;
		case 'f':
			flags = optarg;
			break;
		case 'V':
			verify = true;
			break;
		default:
			usage(APPEND);
			return 1;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc < 2 || argc > 4) {
		usage(APPEND);
		return 1;
	}
	const char *dir     = argv[0];
	const char *wimfile = argv[1];
	const char *name    = (argc >= 3) ? argv[2] : path_basename(dir);
	const char *desc    = (argc >= 4) ? argv[3] : NULL;

	WIMStruct *w = wimlib_open_wim(wimfile, check);
	if (w == NULL)
		return 1;
	if (!wimlib_add_image(w, dir, name, desc, flags, boot))
		goto append_error;
	if (!wimlib_overwrite(w, check))
		goto append_error;
	wimlib_free(w);
	/* XXX config, verify */
	(void)verify;
	(void)config;
	return 0;
append_error:
	wimlib_free(w);
	return 1;
}

/* Extract one image, or all images, from a WIM file into a directory. */
static int imagex_apply(int argc, const char **argv)
{
	int c;
	int link_type = WIM_LINK_TYPE_NONE;
	bool check = false;
	bool verify = false;
	bool verbose = false;
	for_opt(c, apply_options) {
		switch (c) {
		case 'c':
			check = true;
			break;
		case 'V':
			verify = true;
			break;
		case 'h':
			link_type = WIM_LINK_TYPE_HARD;
			break;
		case 's':
			link_type = WIM_LINK_TYPE_SYMBOLIC;
			break;
		case 'v':
			verbose = true;
			break;
		default:
			usage(APPLY);
			return 1;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc != 2 && argc != 3) {
		usage(APPLY);
		return 1;
	}

	const char *wimfile, *image_num_or_name, *dir;
	wimfile = argv[0];
	if (argc == 2) {
		image_num_or_name =  "1";
		dir = argv[1];
	} else {
		image_num_or_name = argv[1];
		dir = argv[2];
	}

	WIMStruct *w = wimlib_open_wim(wimfile, check);
	if (w == NULL)
		return 1;

	int image = wimlib_resolve_image(w, image_num_or_name);
	if (!verify_image_exists(image))
		goto apply_error;

	int num_images = wimlib_get_num_images(w);
	if (argc == 2 && num_images != 1) {
		imagex_error("`%s' contains %d images; Please select one "
				"(or all).\n", wimfile, num_images);
		usage(APPLY);
		goto apply_error;
	}

	if (!wimlib_set_output_dir(w, dir))
		goto apply_error;

	wimlib_set_verbose(w, verbose);
	wimlib_set_link_type(w, link_type);

	if (!wimlib_extract_image(w, image))
		goto apply_error;

	wimlib_free(w);

	/* XXX */
	(void)verify;
	return 0;
apply_error:
	wimlib_free(w);
	return 1;
}


/* Create a WIM file from a directory. */
static int imagex_capture(int argc, const char **argv)
{
	int c;
	bool boot            = false;
	bool check           = false;
	const char *config   = NULL;
	int compression_type = WIM_COMPRESSION_TYPE_NONE;
	const char *flags    = NULL;
	bool norpfix         = false;
	bool verify          = false;
	for_opt(c, capture_options) {
		switch (c) {
		case 'b':
			boot = true;
			break;
		case 'c':
			check = true;
			break;
		case 'x':
			compression_type = get_compression_type(optarg);
			if (compression_type == WIM_COMPRESSION_TYPE_INVALID)
				return 1;
			break;
		case 'C':
			config = optarg;
			break;
		case 'f':
			flags = optarg;
			break;
		case 'n':
			norpfix = true;
			break;
		case 'V':
			verify = true;
			break;
		default:
			usage(CAPTURE);
			return 1;
		}
	}

	argc -= optind;
	argv += optind;
	if (argc < 2 || argc > 4) {
		usage(CAPTURE);
		return 1;
	}
	const char *dir     = argv[0];
	const char *wimfile = argv[1];
	const char *name    = (argc >= 3) ? argv[2] : dir;
	const char *desc    = (argc >= 4) ? argv[3] : NULL;

	WIMStruct *w = wimlib_create_new_wim(compression_type);

	if (!wimlib_add_image(w, dir, name, desc, flags, boot)) {
		imagex_error("Failed to add the image `%s'!\n", dir);
		goto capture_error;
	}

	if (!wimlib_write(w, wimfile, WIM_ALL_IMAGES, check)) {
		imagex_error("Failed to write the WIM file `%s'!\n", wimfile);
		goto capture_error;
	}

	wimlib_free(w);

	/* XXX verify, norpfix, config */
	(void)verify;
	(void)norpfix;
	(void)config;

	return 0;
capture_error:
	wimlib_free(w);
	return 1;
}

/* Remove image(s) from a WIM. */
static int imagex_delete(int argc, const char **argv)
{
	int c;
	bool check = false;
	for_opt(c, delete_options) {
		switch (c) {
		case 'c':
			check = true;
			break;
		default:
			usage(DELETE);
			return 1;
		}
	}
	argc -= optind;
	argv += optind;

	if (argc != 2) {
		if (argc < 1)
			imagex_error("Must specify a WIM file!\n");
		if (argc < 2)
			imagex_error("Must specify an image!\n");
		usage(DELETE);
		return 1;
	}
	const char *wimfile = argv[0];
	const char *image_num_or_name = argv[1];
	WIMStruct *w = wimlib_open_wim(wimfile, check);
	if (w == NULL)
		return 1;

	int image = wimlib_resolve_image(w, image_num_or_name);

	if (!verify_image_exists(image))
		goto delete_error;

	if (!wimlib_delete_image(w, image))
		goto delete_error;

	if (!wimlib_overwrite(w, check))
		goto delete_error;
	wimlib_free(w);

	return 0;
delete_error:
	wimlib_free(w);
	return 1;
}

/* Print the files contained in an image(s) in a WIM file. */
static int imagex_dir(int argc, const char **argv)
{
	if (argc < 2) {
		imagex_error("Must specify a WIM file!\n");
		usage(DIR);
		return 1;
	}
	if (argc > 3) {
		imagex_error("Too many arguments!\n");
		usage(DIR);
		return 1;
	}

	const char *wimfile = argv[1];
	WIMStruct *w = wimlib_open_wim(wimfile, false);
	if (w == NULL)
		return 1;

	int image;
	if (argc == 3) {
		image = wimlib_resolve_image(w, argv[2]);
		if (!verify_image_exists(image))
			goto dir_error;
	} else {
		/* Image was not specified.  If the WIM only contains one image,
		 * choose that one; otherwise, print an error. */
		int num_images = wimlib_get_num_images(w);
		if (num_images != 1) {
			imagex_error("The file `%s' contains %d images; Please "
					"select one.\n", wimfile, num_images);
			usage(DIR);
			goto dir_error;
		}
		image = 1;
	}

	if (!wimlib_print_files(w, image))
		goto dir_error;

	wimlib_free(w);
	return 0;
dir_error:
	wimlib_free(w);
	return 1;
}

/* Exports one, or all, images from a WIM file to a new WIM file or an existing
 * WIM file. */
static int imagex_export(int argc, const char **argv)
{
	int c;
	bool boot = false;
	bool check = false;
	int compression_type = WIM_COMPRESSION_TYPE_NONE;
	bool compression_type_specified = false;
	for_opt(c, export_options) {
		switch (c) {
		case 'b':
			boot = true;
			break;
		case 'c':
			check = true;
			break;
		case 'x':
			compression_type = get_compression_type(optarg);
			if (compression_type == WIM_COMPRESSION_TYPE_INVALID)
				return 1;
			compression_type_specified = true;
			break;
		default:
			usage(EXPORT);
			return 1;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc < 3 || argc > 5) {
		usage(EXPORT);
		return 1;
	}
	const char *src_wimfile           = argv[0];
	const char *src_image_num_or_name = argv[1];
	const char *dest_wimfile          = argv[2];
	const char *dest_name             = (argc >= 4) ? argv[3] : NULL;
	const char *dest_desc             = (argc >= 5) ? argv[4] : NULL;
	WIMStruct *src_w, *dest_w = NULL;

 	src_w = wimlib_open_wim(src_wimfile, check);
	if (src_w == NULL)
		return 1;

	/* Determine if the destination is an existing file or not.  
	 * If so, we try to append the exported image(s) to it; otherwise, we
	 * create a new WIM containing the exported image(s). */
	struct stat stbuf;
	bool wim_is_new;
	if (stat(dest_wimfile, &stbuf) == 0) {
		/* Destination file exists. */
		if (!S_ISREG(stbuf.st_mode)) {
			imagex_error("`%s' is not a regular file!\n",
					dest_wimfile);
			goto export_error;
		}
		dest_w = wimlib_open_wim(dest_wimfile, check);
		wim_is_new = false;

		if (dest_w == NULL)
			goto export_error;

		if (compression_type_specified && compression_type != 
				wimlib_get_compression_type(dest_w)) 
		{
			imagex_error("Cannot specify a compression type that is "
					"not the same as that used in the "
					"destination WIM!\n");
			goto export_error;
		}
		compression_type = wimlib_get_compression_type(dest_w);
	} else {
		/* dest_wimfile is not an existing file, so create a new WIM. */
		if (errno == ENOENT) {
			dest_w = wimlib_create_new_wim(compression_type);
			wim_is_new = true;
			if (dest_w == NULL)
				goto export_error;

		} else {
			imagex_error("Cannot stat file `%s': %s\n",
					dest_wimfile, strerror(errno));
			goto export_error;
		}
	}

	int image = wimlib_resolve_image(src_w, src_image_num_or_name);
	if (!verify_image_exists(image))
		goto export_error;

	if (!wimlib_export_image(src_w, image, dest_w, dest_name, dest_desc, 
								boot))
		goto export_error;


	if (wim_is_new) {
		if (!wimlib_write(dest_w, dest_wimfile, WIM_ALL_IMAGES, check))
			goto export_error;
	} else {
		if (!wimlib_overwrite(dest_w, check))
			goto export_error;
	}
	wimlib_free(src_w);
	wimlib_free(dest_w);
	return 0;
export_error:
	wimlib_free(src_w);
	wimlib_free(dest_w);
	return 1;
}

/* Prints information about a WIM file; also can mark an image as bootable,
 * change the name of an image, or change the description of an image. */
static int imagex_info(int argc, const char **argv)
{
	int c;
	bool boot         = false;
	bool check        = false;
	bool header       = false;
	bool lookup_table = false;
	bool xml          = false;
	bool metadata     = false;
	bool short_header = true;
	const char *xml_out_file = NULL;

	for_opt(c, info_options) {
		switch (c) {
		case 'b':
			boot = true;
			break;
		case 'c':
			check = true;
			break;
		case 'h':
			header = true;
			short_header = false;
			break;
		case 'l':
			lookup_table = true;
			short_header = false;
			break;
		case 'x':
			xml = true;
			short_header = false;
			break;
		case 'X':
			xml_out_file = optarg;
			short_header = false;
			break;
		case 'm':
			metadata = true;
			short_header = false;
			break;
		default:
			usage(INFO);
			return 1;
		}
	}
	const char *wimfile;
	const char *image_num_or_name = "all";
	const char *new_name = NULL;
	const char *new_desc = NULL;

	argc -= optind;
	argv += optind;
	if (argc == 0 || argc > 4) {
		usage(INFO);
		return 1;
	}
	wimfile = argv[0];
	if (argc > 1) {
		image_num_or_name = argv[1];
		if (argc > 2) {
			new_name = argv[2];
			if (argc > 3) {
				new_desc = argv[3];
			} 
		}
	}

	WIMStruct *w = wimlib_open_wim(wimfile, check);

	if (w == NULL)
		return 1;

	int image = wimlib_resolve_image(w, image_num_or_name);
	if (!verify_image_exists(image))
		goto info_error;

	if (image == WIM_ALL_IMAGES && wimlib_get_num_images(w) != 1) {
		if (boot) {
			imagex_error("Cannot specify the --boot flag without "
					"specifying a specific image in a "
					"multi-image WIM!\n");
			goto info_error;
		}
		if (new_name != NULL) {
			imagex_error("Cannot specify the NEW_NAME without "
					"specifying a specific image in a "
					"multi-image WIM!\n");
			goto info_error;
		}
	}

	/* Operations that print information are separated from operations that
	 * recreate the WIM file. */
	if (new_name == NULL && !boot) {

		if (image == WIM_ALL_IMAGES && short_header)
			wimlib_print_wim_information(w);

		if (header)
			wimlib_print_header(w);

		if (lookup_table)
			wimlib_print_lookup_table(w);

		if (xml)
			if (!wimlib_print_xml_data(w, image))
				goto info_error;

		if (xml_out_file != NULL)
			if (!wimlib_extract_xml_data(w, xml_out_file))
				goto info_error;

		if (short_header)
			if (!wimlib_print_available_images(w, image))
				goto info_error;

		if (metadata && !wimlib_print_metadata(w, image))
			goto info_error;
	} else {
		if (image == WIM_ALL_IMAGES)
			image = 1;

		if (boot) {
			printf("Marking image %d as bootable\n", image);
			wimlib_set_boot_idx(w, image);
		}
		if (new_name != NULL) {
			printf("Changing the name of image %d to `%s'\n",
					image, new_name);
			wimlib_set_image_name(w, image, new_name);
		}
		if (new_desc != NULL) {
			printf("Changing the description of image %d to `%s'\n",
					image, new_desc);
			wimlib_set_image_descripton(w, image, new_desc);
		}
		if (!wimlib_overwrite_xml_and_header(w, check))
			goto info_error;
	}
	wimlib_free(w);

	return 0;
info_error:
	wimlib_free(w);
	return 1;
}

/* Mounts an image using a FUSE mount. */
static int imagex_mount_rw_or_ro(int argc, const char **argv)
{
	imagex_error("WARNING: THE CODE FOR MOUNTING WIMS IS A WORK IN PROGRESS.\n");
	bool ro;
	if (strcmp(argv[0], "mount") == 0)
		ro = true;
	else
		ro = false;

	int c;
	bool check = false;
	bool debug = false;
	for_opt(c, mount_options) {
		switch (c) {
		case 'c':
			check = true;
			break;
		case 'd':
			debug = true;
			break;
		default:
			usage(ro ? MOUNT : MOUNTRW);
			return 1;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc != 2 && argc != 3) {
		usage(ro ? MOUNT : MOUNTRW);
		return 1;
	}

	const char *wimfile = argv[0];
	const char *image_num_or_name;
	const char *dir;
	if (argc == 2) {
		image_num_or_name = "1";
		dir = argv[1];
	} else {
		image_num_or_name = argv[1];
		dir = argv[2];
	}
	WIMStruct *w = wimlib_open_wim(wimfile, check);
	if (w == NULL)
		return 1;
	int image = wimlib_resolve_image(w, image_num_or_name);

	if (!verify_image_exists_and_is_single(image)) {
		wimlib_free(w);
		return 1;
	}

	if (!wimlib_mount(w, image, dir, ro, debug, check)) {
		wimlib_free(w);
		return 1;
	}

	return 0;
}

/* Unmounts an image. */
static int imagex_unmount(int argc, const char **argv)
{
	int c;
	bool commit = false;
	for_opt(c, unmount_options) {
		switch (c) {
		case 'c':
			commit = true;
			break;
		default:
			usage(UNMOUNT);
			return 1;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc != 1) {
		usage(UNMOUNT);
		return 1;
	}

	const char *dir = argv[0];

	if (!wimlib_unmount(dir, commit))
		return 1;
	return 0;
}

struct imagex_command {
	const char *name;
	int (*func)(int , const char **);
	int cmd;
};

static struct imagex_command imagex_commands[] = {
	{"append",  imagex_append,	   APPEND},
	{"apply",   imagex_apply,   	   APPLY},
	{"capture", imagex_capture,	   CAPTURE},
	{"delete",  imagex_delete,	   DELETE},
	{"dir",     imagex_dir,		   DIR},
	{"export",  imagex_export,	   EXPORT},
	{"info",    imagex_info,	   INFO},
	{"mount",   imagex_mount_rw_or_ro, MOUNT},
	{"mountrw", imagex_mount_rw_or_ro, MOUNTRW},
	{"unmount", imagex_unmount,	   UNMOUNT},
};

#define for_imagex_command(p) for (p = &imagex_commands[0]; \
		p != &imagex_commands[ARRAY_LEN(imagex_commands)]; p++)

static void help_or_version(int argc, const char **argv)
{
	for (int i = 1; i < argc; i++) {
		const char *p = argv[i];
		if (*p == '-')
			p++;
		else
			continue;
		if (*p == '-')
			p++;
		if (strcmp(p, "help") == 0 || (*p == '?' && *(p + 1) == '\0')) {
			struct imagex_command *cmd;
			for_imagex_command(cmd) {
				if (strcmp(cmd->name, argv[1]) == 0) {
					usage(cmd->cmd);
					exit(0);
				}
			}
			usage_all();
			exit(0);
		}
		if (strcmp(p, "version") == 0) {
			version();
			exit(0);
		}
	}
}


int main(int argc, const char **argv)
{
	if (argc < 2) {
		imagex_error("No command specified!\n");
		usage_all();
		return 1;
	}

	help_or_version(argc, argv);
	argc--;
	argv++;

	struct imagex_command *cmd;

	for_imagex_command(cmd)
		if (strcmp(cmd->name, *argv) == 0)
			return (cmd->func)(argc, argv);

	imagex_error("Unrecognized command: `%s'\n", argv[0]);

	usage_all();
	return 1;
}
