/*
 * apply.c
 *
 * Support for extracting WIM files.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib_internal.h"
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>


/* 
 * Extracts a regular file from the WIM archive. 
 *
 * @dentry:  		The directory entry for the file, which must be a
 *				regular file.
 * @output_path:   	The path to which the file is to be extracted.
 * @lookup_table:	The lookup table for the WIM file.
 * @wim_fp:  		The FILE* for the WIM, opened for reading.
 * @wim_ctype:  	The type of compression used in the WIM.
 * @link_type:		One of WIM_LINK_TYPE_*; specifies what to do with
 * 			files that are hard-linked inside the WIM.
 * @is_multi_image_extraction: 
 * 			True if the image currently being extracted is just one 
 * 			image of a multi-image extraction.  This is needed so
 * 			that cross-image symbolic links can be created
 * 			correctly.
 * @return: 		True on success, false on failure. 
 */
static bool extract_regular_file(const WIMDentry *dentry, 
		WIMLookupTable *lookup_table, const char *output_path, 
		FILE *wim_fp, int wim_ctype, int link_type,
		bool is_multi_image_extraction)
{
	/* Lookup up the dentry in the lookup table. */
	WIMLookupTableEntry *entry = lookup_file_resource(dentry, lookup_table);

	/* If we already extracted the same file or a hard link copy of it, we
	 * may be able to simply create a link.  The exact action is specified
	 * by the current @link_type. */
	if (link_type != WIM_LINK_TYPE_NONE && entry != NULL && 
						entry->out_refcnt != 0)
	{
		wimlib_assert(entry->file_on_disk != NULL);

		if (link_type == WIM_LINK_TYPE_HARD) {
			if (link(entry->file_on_disk, output_path) != 0) {
				wimlib_error("Failed to hard link `%s' to `%s': %m\n",
						output_path, entry->file_on_disk);
				return false;
			}
		} else {
			wimlib_assert(link_type == WIM_LINK_TYPE_SYMBOLIC);
			int num_path_components = 
				get_num_path_components(dentry->full_path_utf8);
			if (is_multi_image_extraction)
				num_path_components++;
			size_t file_on_disk_len = strlen(entry->file_on_disk);
			char buf[file_on_disk_len + 3 * num_path_components + 1];
			char *p = &buf[0];
			for (size_t i = 0; i < num_path_components; i++) {
				*p++ = '.';
				*p++ = '.';
				*p++ = '/';
			}
			strcpy(p, entry->file_on_disk);
			if (symlink(buf, output_path) != 0) {
				wimlib_error("Failed to symlink `%s' to `%s': %m\n",
						buf, entry->file_on_disk);
				return false;
			}

		}
		return true;
	} 

	/* Otherwise, we must actually extract the file contents. */

	FILE *output_fp = fopen(output_path, "wb");
	if (output_fp == NULL) 
	{
		wimlib_error("Failed to open the file `%s' for writing: "
				"%m\n", output_path);
		return false;
	}

	/* Extract empty file, with no lookup table entry... */
	if (entry == NULL) {
		DEBUG("Empty file `%s'\n", output_path);
		fclose(output_fp);
		return true;
	}

	const WIMResourceEntry *res_entry = &entry->resource_entry;
	int ctype = resource_compression_type(wim_ctype, res_entry->flags);
	u8 *file_contents = xmalloc(res_entry->original_size);

	if (!read_full_resource(wim_fp, res_entry->size, 
				res_entry->original_size,
				res_entry->offset, ctype, file_contents)) 
	{
		wimlib_error("Failed to read file resource `%s'!\n",
				dentry->full_path_utf8);
		goto extract_regular_file_error;
	}

	if (fwrite(file_contents, 1, res_entry->original_size, output_fp) != 
			res_entry->original_size) 
	{
		wimlib_error("Failed to write to output file `%s': %m\n", 
				output_path);
		goto extract_regular_file_error;
	}

	if (fclose(output_fp) != 0) 
	{
		wimlib_error("Failed to close output file `%s': %m!\n",
				output_path);
		goto extract_regular_file_error;
	}

	/* Mark the lookup table entry to indicate this file has been extracted. */
	entry->out_refcnt++;
	free(entry->file_on_disk);
	entry->file_on_disk = xstrdup(output_path);
	free(file_contents);
	return true;
extract_regular_file_error:
	free(file_contents);
	fclose(output_fp);
	return false;
}

/* 
 * Extracts a directory from the WIM archive. 
 *
 * @dentry:		The directory entry for the directory.
 * @output_path:   	The path to which the directory is to be extracted to.
 * @return: 		True on success, false on failure. 
 */
static bool extract_directory(WIMDentry *dentry, const char *output_path)
{
	/* Compute the output path directory to the directory. */
	if (mkdir(output_path, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) != 0) 
	{
		switch (errno) {
		case EEXIST: /* Already existing directory is OK */
		case EACCES: /* We may have permissions to extract files inside
				 the directory, but not for the directory
				 itself. */
			return true;
		default:
			wimlib_error("Cannot create directory `%s': %m\n",
					output_path);
			return false;
		}
	}
	return true;
}


/* 
 * Extracts a file or directory from the WIM archive.  For use in
 * for_dentry_in_tree().
 *
 * @dentry:	The dentry to extract.
 * @arg:	A pointer to the WIMStruct for the WIM file.
 */
static bool extract_regular_file_or_directory(WIMDentry *dentry, void *arg)
{
	WIMStruct *w = (WIMStruct*)arg;
	if (w->verbose)
		puts(dentry->full_path_utf8);

	size_t len = strlen(w->output_dir);

	char output_path[len + dentry->full_path_utf8_len + 1];

	memcpy(output_path, w->output_dir, len);
	memcpy(output_path + len, dentry->full_path_utf8, dentry->full_path_utf8_len);
	output_path[len + dentry->full_path_utf8_len] = '\0';


	if (dentry_is_regular_file(dentry)) {
		return extract_regular_file(dentry, &w->lookup_table,
				output_path, w->fp, wimlib_get_compression_type(w), 
					w->link_type, w->is_multi_image_extraction);
	} else {
		wimlib_assert(dentry_is_directory(dentry));
		if (dentry_is_root(dentry)) /* Root doesn't need to be extracted. */
			return true;
		else
			return extract_directory(dentry, output_path);
	}
}

static bool extract_single_image(WIMStruct *w, int image)
{
	wimlib_assert(image >= 1 && image <= w->hdr.image_count);

	DEBUG("Attempting to extract image %u...\n", image);
	if (!wimlib_select_image(w, image))
		return false;
	return for_dentry_in_tree(wim_root_dentry(w),
			extract_regular_file_or_directory, w);
}

/* Extracts all images from the WIM to w->output_dir, with the images placed in
 * subdirectories named by their image names. */
static bool extract_all_images(WIMStruct *w)
{
	const char image_name_max_len = 100;
	DEBUG("Attempting to extract all images from `%s'\n",
			wimlib_get_filename(w));

	size_t output_path_len = strlen(w->output_dir);
	char buf[output_path_len + 1 + image_name_max_len + 1];

	memcpy(buf, w->output_dir, output_path_len);
	buf[output_path_len] = '/';
	buf[output_path_len + 1] = '\0';
	for (unsigned image = 1; image <= w->hdr.image_count; image++) {
		strncat(buf + output_path_len + 1, wimlib_get_image_name(w, image),
				image_name_max_len);
		if (!wimlib_set_output_dir(w, buf))
			return false;
		if (!extract_single_image(w, image))
			return false;
		buf[output_path_len + 1] = '\0';
	}
	return true;
}

/* Extracts a single image or all images from a WIM file. */
bool wimlib_extract_image(WIMStruct *w, int image)
{
	if (image == WIM_ALL_IMAGES) {
		w->is_multi_image_extraction = true;
		return extract_all_images(w);
	} else {
		w->is_multi_image_extraction = false;
		return extract_single_image(w, image);
	}

}

/* Set the output directory for WIM extraction.  The directory is created using
 * mkdir().  Fails if directory cannot be created or already exists. */
bool wimlib_set_output_dir(WIMStruct *w, const char *dir)
{

	if (mkdir(dir, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) != 0) 
	{
		wimlib_error("Cannot create directory `%s': %m\n", dir);
		return false;
	} else {
		w->output_dir = dir;
		return true;
	}
}

void wimlib_set_link_type(WIMStruct *w, int link_type)
{
	w->link_type = link_type;
}

