/*
 * dentry.c
 *
 * A dentry (directory entry) contains the metadata for a file.  In the WIM file
 * format, the dentries are stored in the "metadata resource" section, also
 * known as the "boot metadata", directly after the security data.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib_internal.h"
#include "io.h"
#include <unistd.h>

/* Arguments for do_free_dentry(). */
struct free_dentry_args {
	WIMDentry *root;
	WIMLookupTable *lookup_table;
	bool decrement_refcnt;
};

/* Transfers file attributes from a `stat' buffer to a WIMDentry. */
INTERNAL void stbuf_to_dentry(struct stat *stbuf, WIMDentry *dentry)
{
	dentry->attributes = 0;
	if (S_ISDIR(stbuf->st_mode))
		dentry_set_is_directory(dentry);
	else if (S_ISREG(stbuf->st_mode))
		dentry_set_is_regular_file(dentry);
	else
		wimlib_error("Unknown st_mode\n");
}

/* Transfers file attributes from a WIMDentry to a `stat' buffer. */
INTERNAL void dentry_to_stbuf(WIMDentry *dentry, struct stat *stbuf, 
						WIMLookupTable *table)
{
	if (dentry_is_directory(dentry))
		stbuf->st_mode = S_IFDIR | 0755;
	else if (dentry_is_regular_file(dentry))
		stbuf->st_mode = S_IFREG | 0644;
	WIMLookupTableEntry *lte;
	if (table == NULL)
		lte = NULL;
	else
		lte = lookup_file_resource(dentry, table);

	if (lte != NULL) {
		stbuf->st_nlink = lte->refcnt;
		stbuf->st_size = lte->resource_entry.original_size;
	} else {
		stbuf->st_nlink = 1;
		stbuf->st_size = 0;
	}
	stbuf->st_uid     = getuid();
	stbuf->st_gid     = getgid();
	stbuf->st_atime   = ms_timestamp_to_unix(dentry->last_access_time);
	stbuf->st_mtime   = ms_timestamp_to_unix(dentry->last_write_time);
	stbuf->st_ctime   = ms_timestamp_to_unix(dentry->creation_time);
	stbuf->st_blocks  = 0;
}

INTERNAL void dentry_update_all_timestamps(WIMDentry *dentry)
{
	u64 now = get_timestamp();
	dentry->creation_time       = now;
	dentry->last_access_time    = now;
	dentry->last_write_time     = now;
}

/* 
 * Calls a function on all directory entries in a directory tree.  It is called
 * on a parent before its children.
 *
 * @root: The root of the directory tree.
 * @visitor: The function to call. @args is passed as the second argument.  The
 * 	function must return true on success and false on failure, and the 
 * 	traversal is terminated if a failure result occurs. 
 * @args:  Second argument to @visitor.
 * @return:	True only if all invocations of @visitor returned true.
 */
INTERNAL bool for_dentry_in_tree(WIMDentry *root, 
		bool (*visitor)(WIMDentry*, void*), void *args)
{
	if (!(*visitor)(root, args))
		return false;
	WIMDentry *child = root->children;

	if (child == NULL)
		return true;
	do {
		if (!for_dentry_in_tree(child, visitor, args))
			return false;
		child = child->next;
	} while (child != root->children);
	return true;
}

/* 
 * Like for_dentry_in_tree(), but the visitor function is always called on a
 * dentry's children before on itself.
 */
INTERNAL bool for_dentry_in_tree_depth(WIMDentry *root, 
		bool (*visitor)(WIMDentry*, void*), void *args)
{
	WIMDentry *child = root->children;
	WIMDentry *next;
	if (child != NULL) {
		do {
			next = child->next;
			if (!for_dentry_in_tree_depth(child, visitor, args))
				return false;
			child = next;
		} while (child != root->children);
	}
	if (!(*visitor)(root, args))
		return false;
	return true;
}

/* 
 * Calculate the full path of @dentry, based on its parent's full path and on
 * its UTF-8 file name. 
 */
INTERNAL void calculate_dentry_full_path(WIMDentry *dentry)
{
	int parent_len;
	int len;
	char *parent_full_path;
	char *full_path;

	free(dentry->full_path_utf8);

	if (dentry_is_root(dentry)) {
		dentry->full_path_utf8 = xmalloc(2);
		dentry->full_path_utf8[0] = '/';
		dentry->full_path_utf8[1] = '\0';
		dentry->full_path_utf8_len = 1;
		return;
	}

	if (dentry_is_root(dentry->parent)) {
		parent_len = 0;
		parent_full_path = "";
	} else {
		parent_len = dentry->parent->full_path_utf8_len;
		parent_full_path = dentry->parent->full_path_utf8;
	}

	len = parent_len + 1 + dentry->file_name_utf8_len;
	full_path = xmalloc(len + 1);

	memcpy(full_path, parent_full_path, parent_len);
	full_path[parent_len] = '/';
	memcpy(full_path + parent_len + 1, dentry->file_name_utf8, 
				dentry->file_name_utf8_len);
	full_path[len] = '\0';
	dentry->full_path_utf8 = full_path;
	dentry->full_path_utf8_len = len;
}

/* For use in for_dentry_in_tree(). */
INTERNAL bool do_calculate_dentry_full_path(WIMDentry *dentry, void *ignore)
{
	calculate_dentry_full_path(dentry);
	return true;
}

/* 
 * Recursively calculates the subdir offsets for a directory tree. 
 *
 * @dentry:  The root of the directory tree.
 * @subdir_offset_p:  The current subdirectory offset; i.e., the subdirectory
 * 	offset for @dentry. 
 */
INTERNAL void calculate_subdir_offsets(WIMDentry *dentry, off_t *subdir_offset_p)
{
	dentry->subdir_offset = *subdir_offset_p;
	
	WIMDentry *child = dentry->children;
	if (child != NULL) {

		/* Advance the subdir offset by the amount of space the children
		 * of this dentry take up. */
		do {
			*subdir_offset_p += child->length;
			child = child->next;
		} while (child != dentry->children);

		/* End-of-directory dentry on disk. */
		*subdir_offset_p += 8;

		/* Recursively call calculate_subdir_offsets() on all the
		 * children. */
		do {
			calculate_subdir_offsets(child, subdir_offset_p);
			child = child->next;
		} while (child != dentry->children);
	} else {
		/* On disk, childless directories have a valid subdir_offset
		 * that points to an 8-byte end-of-directory dentry.  Regular
		 * files have a subdir_offset of 0. */
		if (dentry_is_directory(dentry))
			*subdir_offset_p += 8;
		else
			dentry->subdir_offset = 0;
	}
}

/*
 * Returns true if @dentry has the UTF-8 file name @name that has length
 * @name_len.
 */
INTERNAL bool dentry_has_name(const WIMDentry *dentry, 
			const char *name, int name_len)
{
	if (dentry->file_name_utf8_len != name_len)
		return false;
	return memcmp(dentry->file_name_utf8, name, name_len) == 0;
}

/* Returns the child of @dentry that has the file name @name.  
 * Returns NULL if no child has the name. */
INTERNAL WIMDentry *get_dentry_child_with_name(const WIMDentry *dentry, 
							const char *name)
{
	WIMDentry *child = dentry->children;
	if (child != NULL) {
		int len = strlen(name);
		do {
			if (dentry_has_name(child, name, len))
				return child;
			child = child->next;
		} while (child != dentry->children);
	}
	return NULL;
}

/* Retrieves the dentry that has the UTF-8 @path relative to the dentry
 * @cur_dir.  Returns NULL if no dentry having the path is found. */
static WIMDentry *get_dentry_relative_path(WIMDentry *cur_dir, const char *path)
{
	if (*path == '\0')
		return cur_dir;

	WIMDentry *child = cur_dir->children;
	if (child != NULL) {
		int base_len;
		const char *new_path = path_next_part(path, &base_len);
		do {
			if (dentry_has_name(child, path, base_len))
				return get_dentry_relative_path(child, new_path);
			child = child->next;
		} while (child != cur_dir->children);
	}
	return NULL;
}

/* Returns the dentry corresponding to the UTF-8 @path. */
INTERNAL WIMDentry *get_dentry(WIMStruct *w, const char *path)
{
	WIMDentry *root = wim_root_dentry(w);
	while (*path == '/')
		path++;
	return get_dentry_relative_path(root, path);
}

/* Returns the parent directory for the @path. */
INTERNAL WIMDentry *get_parent_dentry(WIMStruct *w, const char *path)
{
	size_t len = strlen(path);
	char buf[len + 1];
	memcpy(buf, path, len + 1);

	to_parent_name(buf, len);

	return get_dentry(w, buf);
}

/* Prints the full path of a dentry. */
INTERNAL bool print_dentry_full_path(WIMDentry *entry, void *ignore)
{
	puts(entry->full_path_utf8);
	return true;
}

/* Prints a directory entry.  @lookup_table is a pointer to the lookup table, or
 * NULL if the resource entry for the dentry is not to be printed. */
INTERNAL bool print_dentry(WIMDentry *dentry, void *lookup_table)
{
	puts("[DIRECTORY ENTRY]");
	printf("Length            = %"PRIu64"\n", dentry->length);
	printf("Attributes        = 0x%x\n", dentry->attributes);
	printf("Security ID       = %d\n", dentry->security_id);
	printf("Subdir offset     = %"PRIu64"\n", dentry->subdir_offset);
	printf("Unused1           = %"PRIu64"\n", dentry->unused1);
	printf("Unused2           = %"PRIu64"\n", dentry->unused2);
	printf("Creation Time     = %"PRIu64"\n", dentry->creation_time);
	printf("Last_Access Time  = %"PRIu64"\n", dentry->last_access_time);
	printf("Last Write Time   = %"PRIu64"\n", dentry->last_write_time);
	printf("Creation Time     = 0x%"PRIx64"\n", dentry->creation_time);
	printf("Last_Access Time  = 0x%"PRIx64"\n", dentry->last_access_time);
	printf("Last Write Time   = 0x%"PRIx64"\n", dentry->last_write_time);
	printf("Hash              = "); 
	print_hash(dentry->hash); 
	putchar('\n');
	printf("Reparse Tag       = %u\n", dentry->reparse_tag);
	printf("Hard Link Group   = %"PRIu64"\n", dentry->hard_link);
	printf("Number of Streams = %hu\n", dentry->streams);
	printf("Filename          = \"");
	print_string((u8*)dentry->file_name, dentry->file_name_len);
	puts("\"");
	printf("Filename Length   = %hu\n", dentry->file_name_len);
	printf("Filename (UTF-8)  = \"%s\"\n", dentry->file_name_utf8);
	printf("Filename (UTF-8) Length = %hu\n", dentry->file_name_utf8_len);
	printf("Short Name        = \"");
	print_string((u8*)dentry->short_name, dentry->short_name_len);
	puts("\"");
	printf("Short Name Length = %hu\n", dentry->short_name_len);
	printf("Full Path (UTF-8) = \"%s\"\n", dentry->full_path_utf8);
	if (lookup_table != NULL) {
		WIMLookupTableEntry *entry = lookup_file_resource(dentry, 
				(WIMLookupTable*)lookup_table);
		if (entry != NULL)
			print_lookup_table_entry(entry, NULL);
	}
	return true;
}

static void dentry_common_init(WIMDentry *dentry)
{
	memset(dentry, 0, sizeof(WIMDentry));
	dentry->refcnt = 1;
	dentry->security_id = -1;
}


INTERNAL void init_dentry(WIMDentry *dentry, const char *name)
{
	dentry_common_init(dentry);
	change_dentry_name(dentry, name);

	dentry_update_all_timestamps(dentry);

	dentry->next             = dentry;
	dentry->prev             = dentry;
	dentry->parent           = dentry;
}

/* 
 * Creates an unlinked directory entry.
 *
 * @name:    The base name of the new dentry.
 * @return:  A pointer to the new dentry.
 */
INTERNAL WIMDentry *create_dentry(const char *name)
{
	WIMDentry *dentry = xmalloc(sizeof(WIMDentry));
	init_dentry(dentry, name);
	return dentry;
}


/* Frees memory allocated inside a dentry.  Does not free the dentry itself. */
static void destroy_dentry(WIMDentry *dentry)
{
	free(dentry->file_name);
	free(dentry->file_name_utf8);
	free(dentry->short_name);
	free(dentry->full_path_utf8);
	for (int i = 0; i < dentry->streams; i++)
		free(dentry->stream_entries[i].stream_name);
	free(dentry->stream_entries);
}

/* Frees memory allocated for a dentry (and inside the dentry). */
INTERNAL void free_dentry(WIMDentry *dentry)
{
	destroy_dentry(dentry);
	free(dentry);
}

/* 
 * This function is passed as an argument to for_dentry_in_tree_depth() in order
 * to free a directory tree.  __args is a pointer to a `struct free_dentry_args'.
 */
static bool do_free_dentry(WIMDentry *dentry, void *__args)
{
	struct free_dentry_args *args = (struct free_dentry_args*)__args;

	if (dentry == args->root) /* Only the root of the tree needs to be unlinked. */
		unlink_dentry(dentry);

	if (args->decrement_refcnt && !dentry_is_directory(dentry))
		lookup_table_decrement_refcnt(args->lookup_table, dentry);

	wimlib_assert(dentry->refcnt >= 1);
	--dentry->refcnt;

	if (dentry->refcnt == 0)
		free_dentry(dentry);
	return true;
}

/* 
 * Unlinks and frees a dentry tree.
 *
 * @root: 		The root of the tree.
 * @lookup_table:  	The lookup table for dentries.
 * @decrement_refcnt:  	True if the dentries in the tree are to have their 
 * 			reference counts in the lookup table decremented.
 */
INTERNAL void free_dentry_tree(WIMDentry *root, WIMLookupTable *lookup_table, 
		bool decrement_refcnt)
{
	if (root == NULL)
		return;

	struct free_dentry_args args;
	args.root                = root;
	args.lookup_table        = lookup_table;
	args.decrement_refcnt    = decrement_refcnt;
	for_dentry_in_tree_depth(root, do_free_dentry, &args);
}

static bool increment_dentry_refcnt(WIMDentry *dentry, void *ignore)
{
	dentry->refcnt++;
	return true;
}

INTERNAL void increment_dentry_tree_refcnt(WIMDentry *tree)
{
	for_dentry_in_tree(tree, increment_dentry_refcnt, NULL);
}


/* 
 * Links a dentry into the directory tree.
 *
 * @dentry: The dentry to link.
 * @parent: The dentry that will be the parent of @dentry.
 */
INTERNAL void link_dentry(WIMDentry *dentry, WIMDentry *parent)
{
	dentry->parent = parent;
	if (parent->children == NULL) {
		/* Only child; link to parent. */
		parent->children = dentry;
		dentry->next = dentry;
		dentry->prev = dentry;
	} else {
		/* Not an only child; link to siblings. */
		dentry->next = parent->children;
		dentry->prev = parent->children->prev;
		dentry->next->prev = dentry;
		dentry->prev->next = dentry;
	}
}

/* Unlink a dentry from the directory tree. */
INTERNAL void unlink_dentry(WIMDentry *dentry)
{
	if (dentry_is_root(dentry))
		return;
	if (dentry_is_only_child(dentry)) {
		dentry->parent->children = NULL;
	} else {
		if (dentry_is_first_sibling(dentry))
			dentry->parent->children = dentry->next;
		dentry->next->prev = dentry->prev;
		dentry->prev->next = dentry->next;
	}
}


/* Recalculates the length of @dentry based on its file name length and short
 * name length.  */
static inline void recalculate_dentry_size(WIMDentry *dentry)
{
	dentry->length = WIM_DENTRY_DISK_SIZE + dentry->file_name_len 
		+ WIM_FILENAME_PADDING + dentry->short_name_len;
	/* Must be multiple of 8. */
	dentry->length += (8 - dentry->length % 8) % 8;
}

/* Changes the name of a dentry to @new_name.  Only changes the file_name and
 * file_name_utf8 fields; does not change the short_name, short_name_utf8, or
 * full_path_utf8 fields.  Also recalculates its length. */
INTERNAL void change_dentry_name(WIMDentry *dentry, const char *new_name)
{
	free(dentry->file_name);
	free(dentry->file_name_utf8);
	size_t utf8_len = strlen(new_name);
	size_t utf16_len;
	dentry->file_name = utf8_to_utf16(new_name, utf8_len, &utf16_len);
	wimlib_assert(dentry->file_name[0] != 0xfe || dentry->file_name[1] != 0xff);
	dentry->file_name_len = utf16_len;
	dentry->file_name_utf8 = xmalloc(utf8_len + 1);
	dentry->file_name_utf8_len = utf8_len;
	memcpy(dentry->file_name_utf8, new_name, utf8_len + 1);
	recalculate_dentry_size(dentry);
}

struct params_calculate_dentry_statistics {
	WIMLookupTable *table;
	u64 *dir_count;
	u64 *file_count;
	u64 *total_bytes;
	u64 *hard_link_bytes;
};

static bool calculate_dentry_statistics(WIMDentry *dentry, void *arg)
{
	struct params_calculate_dentry_statistics *params = arg;
	WIMLookupTableEntry *entry = lookup_file_resource(dentry, params->table);

	if (dentry_is_directory(dentry))
		++*params->dir_count;
	else {
		wimlib_assert(dentry_is_regular_file(dentry));
		++*params->file_count;
	}

	if (entry != NULL) {
		u64 size = entry->resource_entry.original_size;
		*params->total_bytes += size;
		if (++entry->out_refcnt == 1)
			*params->hard_link_bytes += size;
	}
	return true;
}

INTERNAL void calculate_dir_tree_statistics(WIMDentry *root, WIMLookupTable *table, 
			u64 *dir_count_ret, u64 *file_count_ret, 
			u64 *total_bytes_ret, u64 *hard_link_bytes_ret)
{
	struct params_calculate_dentry_statistics params;
	*dir_count_ret         = 0;
	*file_count_ret        = 0;
	*total_bytes_ret       = 0;
	*hard_link_bytes_ret   = 0;
	params.table           = table;
	params.dir_count       = dir_count_ret;
	params.file_count      = file_count_ret;
	params.total_bytes     = total_bytes_ret;
	params.hard_link_bytes = hard_link_bytes_ret;
	lookup_table_zero_out_refcnts(table);
	for_dentry_in_tree(root, calculate_dentry_statistics, &params);
}

/* 
 * Reads a directory entry from the metadata resource.
 *
 * @metadata_resource:	An array that contains the uncompressed metadata
 * 				resource for the WIM file.
 * @metadata_resource_len:	The length of @metadata_resource.
 * @offset:	The offset of the directory entry in the metadata resource.
 * @dentry:	A pointer to a WIMDentry into which the directory entry that is
 * 		read will be returned.
 * @return:	True on success, false on failure.
 */
INTERNAL bool read_dentry(const u8 metadata_resource[], u64 metadata_resource_len, 
					u64 offset, WIMDentry *dentry)
{
	dentry_common_init(dentry);

	/*Make sure the dentry really fits into the metadata resource.*/
	if (offset + 8 > metadata_resource_len) {
		wimlib_error("Directory entry starting at %"PRIu64" ends past the "
				"end of the metadata resource (size %"PRIu64")!\n",
				offset, metadata_resource_len);
		return false;
	}

	/* Before reading the whole entry, we need to read just the length.
	 * This is because an entry of length 8 (that is, just the length field)
	 * terminates the list of sibling directory entries. */

	const u8 *p = &metadata_resource[offset];

	p = get_u64(p, &dentry->length);

	/* A zero length field (really a length of 8, since that's how big the
	 * directory entry is...) indicates that this is the end of directory
	 * dentry.  We do not read it into memory as an actual dentry, so just
	 * return true in that case. */
	if (dentry->length == 0)
		return true;

	if (offset + dentry->length >= metadata_resource_len) {
		wimlib_error("Directory entry at offset %"PRIu64" and with size "
				"%"PRIu64" ends past the end of the metadata resource "
				"(size %"PRIu64")!\n", offset, dentry->length,
				metadata_resource_len);
		return false;
	}

	/* If it is a recognized length, read the rest of the directory entry.
	 * Note: The root directory entry has no name, and its length does not
	 * include the short name length field.  */
	if (dentry->length < WIM_DENTRY_DISK_SIZE)
	{
		wimlib_error("Directory entry has invalid length of "
				"%"PRIu64" bytes\n", dentry->length);
		return false;
	}
	p = get_bytes(p, WIM_DENTRY_DISK_SIZE - 8, &dentry->attributes);


	TO_LE32(dentry->attributes);

	/* We are currently ignoring the security data, so the security ID must
	 * be set to -1 to indicate that no security descriptor is referenced. */
	dentry->security_id = -1;
	/*TO_LE32(dentry->security_id);*/
	TO_LE64(dentry->subdir_offset);

	TO_LE64(dentry->unused1);
	TO_LE64(dentry->unused2);
	TO_LE64(dentry->creation_time);
	TO_LE64(dentry->last_access_time);
	TO_LE64(dentry->last_write_time);

	/* Hash goes here */

	TO_LE32(dentry->reparse_tag);

	/* reparse_reserved does not exist.  See comment in wimlib_internal.h */
	/*TO_LE32(dentry->reparse_reserved);*/

	TO_LE64(dentry->hard_link);
	TO_LE16(dentry->streams);
	TO_LE16(dentry->short_name_len);
	TO_LE16(dentry->file_name_len);

	u64 calculated_size = dentry->file_name_len + 
		dentry->short_name_len + WIM_DENTRY_DISK_SIZE + 
			WIM_FILENAME_PADDING;

	if (dentry->length < calculated_size) {
		wimlib_error("Unexpected end of directory entry! (Expected "
				"%"PRIu64" bytes, got %"PRIu64" bytes. "
				"short_name_len = %lu, file_name_len = %lu)\n", 
				calculated_size, dentry->length,
				dentry->short_name_len, dentry->file_name_len);
		return false;
	}

	/* Read the filename. */
	dentry->file_name = xmalloc(dentry->file_name_len);
	p = get_bytes(p, dentry->file_name_len, dentry->file_name);

	/* Convert filename to UTF-8. */
	size_t utf8_len;
	dentry->file_name_utf8 = utf16_to_utf8(dentry->file_name, 
					dentry->file_name_len, &utf8_len);
	dentry->file_name_utf8_len = utf8_len;

	/*DEBUG("%s, offset = %lu\n", dentry->file_name_utf8, offset);*/

	/* Undocumented padding between file name and short name. */
	p += WIM_FILENAME_PADDING;

	/* Read the short filename. */
	dentry->short_name = xmalloc(dentry->short_name_len);
	p = get_bytes(p, dentry->short_name_len, dentry->short_name);

	/* Handle any stream entries that follow this dir entry. */
	u16 num_streams = dentry->streams;
	u64 num_allocated_stream_names = 0;
	if (num_streams != 0) {
		dentry->stream_entries = xmalloc(num_streams * sizeof(WIMStreamEntry));
		WIMStreamEntry *cur_entry = dentry->stream_entries;
		while (num_streams--) {
			/* Read the base stream entry, excluding the stream name. */
			calculated_size += WIM_STREAM_ENTRY_DISK_SIZE;
			if (calculated_size > dentry->length) {
				wimlib_error("Stream entries go past end of "
						"directory entry (%"PRIu64" streams "
						"left to read)!\n", num_streams + 1);
				goto read_dentry_error;
			}

			p = get_bytes(p, WIM_STREAM_ENTRY_DISK_SIZE, cur_entry);

			TO_LE64(cur_entry->length);
			TO_LE64(cur_entry->unused1);
			TO_LE16(cur_entry->stream_name_len);

			/* Read the name of the stream. */
			calculated_size += cur_entry->stream_name_len;

			if (calculated_size > dentry->length) {
				wimlib_error("Stream entries go past end of "
						"directory entry (%"PRIu64" streams "
						"left to read)!\n", num_streams + 1);
				goto read_dentry_error;
			}

			cur_entry->stream_name = xmalloc(cur_entry->stream_name_len);
			num_allocated_stream_names++;
			p = get_bytes(p, cur_entry->stream_name_len,
							cur_entry->stream_name);
			cur_entry++;
		}
	}

	/* There may be extra bytes at the end of the directory entry;
	 * supposedly these are only used for padding the directory entry length
	 * to be a multiple of 8. */

	return true;

read_dentry_error:
	dentry->streams = num_allocated_stream_names;
	destroy_dentry(dentry);
	return false;
}

/* 
 * Writes a dentry to the output file.
 *
 * @dentry:  The dentry structure.
 * @out:     The open output file stream to write the data to.
 * @return:  True on success, false on failure.
 */
static bool write_dentry(WIMDentry *dentry, FILE *out)
{
	u8 buf[dentry->length];
	u8 *p = buf;
	wimlib_assert(dentry->length >= WIM_DENTRY_DISK_SIZE);
	wimlib_assert(dentry->length % 8 == 0);
	p = put_u64(p, dentry->length);
	p = put_u32(p, dentry->attributes);
	p = put_u32(p, dentry->security_id);
	p = put_u64(p, dentry->subdir_offset);
	p = put_u64(p, dentry->unused1);
	p = put_u64(p, dentry->unused2);
	p = put_u64(p, dentry->creation_time);
	p = put_u64(p, dentry->last_access_time);
	p = put_u64(p, dentry->last_write_time);
	p = put_bytes(p, WIM_HASH_SIZE, dentry->hash);
	p = put_u32(p, dentry->reparse_tag);
	p = put_u64(p, dentry->hard_link);
	p = put_u16(p, dentry->streams);
	p = put_u16(p, dentry->short_name_len);
	p = put_u16(p, dentry->file_name_len);
	p = put_bytes(p, dentry->file_name_len, (u8*)dentry->file_name);
	p = put_u16(p, 0); /* filename padding, 2 bytes. */
	p = put_bytes(p, dentry->short_name_len, (u8*)dentry->short_name);
	WIMStreamEntry *sentry = dentry->stream_entries;
	unsigned num_streams = dentry->streams;
	while (num_streams--) {
		p = put_u64(p, sentry->length);
		p = put_u64(p, sentry->unused1);
		p = put_bytes(p, WIM_HASH_SIZE, sentry->hash);
		p = put_u16(p, sentry->stream_name_len);
		p = put_bytes(p, sentry->stream_name_len, (u8*)sentry->stream_name);
		sentry++;
	}
	if (fwrite(buf, 1, dentry->length, out) != dentry->length)
		goto write_dentry_error;

	return true;
write_dentry_error:
	wimlib_error("Failed to write dentry: %m\n");
	return false;
}

/* Writes an end-of-directory entry to @out. */
static bool write_end_of_directory_entry(FILE *out)
{
	u8 buf[8];
	put_u64(buf, 0);
	if (fwrite(buf, 1, sizeof(buf), out) != sizeof(buf)) {
		wimlib_error("Failed to write end-of-directory entry: %m\n");
		return false;
	}
	return true;
}

/* Recursive function that writes a dentry tree rooted at @tree, not including
 * @tree itself, which has already been written, except in the case of the root
 * dentry, which is written right away, along with an end-of-directory entry. */
INTERNAL bool write_dentry_tree(WIMDentry *tree, FILE *out)
{
	if (dentry_is_root(tree)) {
		if (!write_dentry(tree, out))
			return false;
		if (!write_end_of_directory_entry(out))
			return false;
	}

	/* Nothing to do for a regular file. */
	if (dentry_is_regular_file(tree))
		return true;

	/* Write child dentries and end-of-directory entry. */
	WIMDentry *child = tree->children;
	if (child != NULL) {
		do {
			if (!write_dentry(child, out))
				return false;
			child = child->next;
		} while (child != tree->children);
	}

	if (!write_end_of_directory_entry(out))
		return false;

	/* Recurse on children. */
	if (child != NULL) {
		do {
			if (!write_dentry_tree(child, out))
				return false;
			child = child->next;
		} while (child != tree->children);
	}
	return true;
}

/* Reads the children of a dentry, and all their children, ..., etc. from the
 * metadata resource and into the dentry tree.
 * @metadata_resource:	An array that contains the uncompressed metadata
 * 				resource for the WIM file.
 * @metadata_resource_len:	The length of @metadata_resource.
 * @dentry:	A pointer to a WIMDentry that is the root of the directory tree
 * 		and has already been read from the metadata resource.  It does not 
 * 		need to be the real root, because this procedure is called 
 * 		recursively.
 * @return:	True on success, false on failure. 
 */
INTERNAL bool read_dentry_tree(const u8 metadata_resource[], 
		u64 metadata_resource_len, WIMDentry *dentry)
{
	/* Offset of first child of @dentry. */
	u64 cur_offset = dentry->subdir_offset;

	/* If @dentry is a regular file, nothing more needs to be done for this
	 * branch. */
	if (cur_offset == 0)
		return true;

	WIMDentry  cur_child;
	WIMDentry *prev_child = NULL;
	WIMDentry *first_child = NULL;

	/* Find and read all the children of @dentry. */
	while (1) {

		/* Read next child of @dentry into @cur_child. */
		if (!read_dentry(metadata_resource, metadata_resource_len, 
					cur_offset, &cur_child))
			return false;

		/* Check for end of directory. */
		if (cur_child.length == 0) {
			/* Link last child to first one, and set parent's
			 * children pointer to the first child.  */
			if (prev_child != NULL) {
				prev_child->next = first_child;
				first_child->prev = prev_child;
			}
			dentry->children = first_child;
			return true;
		}

		/* Not end of directory.  Allocate this child permanently and
		 * link it to the parent and previous child. */
		WIMDentry *child = xmalloc(sizeof(WIMDentry));
		memcpy(child, &cur_child, sizeof(WIMDentry));

		if (prev_child == NULL) {
			first_child = child;
		} else {
			prev_child->next = child;
			child->prev = prev_child;
		}
		prev_child = child;

		child->parent = dentry;

		/* If there are children of this child, call this procedure
		 * recursively. */
		if (child->subdir_offset != 0) {
			if (!read_dentry_tree(metadata_resource, 
					metadata_resource_len, child))
				return false;
		}
		/* Advance to the offset of the next child. */
		cur_offset += child->length;
	}
	/* Not reached. */
}
