/*
 * header.c
 *
 * Read, write, or create a WIM header.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib_internal.h"
#include "io.h"

static const u8 wim_magic_chars[WIM_MAGIC_LEN] = { 
			'M', 'S', 'W', 'I', 'M', '\0', '\0', '\0' };

/* 
 * Reads the header for a WIM file.
 *
 * @fp:  	The FILE* for the WIM file, positioned at the beginning of the
 * 			file stream, which is where the header starts.
 * @hdr: 	A pointer to a WIMHeader structure into which the information from the
 * 			header will be written.
 * @return:  	True on success, false on failure.
 */
INTERNAL bool read_header(FILE *fp, WIMHeader *hdr)
{
	u8 magic_buf[WIM_MAGIC_LEN];
	size_t bytes_read = fread(magic_buf, 1, WIM_MAGIC_LEN, fp);

	if (bytes_read != WIM_MAGIC_LEN)
		goto read_header_fread_error;

	/* Byte 8 */

	if (memcmp(magic_buf, wim_magic_chars, WIM_MAGIC_LEN) != 0) {
		wimlib_error("Not a WIM file (invalid magic characters)!\n");
		return false;
	}

	bytes_read = fread(&hdr->hdr_size, 1, sizeof(u32), fp);
	if (bytes_read != sizeof(u32))
		goto read_header_fread_error;

	TO_LE32(hdr->hdr_size);

	/* Byte 12 */

	if (hdr->hdr_size != WIM_HEADER_DISK_SIZE) {
		wimlib_error("Unexpected header length of %u\n", hdr->hdr_size);
		return false;
	}
	{
		/* Read the rest of the header into a buffer. */
		u8 buf[hdr->hdr_size];

		size_t hdr_rem_size = hdr->hdr_size - WIM_MAGIC_LEN - sizeof(u32);

		bytes_read = fread(buf, 1, hdr_rem_size, fp);
		if (bytes_read != hdr_rem_size)
			goto read_header_fread_error;

		const u8 *p = buf;

		p = get_u32(p, &hdr->version);
		p = get_u32(p, &hdr->flags);
		p = get_u32(p, &hdr->chunk_size);
		p = get_bytes(p, WIM_GID_LEN, hdr->guid);
		p = get_u16(p, &hdr->part_number);
		p = get_u16(p, &hdr->total_parts);
		p = get_u32(p, &hdr->image_count);

		/* Byte 48 */

		p = get_resource_entry(p, &hdr->lookup_table_metadata);
		p = get_resource_entry(p, &hdr->xml_metadata);
		p = get_resource_entry(p, &hdr->boot_metametadata);

		/* Byte 120 */

		p = get_u32(p, &hdr->boot_idx);

		/* Byte 124 */

		p = get_resource_entry(p, &hdr->integrity);

		/* Byte 148 */

		/* 60 bytes of unused stuff.  I have no idea if it is secretly
		 * used for something or not, but it's best just to read it in
		 * so it can be copied to the output file.  */
		p = get_bytes(p, WIM_UNUSED_LEN, hdr->unused);

		/* Byte 208 */
	}

	return true;
read_header_fread_error:
	if (feof(fp))
		wimlib_error("Unexpected EOF while reading WIM header!\n");
	else
		wimlib_error("Error reading WIM header: %m\n");
	return false;
}

/* 
 * Writes the header for a WIM file.
 *
 * @hdr: 	A pointer to a WIMHeader structure that describes the header.
 * @out:	The FILE* for the output file, positioned at the appropriate
 * 		place (the beginning of the file).
 * @return:	True on success, false on failure.
 */
INTERNAL bool write_header(const WIMHeader *hdr, FILE *out)
{
	DEBUG("Writing WIM header.\n");
	u8 buf[hdr->hdr_size];
	u8 *p = buf;
	p = put_bytes(p, WIM_MAGIC_LEN, wim_magic_chars);
	p = put_u32(p, hdr->hdr_size);
	p = put_u32(p, hdr->version);
	p = put_u32(p, hdr->flags);
	p = put_u32(p, hdr->chunk_size);
	p = put_bytes(p, WIM_GID_LEN, hdr->guid);
	p = put_u16(p, hdr->part_number);
	p = put_u16(p, hdr->total_parts);
	p = put_u32(p, hdr->image_count);
	p = put_resource_entry(p, &hdr->lookup_table_metadata);
	p = put_resource_entry(p, &hdr->xml_metadata);
	p = put_resource_entry(p, &hdr->boot_metametadata);
	p = put_u32(p, hdr->boot_idx);
	p = put_resource_entry(p, &hdr->integrity);
	p = put_bytes(p, WIM_UNUSED_LEN, hdr->unused);
	if (fwrite(buf, 1, hdr->hdr_size, out) != hdr->hdr_size) {
		wimlib_error("Failed to write WIM header!\n");
		return false;
	}
	return true;
}

/*
 * Creates a header for a WIM file.
 *
 * @hdr:	A pointer to a WIMHeader structer to be initialized.
 * @flags:	Flags to go in the header.
 */
INTERNAL void init_header(WIMHeader *hdr, int hdr_flags)
{
	hdr->hdr_size    = WIM_HEADER_DISK_SIZE;
	hdr->version     = WIM_VERSION;
	hdr->flags       = hdr_flags;
	hdr->chunk_size  = WIM_CHUNK_SIZE;
	hdr->part_number = 1;
	hdr->total_parts = 1;
	hdr->image_count = 0;
	hdr->boot_idx    = 0;
	randomize_byte_array(hdr->guid, sizeof(hdr->guid));
	memset(&hdr->integrity, 0, sizeof(hdr->integrity));
}

/* Prints information from the header of the WIM file associated with @w. */
void wimlib_print_header(const WIMStruct *w)
{
	const WIMHeader *hdr = &w->hdr;
	puts("[WIM HEADER]");
	fputs ("Magic Characters            = MSWIM\\000\\000\\000", stdout);
	printf("Header Size                 = %u\n", hdr->hdr_size);
	printf("Version                     = 0x%x\n", hdr->version);
	printf("Flags                       = 0x%x\n", hdr->flags);

	if (hdr->flags & WIM_HDR_FLAG_COMPRESS_XPRESS)
		printf("    COMPRESS XPRESS FLAG is set\n");

	if (hdr->flags & WIM_HDR_FLAG_COMPRESS_LZX)
		printf("    COMPRESS LZX FLAG is set\n");

	printf("Chunk Size                  = %u\n", hdr->chunk_size);
	fputs ("GUID                        = ", stdout);
	print_byte_field(hdr->guid, WIM_GID_LEN);
	putchar('\n');
	printf("Part Number                 = %hu\n", hdr->part_number);
	printf("Total Parts                 = %hu\n", hdr->total_parts);
	printf("Image Count                 = %u\n", hdr->image_count);
	printf("Lookup Table Size           = %"PRIu64"\n", 
						(u64)hdr->lookup_table_metadata.size);
	printf("Lookup Table Flags          = 0x%hhx\n", 
						hdr->lookup_table_metadata.flags);
	printf("Lookup Table Offset         = %"PRIu64"\n",
						hdr->lookup_table_metadata.offset);
	printf("Lookup Table Original_size  = %"PRIu64"\n", 
						hdr->lookup_table_metadata.original_size);
	printf("XML Data Size               = %"PRIu64"\n", 
						(u64)hdr->xml_metadata.size);
	printf("XML Data Flags              = 0x%hhx\n", 
						hdr->xml_metadata.flags);
	printf("XML Data Offset             = %"PRIu64"\n", 
						hdr->xml_metadata.offset);
	printf("XML Data Original Size      = %"PRIu64"\n", 
						hdr->xml_metadata.original_size);
	printf("Boot Metadata Size          = %"PRIu64"\n", 
						(u64)hdr->boot_metametadata.size);
	printf("Boot Metadata Flags         = 0x%hhx\n", 
						hdr->boot_metametadata.flags);
	printf("Boot Metadata Offset        = %"PRIu64"\n", 
						hdr->boot_metametadata.offset);
	printf("Boot Metadata Original Size = %"PRIu64"\n", 
						hdr->boot_metametadata.original_size);
	printf("Boot Index                  = %u\n", hdr->boot_idx);
	printf("Integrity Size              = %"PRIu64"\n", (u64)hdr->integrity.size);
	printf("Integrity Flags             = 0x%hhx\n", hdr->integrity.flags);
	printf("Integrity Offset            = %"PRIu64"\n", hdr->integrity.offset);
	printf("Integrity Original_size     = %"PRIu64"\n", hdr->integrity.original_size);
}
