/*
 * integrity.c
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib_internal.h"
#include "io.h"
#include "sha1.h"

typedef struct WIMIntegrityTable {
	u32 integrity_table_size;
	u32 num_elements;
	u32 chunk_size;
	u8  hash_list[0];
} WIMIntegrityTable;

#define INTEGRITY_CHUNK_SIZE 1048576

static int verify_integrity(FILE *fp, u64 num_bytes, u32 chunk_size, const u8 *sha1sums)
{
	char *buf = xmalloc(chunk_size);
	u8 resblock[WIM_HASH_SIZE];
	u64 bytes_remaining = num_bytes;
	size_t bytes_to_read;
	while (bytes_remaining != 0) {

		unsigned percent_done = (num_bytes - bytes_remaining) * 100 / num_bytes;
		printf("Verifying integrity of WIM (%"PRIu64" bytes remaining, %u%% "
				"done)\r", bytes_remaining, percent_done);
		fflush(stdout);

		bytes_to_read = min(chunk_size, bytes_remaining);
		bytes_remaining -= bytes_to_read;
		if (fread(buf, 1, bytes_to_read, fp) != bytes_to_read) {
			free(buf);
			putchar('\n');
			if (feof(fp)) {
				wimlib_error("Unexpected EOF while verifying "
						"integrity!\n");
			} else {
				wimlib_error("File stream error while verifying "
						"integrity: %m\n");
			}
			return WIM_INTEGRITY_ERROR;
		}
		sha1_buffer(buf, bytes_to_read, resblock);
		if (memcmp(resblock, sha1sums, WIM_HASH_SIZE) != 0) {
			free(buf);
			return WIM_INTEGRITY_NOT_OK;
		}
		sha1sums += WIM_HASH_SIZE;
	}
	putchar('\n');
	free(buf);
	return WIM_INTEGRITY_OK;
}

/* 
 * Checks the integrity of a WIM file. 
 *
 * Returns:
 * 	WIM_INTEGRITY_NONEXISTENT if there is no integrity information.
 * 	WIM_INTEGRITY_OK if the integrity of the WIM was successfully verified.
 * 	WIM_INTEGRITY_NOT_OK if the integrity was checked, but the file was 
 * 		intact.
 * 	WIM_INTEGRITY_ERROR if there is a file stream error, unexpected
 * 		end-of-file, or inconsistency in the integrity table itself
 */
INTERNAL int check_integrity(WIMStruct *w)
{
	WIMResourceEntry *res_entry = &w->hdr.integrity;
	FILE *fp = w->fp;
	if (res_entry->size == 0)
		return WIM_INTEGRITY_NONEXISTENT;
	
	int ctype = wim_resource_compression_type(w, res_entry);
	if (res_entry->original_size < 12) {
		wimlib_error("Integrity table resource is too short!\n");
		return WIM_INTEGRITY_ERROR;
	}
	u8 *buf = xmalloc(res_entry->original_size);
	if (!read_full_resource(w->fp, res_entry->size, res_entry->original_size,
				res_entry->offset, ctype, buf)) {
		free(buf);
		return WIM_INTEGRITY_ERROR;
	}
	const u8 *p = buf;
	u32 integrity_table_size;
	u32 num_entries;
	u32 chunk_size;
	p = get_u32(p, &integrity_table_size);
	p = get_u32(p, &num_entries);
	p = get_u32(p, &chunk_size);
	if (integrity_table_size != res_entry->original_size) {
		wimlib_error("Inconsistent integrity table sizes: header says %u "
				"but resource entry says %"PRIu64"\n",
				integrity_table_size, res_entry->original_size);
		free(buf);
		return WIM_INTEGRITY_ERROR;
	}
	DEBUG("integrity_table_size = %u, num_entries = %u, chunk_size = %u\n",
			integrity_table_size, num_entries, chunk_size);

	u64 expected_size = num_entries * WIM_HASH_SIZE + 12;

	if (integrity_table_size != expected_size) {
		wimlib_error("Integrity table is %u bytes, but expected %"PRIu64" "
				"bytes to hold %u entries!\n", integrity_table_size,
				expected_size, num_entries);
		free(buf);
		return WIM_INTEGRITY_ERROR;
	}

	u64 end_header_offset = w->hdr.hdr_size;
	u64 end_lookup_table_offset = w->hdr.lookup_table_metadata.offset +
				w->hdr.lookup_table_metadata.size;
	u64 bytes_to_check = end_lookup_table_offset - end_header_offset;
	u64 expected_num_entries = bytes_to_check / chunk_size +
			(bytes_to_check % chunk_size != 0);
	if (num_entries != expected_num_entries) {
		wimlib_error("%"PRIu64 " entries would be required to checksum "
			"the %"PRIu64" bytes from the end of the header to the\n"
			"end of the lookup table with a chunk size of %u, but "
			"there were only %u entries!\n", 
			expected_num_entries, bytes_to_check, chunk_size,
			num_entries);
		free(buf);
		return WIM_INTEGRITY_ERROR;
	}
	if (fseeko(fp, end_header_offset, SEEK_SET) != 0) {
		wimlib_error("Failed to seek to byte %"PRIu64" of WIM: %m\n",
				end_header_offset);
		free(buf);
		return WIM_INTEGRITY_ERROR;
	}
	free(buf);
	return verify_integrity(fp, bytes_to_check, chunk_size, p);
}

/* Writes integrity information to the output stream for a WIM file being
 * written. 
 * @end_header_offset is the offset of the byte after the header, which is the
 * beginning of the region that is checksummed.
 * @end_lookup_table_offset is the offset of the byte after the lookup table,
 * which is the end of the region that is checksummed. */
INTERNAL bool write_integrity(FILE *out, u64 end_header_offset, 
				u64 end_lookup_table_offset)
{
	DEBUG("Seeking to end of header at byte %"PRIu64"\n",
			end_header_offset);
	if (fseeko(out, end_header_offset, SEEK_SET) != 0) {
		wimlib_error("Failed to seek to byte %"PRIu64" of WIM: %m\n",
				end_header_offset);
		return false;
	}
	wimlib_assert(end_header_offset <= end_lookup_table_offset);

	u64 bytes_to_check = end_lookup_table_offset - end_header_offset;
	u32 num_entries = bytes_to_check / INTEGRITY_CHUNK_SIZE +
			(bytes_to_check % INTEGRITY_CHUNK_SIZE != 0);
	u32 integrity_table_size = num_entries * WIM_HASH_SIZE + 3 * sizeof(u32);

	DEBUG("integrity table size = %u\n", integrity_table_size);


	u8 *buf = xmalloc(integrity_table_size);
	u8 *p = buf;
	p = put_u32(p, integrity_table_size);
	p = put_u32(p, num_entries);
	p = put_u32(p, INTEGRITY_CHUNK_SIZE);

	char *chunk_buf = xmalloc(INTEGRITY_CHUNK_SIZE);
	u64 bytes_remaining = bytes_to_check;

	DEBUG("Bytes to check = %"PRIu64"\n", bytes_to_check);

	while (bytes_remaining != 0) {

		unsigned percent_done = (bytes_to_check - bytes_remaining) * 
							100 / bytes_to_check;

		printf("Calculating integrity checksums for WIM (%"PRIu64" bytes "
				"remaining, %u%% done)\r", bytes_remaining, 
								percent_done);
		fflush(stdout);

		size_t bytes_to_read = min(INTEGRITY_CHUNK_SIZE, bytes_remaining);
		size_t bytes_read = fread(chunk_buf, 1, bytes_to_read, out);
		if (bytes_read != bytes_to_read) {
			putchar('\n');
			DEBUG("Only read %"PRIu64" bytes!\n", bytes_read);
			if (feof(out)) {
				wimlib_error("Unexpected EOF while calculating "
						"integrity checksums!\n");
			} else {
				wimlib_error("File stream error while calculating "
						"integrity checksums: %m\n");
			}
			free(chunk_buf);
			free(buf);
			return false;
		}
		sha1_buffer(chunk_buf, bytes_read, p);
		p += WIM_HASH_SIZE;
		bytes_remaining -= bytes_read;
	}
	putchar('\n');
	wimlib_assert(p - buf == integrity_table_size);
	free(chunk_buf);
	if (fseeko(out, 0, SEEK_END) != 0) {
		wimlib_error("Failed to seek to end of WIM to write integrity "
				"table: %m\n");
		free(buf);
		return false;
	}
	if (fwrite(buf, 1, integrity_table_size, out) != integrity_table_size) {
		free(buf);
		wimlib_error("Failed to write integrity table to end of WIM: %m\n");
		return false;
	}
	free(buf);
	return true;
}
