/*
 * lookup_table.c
 *
 * Lookup table that maps dentries to file resources.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib_internal.h"
#include "io.h"

INTERNAL void lookup_table_init(WIMLookupTable *table, size_t capacity)
{
	table->num_entries = 0;
	table->capacity    = capacity;
	table->array       = xzalloc(capacity * sizeof(WIMLookupTableEntry*));
}


INTERNAL WIMLookupTableEntry *new_lookup_table_entry()
{
	WIMLookupTableEntry *lte = xmalloc(sizeof(WIMLookupTableEntry));
	lte->file_on_disk = NULL;
	lte->other_wim_fp = NULL;
	lte->next         = NULL;
	lte->part_number  = 1;
	lte->refcnt       = 1;
	lte->staging_num_times_opened = 0;
	return lte;
}

INTERNAL void free_lookup_table_entry(WIMLookupTableEntry *entry)
{
	if (entry != NULL) {
		free (entry->file_on_disk);
		free (entry);
	}
}

static bool do_free_lookup_table_entry(WIMLookupTableEntry *entry, void *ignore)
{
	free_lookup_table_entry(entry);
	return true;
}


/*
 * Inserts an entry into the lookup table.
 *
 * @lookup_table:	A pointer to the lookup table.
 * @entry:		A pointer to the entry to insert.
 */
INTERNAL void lookup_table_insert(WIMLookupTable *table, 
					WIMLookupTableEntry *entry)
{
	WIMLookupTableEntry **array = table->array;
	size_t pos = *(u64*)entry->hash % table->capacity;

	if (array[pos] == NULL) {
		entry->next = NULL;
		array[pos] = entry;
	} else {
		entry->next = array[pos];
		array[pos] = entry;
	}
	table->num_entries++;
}


INTERNAL void lookup_table_unlink(WIMLookupTable *table, 
						WIMLookupTableEntry *entry)
{
	u64 pos = *(u64*)entry->hash % table->capacity;
	WIMLookupTableEntry *prev = NULL;
	WIMLookupTableEntry *cur_entry = table->array[pos];
	WIMLookupTableEntry *next;
	while (entry != NULL) {
		next = cur_entry->next;
		if (cur_entry == entry) {
			if (prev == NULL)
				table->array[pos] = next;
			else
				prev->next = next;
			table->num_entries--;
			return;
		}
		prev = entry;
		cur_entry = next;
	}
	wimlib_assert(0);
}

INTERNAL void lookup_table_remove(WIMLookupTable *table, WIMLookupTableEntry *entry)
{
	lookup_table_unlink(table, entry);
	free_lookup_table_entry(entry);
}

/* Decrement the reference count for @dentry in the lookup table.  The lookup
 * table entry is freed if there are no references remaining. */
INTERNAL void lookup_table_decrement_refcnt(WIMLookupTable* table, 
					const WIMDentry *dentry)
{
	u64 pos = *(u64*)dentry->hash % table->capacity;
	WIMLookupTableEntry *prev = NULL;
	WIMLookupTableEntry *entry = table->array[pos];
	WIMLookupTableEntry *next;
	while (entry != NULL) {
		next = entry->next;
		if (memcmp(dentry->hash, entry->hash, WIM_HASH_SIZE) == 0) {
			wimlib_assert(entry->refcnt != 0);
			DEBUG("'%s': %u => %u\n", dentry->full_path_utf8, 
					entry->refcnt, entry->refcnt - 1);
			if (--entry->refcnt == 0) {
				free_lookup_table_entry(entry);
				if (prev == NULL)
					table->array[pos] = next;
				else
					prev->next = next;
			}
		}
		prev = entry;
		entry = next;
	}
}

/* 
 * Looks up an entry in the lookup table.
 * 
 * @dentry:		The dentry for which to look up the entry.
 * @lookup_table:	A pointer to the lookup table.
 * @return:		A pointer to the WIMLookupTableEntry if found; otherwise
 * 				NULL.
 */
INTERNAL WIMLookupTableEntry *lookup_file_resource(const WIMDentry *dentry, 
						const WIMLookupTable *lookup_table)
{
	size_t pos = *(u64*)dentry->hash % lookup_table->capacity;
	WIMLookupTableEntry *entry = lookup_table->array[pos];
	while (entry != NULL) {
		if (memcmp(dentry->hash, entry->hash, WIM_HASH_SIZE) == 0)
			return entry;
		entry = entry->next;
	}
	return NULL;
}



/* 
 * Calls a function on all the entries in the lookup table.  Stop early and
 * return false if any call to the function returns false.
 *
 * @table:	A pointer to the lookup table.
 * @visitor:	The function to call.
 * @arg:	An argument to pass to the function.
 * @return:	True if the function returns true every time it is called;
 * 			otherwise, false.
 */
INTERNAL bool for_lookup_table_entry(WIMLookupTable *table, 
		bool (*visitor)(WIMLookupTableEntry *, void *), void *arg)
{
	WIMLookupTableEntry *entry, *next;
	for (size_t i = 0; i < table->capacity; i++) {
		entry = table->array[i];
		while (entry != NULL) {
			next = entry->next;
			if (!(*visitor)(entry, arg))
				return false;
			entry = next;
		}
	}
	return true;
}


/*
 * Reads the lookup table from a WIM file.
 *
 * @fp:  		The FILE* for the WIM file.
 * @offset:  		The offset of the lookup table resource.
 * @size:  		The size of the lookup table resource.
 * @lookup_table_ret:  	A pointer to a WIMLookupTable structure into which the
 * 				lookup table will be returned.
 * @return:		True on success, false on failure.
 */
INTERNAL bool read_lookup_table(FILE *fp, u64 offset, u64 size, 
		WIMLookupTable *lookup_table_ret)
{
	DEBUG("Reading lookup table: offset %"PRIu64", size %"PRIu64"\n",
			offset, size);
	if (fseeko(fp, offset, SEEK_SET) != 0) {
		wimlib_error("Failed to seek to byte %"PRIu64" to read lookup table: "
				"%m\n", offset);
		return false;
	}

	size_t num_entries = size / WIM_LOOKUP_TABLE_ENTRY_DISK_SIZE;

	size_t capacity = num_entries * 2 + 1;

	lookup_table_init(lookup_table_ret, capacity);

	u8 buf[WIM_LOOKUP_TABLE_ENTRY_DISK_SIZE];

	while (num_entries--) {
		if (fread(buf, 1, sizeof(buf), fp) != sizeof(buf)) {
			if (feof(fp))
				wimlib_error("Unexpected EOF in lookup table!\n");
			else
				wimlib_error("Stream read error while reading "
						"lookup table!\n");
			lookup_table_destroy(lookup_table_ret);
			return false;
		}
		const u8 *p = buf;
		WIMLookupTableEntry *cur_entry = new_lookup_table_entry();
		p = get_resource_entry(p, &cur_entry->resource_entry);
		p = get_u16(p, &cur_entry->part_number);
		p = get_u32(p, &cur_entry->refcnt);
		p = get_bytes(p, sizeof(cur_entry->hash), cur_entry->hash);

		lookup_table_insert(lookup_table_ret, cur_entry);
	}
	DEBUG("Done reading lookup table.\n");
	return true;
}


/* 
 * Writes a lookup table entry to the output file.
 *
 * @entry:   A pointer to the lookup table entry.
 * @__out:   The open output file stream to write the data to.
 * @return:  True on success, false on failure.
 */
static bool write_lookup_table_entry(WIMLookupTableEntry *entry, void *__out)
{
	/* Don't write entries that have not had file resources or metadata
	 * resources written for them. */
	if (entry->out_refcnt == 0)
		return true;
	FILE *out = (FILE*)__out;

	if (entry->output_resource_entry.flags & WIM_RESHDR_FLAG_METADATA)
		DEBUG("Writing metadata entry at %lu\n", ftello(__out));

	u8 buf[WIM_LOOKUP_TABLE_ENTRY_DISK_SIZE];
	u8 *p = buf;
	p = put_resource_entry(p, &entry->output_resource_entry);
	p = put_u16(p, entry->part_number);
	p = put_u32(p, entry->out_refcnt);
	p = put_bytes(p, sizeof(entry->hash), entry->hash);
	wimlib_assert(p - buf == WIM_LOOKUP_TABLE_ENTRY_DISK_SIZE);
	if (fwrite(buf, 1, sizeof(buf), out) != sizeof(buf)) {
		wimlib_error("Failed to write lookup table entry: %m\n");
		return false;
	}
	return true;
}

INTERNAL bool write_lookup_table(WIMLookupTable *table, FILE *out)
{
	return for_lookup_table_entry(table, write_lookup_table_entry, out);
}



/*
 * Prints an entry from the lookup table.
 *
 * @entry:	The lookup table entry.
 * @ignore:	Ignored.
 * @return:	Always true.
 */
INTERNAL bool print_lookup_table_entry(WIMLookupTableEntry *entry, void *ignore)
{
	printf("Offset            = %"PRIu64" bytes\n", 
			entry->resource_entry.offset);
	printf("Size              = %"PRIu64" bytes\n", 
			(u64)entry->resource_entry.size);
	printf("Original size     = %"PRIu64" bytes\n", 
			entry->resource_entry.original_size);
	printf("Part Number       = %hu\n", entry->part_number);
	printf("Reference Count   = %u\n", entry->refcnt);
	printf("Hash              = ");
	print_hash(entry->hash);
	putchar('\n');
	printf("Flags             = ");
	u8 flags = entry->resource_entry.flags;
	if (flags & WIM_RESHDR_FLAG_COMPRESSED)
		fputs("WIM_RESHDR_FLAG_COMPRESSED, ", stdout);
	if (flags & WIM_RESHDR_FLAG_FREE)
		fputs("WIM_RESHDR_FLAG_FREE, ", stdout);
	if (flags & WIM_RESHDR_FLAG_METADATA)
		fputs("WIM_RESHDR_FLAG_METADATA, ", stdout);
	if (flags & WIM_RESHDR_FLAG_SPANNED)
		fputs("WIM_RESHDR_FLAG_SPANNED, ", stdout);
	putchar('\n');
	if (entry->file_on_disk != NULL)
		printf("File on Disk      = \"%s\"\n", entry->file_on_disk);
	putchar('\n');
	return true;
}


static bool zero_out_refcnts(WIMLookupTableEntry *entry, void *ignore)
{
	entry->out_refcnt = 0;
	return true;
}

INTERNAL void lookup_table_zero_out_refcnts(WIMLookupTable *table)
{
	for_lookup_table_entry(table, zero_out_refcnts, NULL);
}


INTERNAL void lookup_table_destroy(WIMLookupTable *table)
{
	if (table->array != NULL) {
		for_lookup_table_entry(table, do_free_lookup_table_entry, NULL);
		free(table->array);
		table->array = NULL;
	}
}

/*
 * Prints the lookup table from the WIM file associated with @w.  Returns true on
 * success, false on failure.
 */
void wimlib_print_lookup_table(WIMStruct *w)
{
	for_lookup_table_entry(&w->lookup_table, print_lookup_table_entry, NULL);
}
