/*
 * modify.c
 *
 * Support for modifying WIM files.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib_internal.h"
#include "util.h"
#include <sys/stat.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>

/* 
 * Recursively builds a dentry tree from a directory tree on disk, outside the
 * WIM file.
 *
 * @root:  A dentry that has already been created for the root of the dentry
 * 	tree.
 * @source_path:  The path to the root of the tree on disk. 
 * @root_stat:   A pointer to a `struct stat' that contains the metadata for the
 * 	root of the tree on disk. 
 * @lookup_table: The lookup table for the WIM file.  For each file added to the
 * 		dentry tree being built, an entry is added to the lookup table, 
 * 		unless an identical file is already in the lookup table.  These
 * 		lookup table entries that are added point to the file on disk.
 *
 * @return:	True on success, false on failure.  It is a failure if any of
 *		the files cannot be `stat'ed, or if any of the needed
 *		directories cannot be opened or read.  Failure to add the files
 *		to the WIM may still occur later when trying to actually read 
 *		the regular files in the tree into the WIM as file resources.
 */
static bool build_dentry_tree(WIMDentry *root, const char *source_path, 
			struct stat *root_stat, WIMLookupTable* lookup_table)
{
	stbuf_to_dentry(root_stat, root);

	if (dentry_is_directory(root)) {

		/* Open the directory on disk */
		DIR *dir = opendir(source_path);
		if (dir == NULL) {
			wimlib_error("Failed to open the directory `%s': %m\n",
					source_path);
			return false;
		}

		/* Buffer for names of files in directory. */
		int len = strlen(source_path);
		char name[len + 1 + FILENAME_MAX + 1];
		memcpy(name, source_path, len);
		name[len] = '/';
		struct dirent *p;
		errno = 0;

		/* Create a dentry for each entry in the directory on disk, and recurse
		 * to any subdirectories. */
		while ((p = readdir(dir)) != NULL) {
			if (p->d_name[0] == '.' && (p->d_name[1] == '\0'
				|| (p->d_name[1] == '.' && p->d_name[2] == '\0')))
					continue;
			strcpy(name + len + 1, p->d_name);
			struct stat child_stat;
			if (stat(name, &child_stat) != 0) {
				wimlib_error("cannot stat `%s': %m\n");
				closedir(dir);
				return false;
			}
			WIMDentry *child = create_dentry(p->d_name);
			if (!build_dentry_tree(child, name, &child_stat, lookup_table)) {
				closedir(dir);
				return false;
			}
			link_dentry(child, root);
		}
		closedir(dir);
		if (errno != 0) {
			wimlib_error("Failed to read the directory `%s': %s\n",
					name, errno);
			return false;
		}
	} else {
		/* For each non-directory, we must check to see if the file is
		 * in the lookup table already; if it is, we increment its
		 * refcnt; otherwise, we create a new lookup table entry and
		 * insert it. */
		if (!sha1sum(source_path, root->hash)) {
			wimlib_error("Failed to calculate sha1sum for file `%s'\n", 
						source_path);
			return false;
		}
		#ifdef ENABLE_DEBUG
		/*DEBUG("Calculated sha1sum for file `%s': ", source_path);*/
		/*print_byte_field(root->hash, WIM_HASH_SIZE);*/
		/*putchar('\n');*/
		#endif

		WIMLookupTableEntry *entry = lookup_file_resource(root, 
								lookup_table);
		if (entry == NULL) {
			/*DEBUG("Adding lookup table entry for `%s'\n", source_path);*/
			entry               = new_lookup_table_entry();
			entry->file_on_disk = xstrdup(source_path);
			entry->resource_entry.flags = 0;
			entry->refcnt       = 1;
			entry->part_number  = 1;
			entry->resource_entry.original_size = root_stat->st_size;
			entry->resource_entry.size = root_stat->st_size;
			memcpy(entry->hash, root->hash, WIM_HASH_SIZE);
			lookup_table_insert(lookup_table, entry);
		} else {
			/*DEBUG("Found existing lookup table entry for `%s'\n", */
								/*source_path);*/
			entry->refcnt++;
		}
	}
	return true;
}

struct clone_entry_params {
	WIMStruct *src_w;
	WIMStruct *dest_w;
};

/* 
 * This function takes in a dentry that was previously located only in image(s)
 * in @src_w, but now is being added to @dest_w. If there is in fact already a
 * lookup table entry for this file in the lookup table of the destination WIM
 * file, we simply increment its reference count.  Otherwise, a new lookup table
 * entry is created that references the location of the file resource in the
 * source WIM file through the other_wim_fp field of the lookup table entry.
 */
static bool add_lookup_table_entry_to_dest_wim(WIMDentry *dentry, void *arg)
{
	WIMStruct *src_w, *dest_w;
	src_w = ((struct clone_entry_params*)arg)->src_w;
	dest_w = ((struct clone_entry_params*)arg)->dest_w;

	if (dentry_is_directory(dentry))
		return true;

	WIMLookupTableEntry *src_table_entry = 
			lookup_file_resource(dentry, &src_w->lookup_table);
	if (src_table_entry == NULL)
		return true;

	WIMLookupTableEntry *dest_table_entry = 
			lookup_file_resource(dentry, &dest_w->lookup_table);
	if (dest_table_entry != NULL) {
		dest_table_entry->refcnt++;
		return true;
	}
	dest_table_entry = new_lookup_table_entry();
	dest_table_entry->other_wim_fp = src_w->fp;
	dest_table_entry->other_wim_ctype = wimlib_get_compression_type(src_w);
	dest_table_entry->refcnt = 1;
	memcpy(&dest_table_entry->resource_entry, &src_table_entry->resource_entry,
			sizeof(WIMResourceEntry));
	memcpy(dest_table_entry->hash, dentry->hash, WIM_HASH_SIZE);
	lookup_table_insert(&dest_w->lookup_table, dest_table_entry);
	return true;
}

/*
 * Adds an image (given by its dentry tree) to the boot metadata array of a WIM
 * file, adds an entry to the lookup table for the boot metadata, updates the
 * image count in the header, and selects the new image. 
 *
 * Does not update the XML data.
 *
 * @w:		  The WIMStruct for the WIM file.
 * @root_dentry:  The root of the directory tree for the image.
 */
static void add_new_dentry_tree(WIMStruct *w, WIMDentry *root_dentry)
{
	WIMLookupTableEntry *bmd_lookup_entry = new_lookup_table_entry();
	bmd_lookup_entry->resource_entry.flags = WIM_RESHDR_FLAG_METADATA;
	randomize_byte_array(bmd_lookup_entry->hash, WIM_HASH_SIZE);
	lookup_table_insert(&w->lookup_table, bmd_lookup_entry);

	/* Reallocate the array of boot metadata. */
	w->hdr.image_count++;
	DEBUG("Reallocing boot metadata for image_count = %u\n",
			w->hdr.image_count);

	w->boot_metadata = xrealloc(w->boot_metadata, w->hdr.image_count * 
			sizeof(WIMBootMetadata));

	WIMBootMetadata *new_bmd = &w->boot_metadata[w->hdr.image_count - 1];

	/* Initialize the boot metadata. */
	init_security_data(&new_bmd->security_data);
	new_bmd->lookup_table_entry = bmd_lookup_entry;
	new_bmd->modified           = true;
	new_bmd->root_dentry        = root_dentry;

	/* Change the current image to the new one. */
	wimlib_select_image(w, w->hdr.image_count);
}

/*
 * Exports an image, or all the images (specified by @src_image, which may be
 * WIM_ALL_IMAGES) to the WIM specified by @dest_w.
 *
 * @dest_image_name specifies the new name of the source image when it is copied
 * to the destination WIM.  If @dest_image_name is NULL, the destination names
 * for the image(s) are taken to be the same as the source names for the
 * image(s).  @dest_image_name must be NULL if (src_image == WIM_ALL_IMAGES).
 *
 * @boot specifies whether the image being exported is to be made bootable.  If
 * src_image is WIM_ALL_IMAGES and there are multiple images in the source WIM,
 * specifying boot = true is valid only if one of the exported images is
 * currently marked as bootable in the source WIM; if that is the case, then
 * that image is marked as bootable in the destination WIM.
 *
 * @check specifies whether to check the image for correctness.
 *
 * Return true on success, false on error.
 */
bool wimlib_export_image(WIMStruct *src_w, int src_image, WIMStruct *dest_w,
			const char *dest_image_name, const char *dest_description, 
			bool boot)
{
	if (src_image == WIM_ALL_IMAGES) {
		if (src_w->hdr.image_count > 1) {

			/* multi-image export. */

			int boot_idx;
			if (boot) {

				/* Specifying @boot on a multi-image source WIM
				 * makes the boot index default to the bootable
				 * image in the source WIM.  It is an error if
				 * there is no such bootable image. */

				if (src_w->hdr.boot_idx == 0) {
					wimlib_error("Cannot specify `boot' flag "
							"when exporting multiple "
							"images from a WIM with no "
							"bootable images!\n");
					return false;
				} else {
					boot_idx = src_w->hdr.boot_idx;
				}
			}
			if (dest_image_name != NULL || dest_description != NULL) {
				wimlib_error("Image name or image description "
						"was specified, but we are exporting "
						"multiple images!\n");
				return false;
			}
			for (int i = 1; i <= src_w->hdr.image_count; i++) {
				if (!wimlib_export_image(src_w, i, dest_w, 
							NULL, dest_description,
							(boot && (i == boot_idx))))
					return false;
			}
			return true;
		} else {
			src_image = 1; 
		}
	}

	wimlib_assert(src_image >= 1 && src_image <= src_w->hdr.image_count);


	if (!wimlib_select_image(src_w, src_image)) {
		wimlib_error("Could not select image %d from the WIM `%s' "
				"to export it!\n", src_image, src_w->filename);
		return false;
	}

	if (dest_image_name == NULL) {
		dest_image_name = wimlib_get_image_name(src_w, src_image);
		DEBUG("Using name `%s' for source image %d\n", 
				dest_image_name, src_image);
	}

	DEBUG("Exporting image %d from `%s'\n", src_image, src_w->filename);

	if (wimlib_image_name_in_use(dest_w, dest_image_name)) {
		wimlib_error("There is already an image named `%s' "
				"in the destination WIM!\n", dest_image_name);
		return false;
	}


	WIMDentry *root = wim_root_dentry(src_w);

	increment_dentry_tree_refcnt(root);

	struct clone_entry_params wims;
	wims.src_w = src_w;
	wims.dest_w = dest_w;
	for_dentry_in_tree(root, add_lookup_table_entry_to_dest_wim, &wims);

	add_new_dentry_tree(dest_w, root);

	if (boot) {
		DEBUG("Setting boot_idx to %d\n", dest_w->hdr.image_count);
		dest_w->hdr.boot_idx = dest_w->hdr.image_count;
	}

	xml_export_image(src_w->xml_info, src_image, &dest_w->xml_info,
			dest_image_name, dest_description);

	return true;
}

/* 
 * Deletes an image from the WIM. 
 */
bool wimlib_delete_image(WIMStruct *w, int image)
{
	if (image == WIM_ALL_IMAGES) {
		int num_images = w->hdr.image_count;
		for (int i = 1; i <= num_images; i++) {
			/* Always delete the first image, since by the end
			 * there won't be any more than that!  */
			if (!wimlib_delete_image(w, 1))
				return false;
		}
		return true;
	}

	wimlib_assert(image >= 1 && image <= w->hdr.image_count);

	DEBUG("Deleting image %d\n", image);

	/* Even if the dentry tree is not allocated, we must select it (and
	 * therefore allocate it) so that we can decrement the reference counts
	 * in the lookup table.  */
	if (!wimlib_select_image(w, image))
		return false;


	/* Free the dentry tree, any lookup table entries that have their
	 * refcnt decremented to 0, and the security data. */
	WIMBootMetadata *bmd = wim_get_current_boot_metadata(w);
	free_dentry_tree(bmd->root_dentry, &w->lookup_table, true);
	destroy_security_data(&bmd->security_data);

	/* Get rid of the lookup table entry for the boot metadata */
	lookup_table_remove(&w->lookup_table, bmd->lookup_table_entry);

	/* Get rid of the empty slot in the boot metadata array. */
	for (int i = image - 1; i < w->hdr.image_count - 1; i++)
		memcpy(&w->boot_metadata[i], &w->boot_metadata[i + 1],
				sizeof(WIMBootMetadata));

	/* Decrement the image count. */
	w->hdr.image_count--;
	if (w->hdr.image_count == 0) {
		free(w->boot_metadata);
		w->boot_metadata = NULL;
	}

	/* Fix the boot index. */
	if (w->hdr.boot_idx == image)
		w->hdr.boot_idx = 0;
	else if (w->hdr.boot_idx > image)
		w->hdr.boot_idx--;

	w->current_image = WIM_NO_IMAGE;

	/* Remove the image from the XML information. */
	xml_delete_image(&w->xml_info, image);

	return true;
}

/* 
 * Adds an image to a WIM file from a directory tree on disk.
 * You must specif either @name and @desc, or @wim_info and @image.
 *
 * @w:  	The WIMStruct for the WIM file.
 * @path:  	The path to a directory in the outside filesystem; its contents will
 *  			be the contents of the image.
 * @name:	The name of the image.
 * @desc:	The description of the image.
 * @flags:	What to put in the <FLAGS> element for the image's XML data.
 * @return:	True on success, false on failure.
 */
bool wimlib_add_image(WIMStruct *w, const char *path, 
			const char *name, const char *desc, 
			const char *flags, bool boot)
{
	DEBUG("Adding dentry tree from path `%s'\n", path);

	if (wimlib_image_name_in_use(w, name)) {
		wimlib_error("There is already an image named `%s' in %s!\n",
				name, w->filename);
		return false;
	}

	DEBUG("Creating root_dentry.\n");


	WIMDentry *root_dentry = create_dentry("");
	calculate_dentry_full_path(root_dentry);
	dentry_set_is_directory(root_dentry);

	/* Construct the dentry tree from the outside filesystem. */
	struct stat root_stat;
	if (stat(path, &root_stat) != 0) {
		wimlib_error("Failed to stat `%s': %m\n", path);
		return false;
	}
	if (!S_ISDIR(root_stat.st_mode)) {
		wimlib_error("`%s' is not a directory!\n", path);
		return false;
	}
	DEBUG("Building dentry tree.\n");
	if (!build_dentry_tree(root_dentry, path, &root_stat, &w->lookup_table)) {
		wimlib_error("Failed to build dentry tree for `%s'!\n", path);
		free_dentry_tree(root_dentry, &w->lookup_table, true);
		return false;
	}

	DEBUG("Recalculating full paths of dentries.\n");
	for_dentry_in_tree(root_dentry, do_calculate_dentry_full_path, NULL);

	add_new_dentry_tree(w, root_dentry);

	if (boot)
		w->hdr.boot_idx = w->hdr.image_count;

	xml_add_image(w, root_dentry, name, desc, flags);

	return true;
}
