/*
 * resource.c
 *
 * Read file resources and boot metadata.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib_internal.h"
#include "lzx.h"
#include "io.h"
#include <unistd.h>
#include <errno.h>

/* 
 * Reads a compressed file resource.
 *
 * @fp:      		The FILE* for the WIM file, positioned at the start
 * 				of the file resource.
 * @resource_compressed_size:  	 The total compressed size of the resource.  
 * @resource_uncompressed_size:  The total uncompressed size of the resource.
 * @resource_offset:		 The offset of the start of the resource in the
 * 					stream @fp.
 * @resource_ctype:	The compression ctype. XXX Currently only LZX
 * 				is supported.
 * @len:		The number of bytes of uncompressed data to
 * 				read.
 * @offset:		The offset of the bytes to read, within the uncompressed
 * 				resource.
 * @contents_len:	An array that must have length of at least @len. The 
 * 				uncompressed data is written into it.
 * @return:  		True on success, false on failure.  Failures may be
 * 				caused by errors in fseeko(), fread(), or
 * 				due to the compressed data being invalid.
 */
static bool read_compressed_resource(FILE *fp, u64 resource_compressed_size, 
				u64 resource_uncompressed_size, u64 resource_offset,
				int resource_ctype, u64 len, u64 offset, 
				u8  contents_ret[])
{
	DEBUG2("resource_uncompressed_size = %"PRIu64", "
			"len = %"PRIu64", offset = %"PRIu64"\n",
			resource_uncompressed_size,
			len,
			offset);

	/* XXX need XPRESS support */
	wimlib_assert(resource_ctype == WIM_COMPRESSION_TYPE_LZX);

	/* Trivial case */
	if (len == 0)
		return true;


	/* The structure of a compressed resource consists of a table of chunk
	 * offsets followed by the chunks themselves.  Each chunk consists of
	 * compressed data, and there is one chunk for each WIM_CHUNK_SIZE =
	 * 32768 bytes of the uncompressed file, with the last chunk having any
	 * remaining bytes.
	 *
	 * The chunk offsets are measured relative to the end of the chunk
	 * table.  The first chunk is omitted from the table in the WIM file
	 * because its offset is implicitly given by the fact that it directly
	 * follows the chunk table and therefore must have an offset of 0. 
	 */

	/* Calculate how many chunks the resource conists of in its entirety. */
	u64 num_chunks = (resource_uncompressed_size + WIM_CHUNK_SIZE - 1) /
								WIM_CHUNK_SIZE;
	/* As mentioned, the first chunk has no entry in the chunk table. */
	u64 num_chunk_entries = num_chunks - 1;


	/* The index of the chunk that the read starts at. */
	u64 start_chunk = offset / WIM_CHUNK_SIZE;
	/* The byte offset at which the read starts, within the start chunk. */
	u64 start_chunk_offset = offset % WIM_CHUNK_SIZE;

	/* The index of the chunk that contains the last byte of the read. */
	u64 end_chunk   = (offset + len - 1) / WIM_CHUNK_SIZE;
	/* The byte offset of the last byte of the read, within the end chunk */
	u64 end_chunk_offset = (offset + len - 1) % WIM_CHUNK_SIZE;

	/* Number of chunks that are actually needed to read the requested part
	 * of the file. */
	u64 num_needed_chunks = end_chunk - start_chunk + 1;
	/* If the end chunk is not the last chunk, an extra chunk entry is
	 * needed because we need to know the offset of the chunk after the last
	 * chunk read to figure out the size of the last read chunk. */
	if (end_chunk != num_chunks - 1)
		num_needed_chunks++;

	/* Declare the chunk table.  It will only contain offsets for the chunks
	 * that are actually needed for this read. */
	u64 chunk_offsets[num_needed_chunks];

	/* Set the implicit offset of the first chunk if it is included in the
	 * needed chunks.
	 *
	 * Note: M$'s documentation includes a picture that shows the first
	 * chunk starting right after the chunk entry table, labeled as offset
	 * 0x10.  However, in the actual file format, the offset is measured
	 * from the end of the chunk entry table, so the first chunk has an
	 * offset of 0. */
	if (start_chunk == 0)
		chunk_offsets[0] = 0;

	/* According to M$'s documentation, if the uncompressed size of
	 * the file is greater than 4 GB, the chunk entries are 8-byte
	 * integers.  Otherwise, they are 4-byte integers. */
	u64 chunk_entry_size = (resource_uncompressed_size >= (u64)1 << 32) ? 
									8 : 4;

	/* Size of the full chunk table in the WIM file. */
	u64 chunk_table_size = chunk_entry_size * num_chunk_entries;

	/* Read the needed chunk offsets from the table in the WIM file. */
	if (end_chunk != 0) {

		/* Index, in the WIM file, of the first needed entry in the
		 * chunk table. */
		u64 start_table_idx = (start_chunk == 0) ? 0 : start_chunk - 1;

		/* Number of entries we need to actually read from the chunk
		 * table (excludes the implicit first chunk). */
		u64 num_needed_chunk_entries = (start_chunk == 0) ? 
					num_needed_chunks - 1 : num_needed_chunks;

		/* Skip over unneeded chunk table entries. */
		u64 file_offset_of_needed_chunk_entries = resource_offset + 
					start_table_idx * chunk_entry_size;
		if (fseeko(fp, file_offset_of_needed_chunk_entries, 
					SEEK_SET) != 0) 
		{
			wimlib_error("Failed to seek to byte %"PRIu64" "
					"to read chunk table of compressed "
					"resource: %m\n", 
					file_offset_of_needed_chunk_entries);
			return false;
		}

		/* Number of bytes we need to read from the chunk table. */
		size_t size = num_needed_chunk_entries * chunk_entry_size;

		u8 chunk_tab_buf[size];

		if (fread(chunk_tab_buf, 1, size, fp) != size)
			goto read_compressed_resource_fread_error;

		/* Now fill in chunk_offsets from the entries we have read in
		 * chunk_tab_buf. */

		u64 *chunk_tab_p = chunk_offsets;
		if (start_chunk == 0)
			chunk_tab_p++;

		if (chunk_entry_size == 4) {
			u32 *entries = (u32*)chunk_tab_buf;
			while (num_needed_chunk_entries--)
				*chunk_tab_p++ = to_le32(*entries++);
		} else {
			u64 *entries = (u64*)chunk_tab_buf;
			while (num_needed_chunk_entries--)
				*chunk_tab_p++ = to_le64(*entries++);
		}
	}

	/* Done with the chunk table now.  We must now seek to the first chunk
	 * that is needed for the read. */

	u64 file_offset_of_first_needed_chunk = resource_offset + 
				chunk_table_size + chunk_offsets[0];
	if (fseeko(fp, file_offset_of_first_needed_chunk, SEEK_SET) != 0) 
	{
		wimlib_error("Failed to seek to byte %"PRIu64" "
				"to read first chunk of compressed "
				"resource: %m\n", 
				file_offset_of_first_needed_chunk);
		return false;
	}

	/* Pointer to current position in the output buffer for uncompressed
	 * data. */
	u8 *out_p = (u8*)contents_ret;

	/* Buffer for compressed data.  While most compressed chunks will have a
	 * size much less than WIM_CHUNK_SIZE, WIM_CHUNK_SIZE - 1 is the maximum
	 * size in the worst-case.  This assumption is valid only if chunks that
	 * happen to compress to more than the uncompressed size (i.e. a
	 * sequence of random bytes) are always stored uncompressed. But this seems
	 * to be the case in M$'s WIM files, even though it is undocumented. */
	u8 compressed_buf[WIM_CHUNK_SIZE - 1];

	/* Initialize the LZX decompressor. */
	struct LZXstate *lzx_state = LZXinit(15);

	/* Decompress all the chunks. */
	for (u64 i = start_chunk; i <= end_chunk; i++) {

		DEBUG2("Chunk %"PRIu64" (%"PRIu64" of %"PRIu64")\n", 
				i, i - start_chunk + 1, 
				end_chunk - start_chunk + 1);

		/* Calculate the sizes of the compressed chunk and of the
		 * uncompressed chunk. */
		unsigned compressed_chunk_size, uncompressed_chunk_size;
		if (i != num_chunks - 1) {
			/* All the chunks except the last one in the resource
			 * expand to WIM_CHUNK_SIZE uncompressed, and the amount
			 * of compressed data for the chunk is given by the
			 * difference of offsets in the chunk offset table. */
			compressed_chunk_size = chunk_offsets[i + 1 - start_chunk] - 
						chunk_offsets[i - start_chunk];
			uncompressed_chunk_size = WIM_CHUNK_SIZE;
		} else {
			/* The last compressed chunk consists of the remaining
			 * bytes in the file resource, and the last uncompressed
			 * chunk has size equal to however many bytes are left-
			 * that is, the remainder of the uncompressed size when
			 * divided by WIM_CHUNK_SIZE. */
			compressed_chunk_size = resource_compressed_size - 
						chunk_offsets[i - start_chunk];

			uncompressed_chunk_size = resource_uncompressed_size % 
								WIM_CHUNK_SIZE;
			/* If the remainder is 0, the last chunk actually
			 * uncompresses to a full WIM_CHUNK_SIZE bytes. */
			if (uncompressed_chunk_size == 0)
				uncompressed_chunk_size = WIM_CHUNK_SIZE;

			DEBUG2("compressed_chunk_size = %u, uncompressed_chunk_size "
					"= %u\n", compressed_chunk_size, 
					uncompressed_chunk_size);

			/* In an undocumented case, the compressed chunk size of
			 * the last chunk can be greater than or equal to the
			 * uncompressed chunk size, and in this case, the data
			 * is not compressed. 
			 *
			 * Note this means that if the compressed_chunk_size is
			 * greater than the uncompressed_chunk_size, there are
			 * extra bytes on the end that simply are not used for
			 * anything, so we can just set compressed_chunk_size =
			 * uncompressed_chunk_size. */
			if (compressed_chunk_size >= uncompressed_chunk_size)
				compressed_chunk_size = uncompressed_chunk_size;

		}

		/* Figure out how much of this chunk we actually need to read */
		u64 start_offset;
		if (i == start_chunk)
			start_offset = start_chunk_offset;
		else
			start_offset = 0;
		u64 end_offset;
		if (i == end_chunk)
			end_offset = end_chunk_offset;
		else
			end_offset = WIM_CHUNK_SIZE - 1;

		u64 partial_chunk_size = end_offset + 1 - start_offset;
		bool is_partial_chunk = partial_chunk_size != compressed_chunk_size;

		/* This is undocumented, but chunks can be uncompressed.  This
		 * appears to be the case when the compressed chunk size is
		 * WIM_CHUNK_SIZE (32768) for non-last chunks, or when the
		 * compressed chunk size is greater than or equal to the
		 * uncompressed chunk size for the last chunk. */
		if (compressed_chunk_size == 32768 || (i == num_chunks - 1 && 
				compressed_chunk_size >= uncompressed_chunk_size)) 
		{
			/* Uncompressed chunk */

			if (start_offset != 0) {
				if (fseeko(fp, start_offset, SEEK_SET) != 0) {
					wimlib_error("Uncompressed partial chunk "
							"fseek() error: %m\n");
					return false;
				}
			}
			if (fread(out_p, 1, partial_chunk_size, fp) != 
					partial_chunk_size)
				goto read_compressed_resource_fread_error;
		} else {
			/* Compressed chunk */

			/* Read the compressed data into compressed_buf. */
			if (fread(compressed_buf, 1, compressed_chunk_size, 
						fp) != compressed_chunk_size)
				goto read_compressed_resource_fread_error;

			/* Return value from LZXdecompress() */
			int status;

			/* For partial chunks we must buffer the uncompressed
			 * data because we don't need all of it. */
			if (is_partial_chunk) {
				u8 uncompressed_buf[uncompressed_chunk_size];
				status = LZXdecompress(lzx_state, compressed_buf,
						uncompressed_buf, compressed_chunk_size,
						uncompressed_chunk_size);
				memcpy(out_p, uncompressed_buf + start_offset,
						partial_chunk_size);
			} else {
				status = LZXdecompress(lzx_state, compressed_buf, 
					out_p, compressed_chunk_size, 
					uncompressed_chunk_size);
			}

			if (status != DECR_OK) {
				wimlib_error("Error during LZX decompression.\n");
				switch (status) {
				case DECR_ILLEGALDATA:
					wimlib_error("DECR_ILLEGALDATA\n");
					break;
				case DECR_DATAFORMAT:
					wimlib_error("DECR_DATAFORMAT\n");
					break;
				}
				return false;
			}

			LZXreset(lzx_state);
		}

		/* Advance the pointer into the uncompressed output data by the
		 * number of uncompressed bytes that were written.  */
		out_p += partial_chunk_size;
	}

	LZXteardown(lzx_state);

	return true;

read_compressed_resource_fread_error:
	if (feof(fp))
		wimlib_error("Unexpected EOF in compressed file resource\n");
	else
		wimlib_error("Error reading compressed file resource: %m\n");
	return false;
}

/* 
 * Reads uncompressed data from an open file stream.
 *
 * @fp:      		The FILE* for the WIM file.
 * @offset:		The offset of the data within the file.
 * @len:    		The length of the data to read, in bytes.
 * @contents_ret:       An array into which the data is written. It must have 
 * 			length of at least @len bytes. 
 *
 * @return:  		True on success, false on failure (errors in fseeko() or
 * 			fread()).
 */
static bool read_uncompressed_resource(FILE *fp, u64 offset, u64 len, 
					u8 contents_ret[])
{
	if (fseeko(fp, offset, SEEK_SET) == -1) {
		wimlib_error("Failed to seek to byte %"PRIu64" of input file "
				"to read uncompressed resource "
				"(len = %"PRIu64")!\n", offset, len);
		return false;
	}
	if (fread(contents_ret, 1, len, fp) != len) {
		if (feof(fp))
			wimlib_error("Unexpected EOF in uncompressed file resource!\n");
		else
			wimlib_error("Failed to read %"PRIu64" bytes from "
					"uncompressed resource at offset "
					"%"PRIu64"\n", len, offset);
		return false;
	}
	return true;
}

/* 
 * Reads a file resource.
 *
 * @fp:  		The FILE* for the WIM file.
 * @resource_size:		The compressed size of the resource.
 * @resource_original_size:	The uncompressed size of the resource.
 * @resource_offset:		The offset of the resource in the stream @fp.
 * @resource_ctype:		The compression type of the resource.
 * 				(WIM_COMPRESSION_TYPE_*)
 * @len:		How many bytes of the resource should be read.
 * @offset:        	The offset within the resource at which the read
 * 				will occur.
 *
 * 			To read the whole file resource, specify offset =
 * 			0 and len = resource_original_size, or call
 * 			read_full_resource().
 *
 * @contents_ret:  	An array, that must have length at least @len,
 * 				into which the uncompressed contents of
 * 				the file resource starting at @offset and 
 * 				continuing for @len bytes will be written.
 *
 * @return:  		True on success, false on failure. Failure may be due to
 * 			being unable to read the data from the WIM file at the
 * 			specified length and offset, or it may be due to the
 * 			compressed data (if the data is compressed) being
 * 			invalid.
 */
INTERNAL bool read_resource(FILE *fp, u64 resource_size, 
				u64 resource_original_size,
				u64 resource_offset, int resource_ctype, 
				u64 len, u64 offset, void *contents_ret)
{
	wimlib_assert(len + offset <= resource_original_size);

	if (resource_ctype == WIM_COMPRESSION_TYPE_NONE) {
		wimlib_assert(resource_size == resource_original_size);
		return read_uncompressed_resource(fp, 
				resource_offset + offset, 
				len, contents_ret);
	} else {
		return read_compressed_resource(fp, resource_size,
				resource_original_size, resource_offset,
				resource_ctype, len, offset, contents_ret);
	}
}

INTERNAL bool read_full_resource(FILE *fp, u64 resource_size, 
				u64 resource_original_size,
				u64 resource_offset, 
				int resource_ctype, void *contents_ret)
{
	return read_resource(fp, resource_size, resource_original_size, 
				resource_offset, resource_ctype,
				resource_original_size, 0, contents_ret);
}

INTERNAL void *get_wim_resource(WIMStruct *w, const WIMResourceEntry *entry)
{
	u8 *buf = xmalloc(entry->original_size);
	int ctype = wim_resource_compression_type(w, entry);
	if (!read_full_resource(w->fp, entry->size, entry->original_size,
			entry->offset, ctype, buf)) 
	{
		free(buf);
		buf = NULL;
	}
	return buf;
}

/* Like extract_full_resource_to_fd(), but only extract @size bytes of the
 * uncompressed file */
INTERNAL bool extract_resource_to_fd(WIMStruct *w, 
				const WIMResourceEntry *entry, int fd,
				u64 size)
{
	errno = 0;

	u64 num_chunks = size / WIM_CHUNK_SIZE;
	u64 n;
	u8 buf[WIM_CHUNK_SIZE];
	int res_ctype = wim_resource_compression_type(w, entry);
	u64 offset = 0;
	for (u64 i = 0; i < num_chunks; i++) {
		n = (i == num_chunks - 1) ? size % WIM_CHUNK_SIZE : WIM_CHUNK_SIZE;

		if (!read_resource(w->fp, entry->size, entry->original_size,
				entry->offset, res_ctype, n, offset, buf))
			return false;
		offset += n;
		if (write(fd, buf, n) != n)
			return false;
	}
	return true;
}

/* Extracts the file resource specified by @entry to the open file @fd.  
 * Returns false and sets errno on error. */
INTERNAL bool extract_full_resource_to_fd(WIMStruct *w, 
				const WIMResourceEntry *entry, int fd)
{
	return extract_resource_to_fd(w, entry, fd, entry->original_size);
}

/* Reads the contents of a WIMResourceEntry, as represented in the on-disk
 * format, from the memory pointed to by @p, and fills in the fields of @entry.
 * A pointer to the byte after the memory read at @p is returned. */
INTERNAL const u8 *get_resource_entry(const u8 *p, WIMResourceEntry *entry)
{
	u64 size;
	u8 flags;
	p = get_u56(p, &size);
	p = get_u8(p, &flags);
	entry->size = size;
	entry->flags = flags;
	p = get_u64(p, &entry->offset);
	p = get_u64(p, &entry->original_size);
	return p;
}

/* Copies the WIMResourceEntry @entry to the memory pointed to by @p in the
 * on-disk format.  A pointer to the byte after the memory written at @p is
 * returned. */
INTERNAL u8 *put_resource_entry(u8 *p, const WIMResourceEntry *entry)
{
	p = put_u56(p, entry->size);
	p = put_u8(p, entry->flags);
	p = put_u64(p, entry->offset);
	p = put_u64(p, entry->original_size);
	return p;
}

/* Given the compression type for the WIM file as a whole as the flags field of
 * a resource entry, returns the compression type for that resource entry. */
INTERNAL int resource_compression_type(int wim_ctype, int reshdr_flags)
{
	if (wim_ctype == WIM_COMPRESSION_TYPE_NONE) {
		wimlib_assert(!(reshdr_flags & WIM_RESHDR_FLAG_COMPRESSED));
		return WIM_COMPRESSION_TYPE_NONE;
	} else {
		if (reshdr_flags & WIM_RESHDR_FLAG_COMPRESSED)
			return wim_ctype;
		else
			return WIM_COMPRESSION_TYPE_NONE;
	}
}

/* 
 * Reads the boot metadata from the WIM file.  The boot metadata doesn't really
 * have much to do with booting at all; it is also just called the 'metadata
 * resource'.  The metadata resource consists of the security data, followed by
 * the directory entry for the root directory, followed by all the other
 * directory entries in the filesystem.  The subdir_offset field of each
 * directory entry gives the start of its child entries from the beginning of
 * the metadata resource.  An end-of-directory is signaled by a directory entry
 * of length '0', really of length 8, because that's how long the 'length' field
 * is.
 *
 * @fp:		The FILE* for the input WIM file.
 * @res_entry:	The resource entry for the metadata resource (a.k.a the metadata
 * 			for the metadata)
 * @wim_ctype:	The compression type of the WIM file.
 * @sd:		A pointer to a WIMSecurityData structure into which the security
 * 		data for the WIM file is returuned.
 * @root_dentry_p:	A pointer to a pointer to a WIMDentry structure into which the 
 * 		root dentry is allocated and returned.
 *
 * @return:	True on success, false on failure.
 */
INTERNAL bool read_boot_metadata(FILE *fp, const WIMResourceEntry *res_entry,
		int wim_ctype, WIMSecurityData *sd, WIMDentry **root_dentry_p)
{
	DEBUG("Reading metadata resource: length = %lu, offset = %lu\n",
			res_entry->original_size, res_entry->offset);

	/* Allocate memory for the uncompressed metadata resource. */
	u8 *buf = xmalloc(res_entry->original_size);

	/* Set root dentry to NULL by default, so we know not to free it if
	 * there is an error. */
	*root_dentry_p = NULL;

	/* Determine the compression type of the metadata resource. */
	int ctype = resource_compression_type(wim_ctype, res_entry->flags);

	/* Read the metadata resource into memory.  (It may be compressed.) */
	if (!read_full_resource(fp, res_entry->size, res_entry->original_size, 
				res_entry->offset, ctype, buf))
		goto read_boot_metadata_error;

	DEBUG("Finished reading metadata resource into memory.\n");

	/* Read the security data into a WIMSecurityData structure. */
	if (!read_security_data(buf, res_entry->original_size, sd))
		goto read_boot_metadata_error;

	/* Read the root directory entry into a WIMDentry structure.  It
	 * starts after security data, on an 8-byte aligned address. */
	u64 dentry_offset = sd->total_length;
	dentry_offset += (8 - dentry_offset % 8) % 8;
	*root_dentry_p = xmalloc(sizeof(WIMDentry));
	if (!read_dentry(buf, res_entry->original_size, 
				dentry_offset, *root_dentry_p))
		goto read_boot_metadata_error;

	/* This is the root dentry, so set its pointers correctly. */
	(*root_dentry_p)->parent = *root_dentry_p;
	(*root_dentry_p)->next   = *root_dentry_p;
	(*root_dentry_p)->prev   = *root_dentry_p;

	/* Now read the entire directory entry tree. */
	if (!read_dentry_tree(buf, res_entry->original_size, *root_dentry_p))
		goto read_boot_metadata_error;

	/* Calculate the full paths in the dentry tree. */
	for_dentry_in_tree(*root_dentry_p, do_calculate_dentry_full_path, NULL);

	/* Free the buffer for the metadata resource and return successfully. */
	free(buf);
	return true;
read_boot_metadata_error:
	free(buf);
	free(*root_dentry_p);
	return false;
}


