/*
 * util.h
 *
 * Header for util.c.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#ifndef _WIMLIB_UTIL_H
#define _WIMLIB_UTIL_H

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <assert.h>
#include <time.h>
#include "config.h"


typedef uint8_t  u8;
typedef uint16_t u16;
typedef uint32_t u32;
typedef uint64_t u64;

#ifdef __GNU_C__
#	define offsetof(type, member) \
		__builtin_offsetof(type, member)
#else
#	define offsetof(type, member) \
		((unsigned long)&(((type*)0)->member) - 0)
#endif

#define min(a, b) ({ typeof(a) __a = (a); typeof(b) __b = (b); \
					(__a < __b) ? __a : __b; })
#define max(a, b) ({ typeof(a) __a = (a); typeof(b) __b = (b); \
					(__a > __b) ? __a : __b; })

#define ARRAY_LEN(array) (sizeof(array) / sizeof((array)[0]))



#if defined(ENABLE_DEBUG) || defined(ENABLE_MORE_DEBUG)
#	define DEBUG(format, ...)  \
({ \
	fprintf(stdout, "[%s %d] %s(): ", __FILE__, __LINE__, __func__); \
	fprintf(stdout, format, ## __VA_ARGS__); \
	fflush(stdout); \
	})

#else
#	define DEBUG(format, ...)
#endif

#ifdef ENABLE_MORE_DEBUG
#	define DEBUG2(format, ...) DEBUG(format, ## __VA_ARGS__)
#else
#	define DEBUG2(format, ...)
#endif

#ifdef ENABLE_ASSERTIONS
#	define wimlib_assert(expr) assert(expr)
//#	define wimlib_assert(expr) ({ if (!(expr)) { \
		//FILE *f = fopen("log", "wb"); \
		//fprintf(f, "assertion failed at %s %d\n", __FILE__, __LINE__); \
		//fclose(f); \
	//} })
#else
#	define wimlib_assert(expr)
#endif


/* Used for buffering FILE IO in some places. */
#define BUFFER_SIZE 4096

void *xmalloc(uint64_t size);
void *xzalloc(u64 size);
void *xrealloc(void *ptr, u64 size);
char *xstrdup(const char *s);


char *utf16_to_utf8(const char *utf16_str, size_t utf16_len,
					size_t *utf8_len_ret);

char *utf8_to_utf16(const char *utf8_str, size_t utf8_len, 
					size_t *utf16_len_ret);

u64 get_timestamp();
time_t ms_timestamp_to_unix(u64 timestamp);

void randomize_byte_array(u8 *p, size_t n);
void randomize_byte_array_with_alnum(u8 *p, size_t n);
bool sha1sum(const char *filename, void *buf);

const char *path_next_part(const char *path, int *first_part_len_ret);
const char *path_basename(const char *path);
void to_parent_name(char buf[], int len);
void print_string(const u8 string[], u64 len);
int get_num_path_components(const char *path);

#define INTERNAL __attribute__((visibility("hidden")))



static inline void print_byte_field(const u8 field[], size_t len)
{
	while (len--)
		printf("%02hhx", *field++);
}


#endif /* _WIMLIB_UTIL_H */
