/*
 * wim.c
 *
 * Miscellaneous routines for wimlib.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib_internal.h"
#include "io.h"
#include <stdarg.h>
#include <errno.h>

/* True if WIMLIB is to print an informational message when an error occurs.
 * This can be turned off by calling wimlib_set_error_messages(false). */
static bool print_errors = true;

/* Print formatted error message to stderr. */
INTERNAL void wimlib_error(const char *format, ...)
{
	if (print_errors) {
		va_list va;
		va_start(va, format);
		fputs("WIMLIB: ", stderr);
		vfprintf(stderr, format, va);
		va_end(va);
	}
}

void wimlib_set_print_errors(bool show_error_messages)
{
	print_errors = show_error_messages;
}




/* 
 * Calls a function on images in the WIM.  If @image is WIM_ALL_IMAGES, @func is
 * called on the WIM once for each image, with each image selected as the
 * current image in turn.  If @image is a certain image, @func is called on the
 * WIM only once, with that image selected.
 */
static bool for_image(WIMStruct *w, int image, bool (*func)(WIMStruct *))
{
	if (image == WIM_ALL_IMAGES) {
		for (int i = 1; i <= w->hdr.image_count; i++) {
			if (!wimlib_select_image(w, i))
				return false;
			if (!(*func)(w))
				return false;
		}
	} else {
		wimlib_assert(image >= 1 && image <= w->hdr.image_count);
		if (!wimlib_select_image(w, image))
			return false;
		if (!(*func)(w))
			return false;
	}
	return true;
}

/* Returns the compression type given in the flags of a WIM header. */
INTERNAL int wim_hdr_flags_compression_type(int wim_hdr_flags)
{
	if (wim_hdr_flags & WIM_HDR_FLAG_COMPRESSION) {
		if (wim_hdr_flags & WIM_HDR_FLAG_COMPRESS_LZX) {
			return WIM_COMPRESSION_TYPE_LZX;
		} else {
			wimlib_assert(wim_hdr_flags & WIM_HDR_FLAG_COMPRESS_XPRESS);
			return WIM_COMPRESSION_TYPE_XPRESS;
		}
	} else {
		return WIM_COMPRESSION_TYPE_NONE;
	}
}

void wimlib_set_verbose(WIMStruct *w, bool verbose)
{
	w->verbose = verbose;
}

/* Creates a WIMStruct for a new WIM file that uses the compression type
 * @ctype. */
WIMStruct *wimlib_create_new_wim(int ctype)
{
	DEBUG("Creating new WIM with %s compression\n", 
			wimlib_get_compression_type_str(ctype));
	/* Allocate the WIMStruct. */
	WIMStruct *w = wimlib_new_struct();

	/* Initialize the header and lookup table. */
	int hdr_flags;
	switch (ctype) {
	case WIM_COMPRESSION_TYPE_NONE:
		hdr_flags = 0;
		break;
	case WIM_COMPRESSION_TYPE_LZX:
		hdr_flags = (WIM_HDR_FLAG_COMPRESSION | WIM_HDR_FLAG_COMPRESS_LZX);
		break;
	case WIM_COMPRESSION_TYPE_XPRESS:
		hdr_flags = (WIM_HDR_FLAG_COMPRESSION | WIM_HDR_FLAG_COMPRESS_XPRESS);
		break;
	default:
		wimlib_assert(0);
	}
	init_header(&w->hdr, hdr_flags);
	lookup_table_init(&w->lookup_table, 9001);
	return w;
}


/*
 * Returns a pointer to a newly allocated WIMStruct.  It must be freed using
 * wimlib_destroy(). 
 */
WIMStruct *wimlib_new_struct()
{
	WIMStruct *w = xzalloc(sizeof(WIMStruct));
	w->link_type      = WIM_LINK_TYPE_HARD;
	w->current_image  = WIM_NO_IMAGE;
	return w;
}


char *wimlib_get_filename(WIMStruct *w)
{
	return w->filename;
}

unsigned wimlib_get_num_images(const WIMStruct *w)
{
	return w->hdr.image_count;
}

unsigned wimlib_get_current_image(const WIMStruct *w)
{
	return w->current_image + 1;
}

bool wimlib_select_image(WIMStruct *w, unsigned image)
{
	DEBUG("Selecting image %u\n", image);

	if (image - 1 == w->current_image)
		return true;

	if (image > w->hdr.image_count || image < 1) {
		wimlib_error("Cannot select image %u: There are only %u images!\n",
				image, w->hdr.image_count);
		return false;
	}


	/* If a valid image is selected, it can be freed if it is not modified.
	 * */
	if (w->current_image != WIM_NO_IMAGE && !wim_current_image_is_modified(w)) {
		DEBUG("Freeing image %u\n", w->current_image);
		WIMBootMetadata *bmd = wim_get_current_boot_metadata(w);
		free_dentry_tree(bmd->root_dentry, NULL, false);
		destroy_security_data(&bmd->security_data);
		bmd->root_dentry = NULL;
	}

	w->current_image = image - 1;

	if (wim_root_dentry(w) != NULL)
		return true;
	else
		return read_boot_metadata(w->fp, 
				wim_metadata_resource_entry(w),
				wimlib_get_compression_type(w), 
				wim_security_data(w), 
				wim_root_dentry_p(w));
}


/* Returns the compression type of the WIM file. */
int wimlib_get_compression_type(const WIMStruct *w)
{
	return wim_hdr_flags_compression_type(w->hdr.flags);
}

const char *wimlib_get_compression_type_str(int ctype)
{
	static const char *compression_types[] = {
		[WIM_COMPRESSION_TYPE_NONE]   = "None",
		[WIM_COMPRESSION_TYPE_LZX]    = "LZX",
		[WIM_COMPRESSION_TYPE_XPRESS] = "XPRESS",
	};
	wimlib_assert(ARRAY_LEN(compression_types) == 3);
	wimlib_assert(ctype >= 0 && ctype <= 2);
	return compression_types[ctype];
}


INTERNAL int wim_resource_compression_type(const WIMStruct *w, 
					const WIMResourceEntry *entry)
{
	int wim_ctype = wimlib_get_compression_type(w);
	return resource_compression_type(wim_ctype, entry->flags);
}


/*
 * Returns the number of an image in the WIM file, given a string that is either
 * the number of the image, or the name of the image.  The images are numbered
 * starting at 1.
 */
int wimlib_resolve_image(WIMStruct *w, const char *image_name_or_num)
{
	if (strcmp(image_name_or_num, "all") == 0)
		return WIM_ALL_IMAGES;
	char *p;
	int image = strtoul(image_name_or_num, &p, 10);
	if (*p == '\0') {
		if (image < 1 || image > w->hdr.image_count)
			return WIM_NO_IMAGE;
		return image;
	} else {
		for (int i = 1; i <= w->hdr.image_count; i++) {
			if (strcmp(image_name_or_num, 
					wimlib_get_image_name(w, i)) == 0)
				return i;
		}
		return WIM_NO_IMAGE;
	}
}


void wimlib_print_wim_information(const WIMStruct *w)
{
	puts("WIM Information:");
	puts("----------------");
	printf("Path:           %s\n", w->filename);
	const WIMHeader *hdr = &w->hdr;
	fputs ("GUID:           0x", stdout);
	print_byte_field(hdr->guid, WIM_GID_LEN);
	putchar('\n');
	printf("Image Count:    %d\n", hdr->image_count);
	printf("Compression:    %s\n", wimlib_get_compression_type_str(
						wimlib_get_compression_type(w)));
	printf("Part Number:    %d/%d\n", hdr->part_number, hdr->total_parts);
	printf("Boot Index:     %d\n", hdr->boot_idx);
	printf("Size:           %"PRIu64" bytes\n", xml_info_get_total_bytes(w->xml_info));
	printf("Integrity Info: %s\n", (w->hdr.integrity.size != 0) ? "yes" : "no");
	putchar('\n');
}


bool wimlib_print_available_images(WIMStruct *w, int image)
{
	if (image == WIM_ALL_IMAGES) {
		puts("Available Images:");
		puts("-----------------");
	} else {
		int n = printf("Information for Image %d\n", image);
		for (int i = 0; i < n - 1; i++)
			putchar('-');
		putchar('\n');

	}
	print_image_info(w->xml_info, image);
	return true;
}

static bool print_metadata(WIMStruct *w)
{
	print_security_data(wim_security_data(w));
	for_dentry_in_tree(wim_root_dentry(w), print_dentry, &w->lookup_table);
	return true;
}


static bool print_files(WIMStruct *w)
{
	return for_dentry_in_tree(wim_root_dentry(w), print_dentry_full_path, 
					NULL);
}


/* Prints the metadata for the specified image, which may be WIM_ALL_IMAGES, but
 * not WIM_NO_IMAGE. */
bool wimlib_print_metadata(WIMStruct *w, int image)
{
	return for_image(w, image, print_metadata);
}

/* 
 * Prints the full paths of all files in the specified image, which may be
 * WIM_ALL_IMAGES, but not WIM_NO_IMAGE. 
 */
bool wimlib_print_files(WIMStruct *w, int image)
{
	return for_image(w, image, print_files);
}

/* Sets the index of the bootable image. */
void wimlib_set_boot_idx(WIMStruct *w, int boot_idx)
{
	wimlib_assert(boot_idx == 0 || 
			(boot_idx >= 1 && boot_idx <= w->hdr.image_count));
	w->hdr.boot_idx = boot_idx;
}

static int sort_boot_metadata_by_position(const void *p1, const void *p2)
{
	WIMBootMetadata *bmd1 = (WIMBootMetadata*)p1;
	WIMBootMetadata *bmd2 = (WIMBootMetadata*)p2;
	u64 offset1 = bmd1->lookup_table_entry->resource_entry.offset;
	u64 offset2 = bmd2->lookup_table_entry->resource_entry.offset;
	if (offset1 < offset2)
		return -1;
	else if (offset1 > offset2)
		return 1;
	else
		return 0;
}

/* 
 * If @entry points to a metadata resource, append it to the list of metadata
 * resources in the WIMStruct.
 */
static bool append_metadata_resource_entry(WIMLookupTableEntry *entry, 
						void *wim_struct_p)
{
	WIMStruct *w = (WIMStruct*)wim_struct_p;
	if (entry->resource_entry.flags & WIM_RESHDR_FLAG_METADATA) {
		if (w->current_image >= w->hdr.image_count) {
			wimlib_error("Expected only %u images, but found more!\n",
					w->hdr.image_count);
			return false;
		} else {
			DEBUG("Found metadata resource for image %u at "
				"offset %"PRIu64"\n", w->current_image + 1, 
				entry->resource_entry.offset);
			w->boot_metadata[w->current_image++].lookup_table_entry = entry;
			return true;
		}
	} else {
		/* Do nothing if not a metadata resource. */
		return true;
	}
}

/* 
 * Begins the reading of a WIM file; opens the file and reads its header.
 *
 * @w:  A pointer to a WIMStruct that keeps track of information about the file.
 * @in_wim_path: The name of the file to open.
 * @check_integrity:	Whether the integrity of the WIM should be checked.  
 * 			If this option is given, the function will fail if there
 * 			is an error reading the integrity data, or if the
 * 			integrity data does not correctly give the integrity of
 * 			the data in the WIM.  If there is not any integrity data
 * 			in the WIM, the function will not fail, but a warning
 * 			will be printed.
 * @return: True if no error occurred; false if error occurred.
 */
bool wimlib_begin_read(WIMStruct *w, const char *in_wim_path, 
							bool check)
{
	DEBUG("Reading the WIM file `%s'\n", in_wim_path);

	w->filename           = xstrdup(in_wim_path);
	w->fp                 = fopen(in_wim_path, "rb");
	if (w->fp == NULL) {
		wimlib_error("Failed to open the file \"%s\" for reading: %s\n",
				in_wim_path, strerror(errno));
		return false;
	}
	if (!read_header(w->fp, &w->hdr))
		return false;

	DEBUG("Wim file contains %u images\n", w->hdr.image_count);

	if (w->hdr.chunk_size != WIM_CHUNK_SIZE) {
		wimlib_error("Unexpected chunk size of %u! Ask the author to "
				"implement support for other chunk sizes. (Or it "
				"might just be that the WIM header is invalid.)\n",
				w->hdr.chunk_size);
		return false;
	}


	switch (wimlib_get_compression_type(w)) {
	case WIM_COMPRESSION_TYPE_XPRESS:
		wimlib_error("Unsupported compression method: XPRESS\n");
		return false;
	case WIM_COMPRESSION_TYPE_INVALID:
		wimlib_error("Unsupported compression method: Unknown\n");
		return false;
	default:
		break;
	}


	if (check) {
		switch (check_integrity(w)) {
		case WIM_INTEGRITY_NONEXISTENT:
			wimlib_error("WARNING: No integrity information- skipping "
					"integrity check.\n");
			break;
		case WIM_INTEGRITY_OK:
			break;
		case WIM_INTEGRITY_NOT_OK:
			wimlib_error("WIM file failed integrity check!\n");
			return false;
		case WIM_INTEGRITY_ERROR:
			return false;
		}
	}

	if (resource_is_compressed(&w->hdr.lookup_table_metadata)) {
		wimlib_error("Didn't expect a compressed lookup table!\n");
		return false;
	}
	if (!read_lookup_table(w->fp, w->hdr.lookup_table_metadata.offset,
			w->hdr.lookup_table_metadata.size, &w->lookup_table))
		return false;

	DEBUG("Allocating %u root dentries and %u security data structures\n",
			w->hdr.image_count, w->hdr.image_count);

	w->boot_metadata = xzalloc(w->hdr.image_count * sizeof(WIMBootMetadata));
	w->current_image = 0;

	DEBUG("Looking for metadata resources in the lookup table.\n");

	/* Find the images in the WIM by searching the lookup table. */
	if (!for_lookup_table_entry(&w->lookup_table, 
				append_metadata_resource_entry, w))
		return false;

	/* Make sure all the expected images were found.  (We already have
	 * returned false if *extra* images were found) */
	if (w->current_image != w->hdr.image_count) {
		wimlib_error("Only found %u images in WIM, but expected %u!\n",
				w->current_image, w->hdr.image_count);
		return false;
	}

	/* Sort images by the position of their metadata resources.  I'm
	 * assuming that is what determines the other of the images in the WIM
	 * file, rather than their order in the lookup table, which may be
	 * random because of hashing. */
	qsort(w->boot_metadata, w->hdr.image_count, sizeof(WIMBootMetadata), 
			sort_boot_metadata_by_position);

	/* Read the XML data. */
	if (!read_xml_data(w->fp, &w->hdr.xml_metadata, &w->xml_data, &w->xml_info)) {
		wimlib_error("Missing or invalid XML data!\n");
		return false;
	}

	/* No image is currently selected; all the root dentries are NULL
	 * pointers. */
	w->current_image = WIM_NO_IMAGE;
	DEBUG("Done beginning read of WIM file.\n");
	return true;
}


/* 
 * Opens a WIM file and creates a WIMStruct for it.
 *
 * Returns a pointer to the WIMStruct, or NULL if there was an error. 
 */
WIMStruct *wimlib_open_wim(const char *wim_file, bool check_integrity)
{
	DEBUG("wim_file = `%s', check_integrity = %d\n",
			wim_file, check_integrity);
	WIMStruct *w = wimlib_new_struct();
	if (!wimlib_begin_read(w, wim_file, check_integrity)) {
		wimlib_error("Could not begin reading the WIM file `%s'!\n", 
								wim_file);
		wimlib_free(w);
		return NULL;
	}
	return w;
}

void wimlib_destroy(WIMStruct *w)
{
	DEBUG2("Closing files.\n");
	if (w->fp != NULL)
		fclose(w->fp);
	if (w->out_fp != NULL)
		fclose(w->out_fp);

	DEBUG2("Destroying lookup table.\n");
	lookup_table_destroy(&w->lookup_table);

	DEBUG2("Freeing filename.\n");
	free(w->filename);

	DEBUG2("Freeing XML data and information structure.\n");
	free(w->xml_data);
	free_xml_info(w->xml_info);

	if (w->boot_metadata != NULL) {
		DEBUG2("Freeing boot metadata.\n");
		for (int i = 0; i < w->hdr.image_count; i++) {
			DEBUG2("Destroying security data for image %d\n", i + 1);
			destroy_security_data(&w->boot_metadata[i].security_data);
			DEBUG2("Freeing dentry tree for image %d\n", i + 1);
			free_dentry_tree(w->boot_metadata[i].root_dentry, 
								NULL, false);
		}
	}
	memset(w, 0, sizeof(*w));
}

/* Frees the memory for the WIMStruct, including all internal memory; also
 * closes all files associated with the WIMStruct.  */
void wimlib_free(WIMStruct *w)
{
	if (w != NULL) {
		wimlib_destroy(w);
		free(w);
	}
}
