/*
 * wimlib.h
 *
 * External header for wimlib.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#ifndef _WIMLIB_H
#define _WIMLIB_H

#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>

#ifndef _WIMLIB_INTERNAL_H
typedef struct WIMStruct WIMStruct;
#endif

WIMStruct *wimlib_new_struct();
WIMStruct *wimlib_create_new_wim(int ctype);
bool wimlib_begin_read(WIMStruct *w, const char *filename, bool check);
WIMStruct *wimlib_open_wim(const char *wim_file, bool check);

bool wimlib_select_image(WIMStruct *w, unsigned image);
int wimlib_resolve_image(WIMStruct *w, const char *image_name_or_num);

void wimlib_print_header(const WIMStruct *w);
void wimlib_print_wim_information(const WIMStruct *w);
void wimlib_print_lookup_table(WIMStruct *w);
bool wimlib_print_metadata(WIMStruct *w, int image);
bool wimlib_print_xml_data(WIMStruct *w, int image);
bool wimlib_print_files(WIMStruct *w, int image);
bool wimlib_print_available_images(WIMStruct *w, int image);


const char *wimlib_get_compression_type_str(int ctype);
const char *wimlib_get_image_name(WIMStruct *w, int image);
unsigned wimlib_get_num_images(const WIMStruct *w);
unsigned wimlib_get_current_image(const WIMStruct *w);
char *wimlib_get_filename(WIMStruct *w);
int wimlib_get_compression_type(const WIMStruct *w);

bool wimlib_image_name_in_use(WIMStruct *w, const char *name);

void wimlib_set_image_name(WIMStruct *w, int image, const char *name);
void wimlib_set_image_descripton(WIMStruct *w, int image, 
					const char *description);
void wimlib_set_boot_idx(WIMStruct *w, int boot_idx);
void wimlib_set_error_messages(bool show_messages);
void wimlib_set_verbose(WIMStruct *w, bool verbose);
void wimlib_set_link_type(WIMStruct *w, int link_type);
bool wimlib_set_output_dir(WIMStruct *w, const char *dir);

bool wimlib_extract_xml_data(WIMStruct *w, const char *out_file);
bool wimlib_extract_image(WIMStruct *w, int image);


bool wimlib_add_image(WIMStruct *w, const char *path, const char *name, 
			const char *desc, const char *flags, bool boot);

bool wimlib_export_image(WIMStruct *src_w, int src_image, WIMStruct *dest_w,
			const char *dest_wimfile, const char *dest_name, 
							bool boot);
bool wimlib_delete_image(WIMStruct *w, int image);


bool wimlib_write(WIMStruct *w, const char *out_wim_path, int image, 
				bool check_integrity);
bool wimlib_overwrite(WIMStruct *w, bool check_integrity);
bool wimlib_overwrite_xml_and_header(WIMStruct *w, bool check);


void wimlib_destroy(WIMStruct *w);
void wimlib_free(WIMStruct *w);


/* FUSE */
bool wimlib_mount(WIMStruct *wim_struct_p, int image, const char *dir, bool ro,
						bool debug, bool check);
bool wimlib_unmount(const char *dir, bool commit);


#define WIM_LINK_TYPE_HARD     0
#define WIM_LINK_TYPE_SYMBOLIC 1
#define WIM_LINK_TYPE_NONE     2

#define WIM_COMPRESSION_TYPE_INVALID   -1
#define WIM_COMPRESSION_TYPE_NONE	0
#define WIM_COMPRESSION_TYPE_LZX	1
#define WIM_COMPRESSION_TYPE_XPRESS	2


/* Used to indicate that no WIM image is currently selected. */
#define WIM_NO_IMAGE   0xfffffffe
/* Used to specify all images in the WIM. */
#define WIM_ALL_IMAGES 0xfffffffd

enum wim_integrity_status {
	WIM_INTEGRITY_NONEXISTENT = 0,
	WIM_INTEGRITY_OK,
	WIM_INTEGRITY_NOT_OK,
	WIM_INTEGRITY_ERROR,
};


#endif /* _WIMLIB_H */
