/*
 * wimlib_internal.h
 *
 * Internal header for wimlib.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#ifndef _WIMLIB_INTERNAL_H
#define _WIMLIB_INTERNAL_H

#include "util.h"

struct stat;

/* Lengths of array fields of constant length. */
#define WIM_HASH_SIZE  20
#define WIM_MAGIC_LEN  8
#define WIM_GID_LEN    16
#define WIM_UNUSED_LEN 60


/* Length of the WIM header on disk. */
#define WIM_HEADER_DISK_SIZE (148 + WIM_UNUSED_LEN)

/* Size of each lookup table entry in the WIM file. */
#define WIM_LOOKUP_TABLE_ENTRY_DISK_SIZE 50

/* Size of the WIMStreamEntry excluding the stream name. */
#define WIM_STREAM_ENTRY_DISK_SIZE (2 * sizeof(u64) + WIM_HASH_SIZE + sizeof(u16))

/* Size of the WIMDentry up to and including the file_name_len. */
#define WIM_DENTRY_DISK_SIZE 102

/* Compressed resources in the WIM are divided into separated compressed chunks
 * of this size. */

#define WIM_CHUNK_SIZE 32768

/* Version of the WIM file.  I don't know if there has ever been a different
 * version or not. */
#define WIM_VERSION 0x10d00



/* Metadata for a resource in a WIM file. */
typedef struct WIMResourceEntry {
	/* Size, in bytes, of the resource in the WIM file. */
	u64 size  : 56;

	/* Bitwise or of one or more of the WIM_RESHDR_FLAG_* flags. */
	u64 flags : 8;

	/* Offset, in bytes, of the resource in the WIM file. */
	u64 offset;

	/* Uncompressed size of the resource in the WIM file.  Is the same as
	 * @size if the resource is uncompressed. */
	u64 original_size;
} WIMResourceEntry;

/* Flags for the `flags' field of the WIMResourceEntry structure. */

/* ??? */
#define WIM_RESHDR_FLAG_FREE            0x01

/* Indicates that a file resource is a metadata resource. */
#define WIM_RESHDR_FLAG_METADATA        0x02

/* Indicates that a file resource is compressed. */
#define WIM_RESHDR_FLAG_COMPRESSED	0x04

/* ??? */
#define WIM_RESHDR_FLAG_SPANNED         0x08


/* Header at the very beginning of the WIM file. */
typedef struct WIMHeader
{ 
	/* Identifies the file as WIM file. Must be exactly
	 * {'M', 'S', 'W', 'I', 'M', 0, 0, 0}  */
	//u8  magic[WIM_MAGIC_LEN]; 	

	/* size of WIM header in bytes. */
	u32 hdr_size;

	/* Version of the WIM file.  M$ provides no documentation about exactly
	 * what this field affects about the file format, other than the fact
	 * that more recent versions have a higher value. */
	u32 version;

	/* Bitwise OR of one or more of the WIM_HDR_FLAG_* defined below. */
	u32 flags;

	/* The size of the pieces that the uncompressed files were split up into
	 * when they were compressed.  This should be the same as
	 * WIM_CHUNK_SIZE.  M$ incorrectly documents this as "the size of the
	 * compressed .wim file in bytes".*/
	u32 chunk_size;
	
	/* A unique identifier for the WIM file. */
	u8  guid[WIM_GID_LEN];

	/* Part number of the WIM file in a spanned set. */
	u16 part_number;

	/* Total number of parts in a spanned set. */
	u16 total_parts;

	/* Number of images in the WIM file. */
	u32 image_count;

	/* Location, size, and flags of the lookup table. */
	WIMResourceEntry lookup_table_metadata;

	/* Location, size, and flags for the XML data. */
	WIMResourceEntry xml_metadata;

	/* Location, size, and flags for the boot metadata. */
	WIMResourceEntry boot_metametadata;

	/* The index of the bootable image in the WIM file. If 0, there are no
	 * bootable images available. */
	u32 boot_idx; 

	/* The location of the integrity table used to verify files. */
	WIMResourceEntry integrity;

	/* Reserved for future disuse */
	u8 unused[WIM_UNUSED_LEN];
} WIMHeader;

/* Flags for the `flags' field of the WIMHeader. */


/* Reserved for future use by M$ */
#define WIM_HDR_FLAG_RESERVED           0x00000001

/* Files and metadata in the WIM are compressed. */
#define WIM_HDR_FLAG_COMPRESSION        0x00000002

/* WIM is read-only. */
#define WIM_HDR_FLAG_READONLY           0x00000004

/* Resource data specified by images in this WIM may be contained in a different
 * WIM */
#define WIM_HDR_FLAG_SPANNED            0x00000008

/* The WIM contains resources only; no filesystem metadata. */
#define WIM_HDR_FLAG_RESOURCE_ONLY      0x00000010

/* The WIM contains metadata only. */
#define WIM_HDR_FLAG_METADATA_ONLY      0x00000020

/* Lock field to prevent multiple writers from writing the WIM concurrently. */
#define WIM_HDR_FLAG_WRITE_IN_PROGRESS  0x00000040 

/* Reparse point fixup ??? */
#define WIM_HDR_FLAG_RP_FIX             0x00000080

/* Unknown compression type */
#define WIM_HDR_FLAG_COMPRESS_RESERVED  0x00010000

/* Resources within the WIM are compressed using "XPRESS" compression, which is
 * a LZ77-based compression algorithm. */
#define WIM_HDR_FLAG_COMPRESS_XPRESS    0x00020000

/* Resources within the WIM are compressed using "LZX" compression.  This is also
 * a LZ77-based algorithm. */
#define WIM_HDR_FLAG_COMPRESS_LZX       0x00040000


/* A lookup table that is used to translate the hash codes of dentries into the
 * offsets and sizes of uncompressed or compressed file resources.  It is
 * implemented as a hash table. */
typedef struct WIMLookupTable {
	struct WIMLookupTableEntry **array;
	u64 num_entries;
	u64 capacity;
} WIMLookupTable;

/* An entry in the lookup table in the WIM file. */
typedef struct WIMLookupTableEntry {

	/* The next WIMLookupTableEntry in the hash bucket.  NULL if this is the
	 * last one. */
	struct WIMLookupTableEntry *next;

	/* @resource_entry is read from the lookup table in the WIM
	 * file; it says where to find the file resource in the WIM
	 * file, and whether it is compressed or not. */
	WIMResourceEntry resource_entry;

	/* Currently ignored; set to 1 in new lookup table entries. */
	u16 part_number;

	/* Number of times this lookup table entry is referenced by dentries. */
	u32 refcnt;

	/* SHA1 hash of the file resource pointed to by this lookup table entry */
	u8  hash[WIM_HASH_SIZE];

	/* If @file_on_disk != NULL, the file resource indicated by this lookup
	 * table entry is not in the WIM file, but rather a file on disk; this
	 * occurs for files that are added to the WIM.  In that case,
	 * file_on_disk is the name of the file in the outside filesystem.  
	 * It will not be compressed, and its size will be given by
	 * resource_entry.size and resource_entry.original_size. */
	union {
		char *file_on_disk;
		char *staging_file_name;
	};

	union {
		struct { /* Used for wimlib_export. */

			/* If (other_wim_fp != NULL), the file resource indicated
			 * by this lookup table entry is in a different WIM
			 * file, and other_wim_fp is the FILE* for it. */
			FILE *other_wim_fp;

			/* Compression type used in other WIM. */
			int   other_wim_ctype;
		};

		struct { /* Used for read-write mounts. */


			/* Offset of the stream file_on_disk_fd. */
			off_t staging_offset;


			/* If file_on_disk_fd, if it is not -1, is the file
			 * descriptor, opened for reading, for file_on_disk. */
			int staging_fd;

			/* Number of times the file has been opened.
			 * file_on_disk_fd can be closed when num_times_opened
			 * is decremented to 0.  */
			int staging_num_times_opened;
		};
	};

	/* When a WIM file is written, out_refcnt starts at 0 and is incremented
	 * whenever the file resource pointed to by this lookup table entry
	 * needs to be written.  Naturally, the file resource only need to be
	 * written when out_refcnt is 0.  Incrementing it further is needed to
	 * find the correct reference count to write to the lookup table in the
	 * output file, which may be less than the regular refcnt if not all
	 * images in the WIM file are written. 
	 *
	 * output_resource_entry is the WIMResourceEntry for the position of the
	 * file resource when written to the output file. */
	u32 out_refcnt;
	WIMResourceEntry output_resource_entry;
} WIMLookupTableEntry;

/* Structure for security data.  Each image in the WIM file has its own security
 * data. */
typedef struct WIMSecurityData {
	/* The total length of the security data, in bytes.  A typical size is
	 * 2048 bytes. */
	u32 total_length;

	/* The number of security descriptors in the array descriptors, below. */
	u32 num_entries;

	/* Array of sizes of the descriptors in the array descriptors. */
	u64 *sizes;

	/* Array of descriptors. */
	u8 **descriptors;
} WIMSecurityData;


/* In-memory structure for a directory entry.  Keep all the fields up until and
 * including file_name_len compatible with the on-disk structure; extra fields
 * are following those. */
typedef struct WIMDentry {
	/* Size of directory entry, in bytes.  Typical size is around 104 to 120
	 * bytes. */
	u64 length;

	/* It is possible for the length field to be 0, and in that case the
	 * on-disk directory entry will end here, meaning the actual length is
	 * 8.  This situation, which is undocumented, indicates the end of a
	 * list of sibling nodes in a directory.  */

	/* The file attributes associated with this file. */
	u32 attributes;

	/* The index of the node in the security table that contains this file's
	 * security information.  If -1, no security information exists for this
	 * file. */
	int32_t security_id;

	/* The offset, from the start of the metadata section, of this directory
	 * entry's child files.  0 if the directory entry has no children. */
	u64 subdir_offset;

	/* Reserved */
	u64 unused1;
	u64 unused2;

	/* Timestamps for the entry.  The timestamps are the number of
	 * 100-nanosecond intervals that have elapsed since 12:00 A.M., January
	 * 1st, 1601, UTC. */
	u64 creation_time;
	u64 last_access_time;
	u64 last_write_time;

	/* A hash of the file's contents. */
	u8 hash[WIM_HASH_SIZE];

	/* Identity of a reparse point. */
	u32 reparse_tag;

	/* Although M$'s documentation does not tell you this, it seems that the
	 * reparse_reserved field does not actually exist.  So the hard_link
	 * field directly follows the reparse_tag on disk. */
	//u32 reparse_reserved;

	/* If the reparse_reserved field existed, there would be a 4-byte gap
	 * here to align hard_link on an 8-byte field.  However,
	 * reparse_reserved does not actually exist, so there is no gap here. */

	/* If the file is part of a hard link set, all the directory entries in
	 * the set will share the same value for this field. */
	u64 hard_link;

	/* Number of WIMStreamEntry structures that follow this WIMDentry. */
	u16 streams;

	/* Length of short filename, in bytes. */
	u16 short_name_len;

	/* Length of file name, in bytes. */
	u16 file_name_len;

	/* Currently at offset 102.  This is the end of the on-disk structure.
	 * On disk, this is followed by the filename, not the remaining members
	 * of this structure.  After the filename, which is file_name_len bytes,
	 * there is an undocumented 2-byte gap, then the short name, which is
	 * short_name_len bytes. */

	/* Length of the filename converted into UTF-8, in bytes. */
	u16 file_name_utf8_len;

	/* Pointer to the short filename */
	char *short_name;

	/* Pointer to the filename. */
	char *file_name;

	/* Pointer to the filename converted to UTF-8. */
	char *file_name_utf8;

	/* Full path to this dentry. */
	char *full_path_utf8;
	u32   full_path_utf8_len;

	/* Stream entries for this dentry. */
	struct WIMStreamEntry *stream_entries;

	/* Number of references to the dentry tree itself, as in multiple
	 * WIMStructs */
	int refcnt;
	
	/* Whether this is a directory or not. */
	//bool is_directory;

	/* The parent of this directory entry. */
	struct WIMDentry *parent;

	/* Linked list of sibling directory entries. */
	struct WIMDentry *next;
	struct WIMDentry *prev;

	/* Pointer to a child of this directory entry. */
	struct WIMDentry *children;
} WIMDentry;

#define WIM_FILE_ATTRIBUTE_READONLY            0x1
#define WIM_FILE_ATTRIBUTE_HIDDEN              0x2
#define WIM_FILE_ATTRIBUTE_SYSTEM              0x4
#define WIM_FILE_ATTRIBUTE_DIRECTORY           0x10
#define WIM_FILE_ATTRIBUTE_ARCHIVE             0x20
#define WIM_FILE_ATTRIBUTE_DEVICE              0x40
#define WIM_FILE_ATTRIBUTE_NORMAL              0x80
#define WIM_FILE_ATTRIBUTE_TEMPORARY           0x100
#define WIM_FILE_ATTRIBUTE_SPARSE_FILE         0x200
#define WIM_FILE_ATTRIBUTE_REPARSE_POINT       0x400
#define WIM_FILE_ATTRIBUTE_COMPRESSED          0x800
#define WIM_FILE_ATTRIBUTE_OFFLINE             0x1000
#define WIM_FILE_ATTRIBUTE_NOT_CONTENT_INDEXED 0x2000
#define WIM_FILE_ATTRIBUTE_ENCRYPTED           0x4000
#define WIM_FILE_ATTRIBUTE_VIRTUAL             0x10000

/* Undocumented padding between the file name and the short name that are part
 * of a directory entry in the WIM file.  */
#define WIM_FILENAME_PADDING 2

/* A directory entry may be followed by a number of alternate stream entries.
 * This feature seems to be rarely used and is currently unsupported, although
 * the stream entries are read into memory. */
typedef struct WIMStreamEntry {

	/* Length of this stream entry. */
	 u64	  length;

	 /* Reserved. */
	 u64	  unused1;

	 /* Hash code to look up this entry in the lookup table. */
	 u8       hash[WIM_HASH_SIZE];

	 /* Length of the stream name. */
	 u16	  stream_name_len;

	 /* Name of the stream. */
	 char	 *stream_name;
} WIMStreamEntry;


/* Information that is part of a metadata resource. */
typedef struct {

	/* The security data for the image. */
	WIMSecurityData      security_data;

	/* The root dentry for the image. */
	WIMDentry           *root_dentry;

	/* True if the filesystem of the image has been modified.  If this is
	 * the case, the memory for the filesystem is not freed when switching
	 * to a different WIM image. */
	bool		     modified;

	/* A pointer to the lookup table entry for the metadata resource. */
	WIMLookupTableEntry *lookup_table_entry;

} WIMBootMetadata;

/* The opaque structure exposed to the wimlib API. */
typedef struct {
	/* The name of the WIM file that has been opened. */
	char                *filename;

	/* A pointer to the file indicated by @filename, opened for reading. */
	FILE                *fp;

	/* The header of the WIM file. */
	WIMHeader            hdr;

	/* The currently selected image. Indexed starting at 0, since it is an
	 * index into the boot_metadata array; this is unlike how the images are
	 * indexed starting at 1 in the public API and the imagex program. */
	int                  current_image;

	/* The lookup table for the WIM file. */ 
	WIMLookupTable       lookup_table;

	/* Pointer to the XML data read from the WIM file. */
	u8                  *xml_data;

	/* Information retrieved from the XML data, arranged
	 * in an orderly manner. */
	struct WIMInfo      *xml_info;

	/* Array of the boot metadata of length image_count.  Each image in the
	 * WIM has a boot metadata associated with it. */
	WIMBootMetadata     *boot_metadata;

	/* True if files names are to be printed when doing extraction. 
	 * May be used for other things later. */
	bool                 verbose;

	/* The type of links to create when extracting files (hard, symbolic, or
	 * none.) */
	int                  link_type;

	/* Name of the output directory for extraction. */
	const char          *output_dir;

	/* Set to true when extracting multiple images */
	bool                 is_multi_image_extraction;

	/* FILE pointer for the WIM file that is being written. */
	FILE                *out_fp;

} WIMStruct;


/* Inline utility functions for WIMStructs. */
static inline WIMDentry *wim_root_dentry(WIMStruct *w)
{
	return w->boot_metadata[w->current_image].root_dentry;
}

static inline WIMDentry **wim_root_dentry_p(WIMStruct *w)
{
	return &w->boot_metadata[w->current_image].root_dentry;
}

static inline WIMSecurityData *wim_security_data(WIMStruct *w)
{
	return &w->boot_metadata[w->current_image].security_data;
}

static inline WIMLookupTableEntry *wim_metadata_lookup_table_entry(WIMStruct *w)
{
	return w->boot_metadata[w->current_image].lookup_table_entry;
}

static inline WIMResourceEntry *wim_metadata_resource_entry(WIMStruct *w)
{
	return &w->boot_metadata[w->current_image].lookup_table_entry->resource_entry;
}

static inline WIMResourceEntry *wim_metadata_output_resource_entry(WIMStruct *w)
{
	return &w->boot_metadata[w->current_image].lookup_table_entry->output_resource_entry;
}


/* Return true if a WIMResourceEntry indicates a compressed resource. */
static inline bool resource_is_compressed(const WIMResourceEntry *entry)
{
	wimlib_assert(entry->original_size == entry->size || 
				entry->flags & WIM_RESHDR_FLAG_COMPRESSED);

	return (entry->flags & WIM_RESHDR_FLAG_COMPRESSED) != 0;
}


static inline WIMBootMetadata *wim_get_current_boot_metadata(WIMStruct *w)
{
	return &w->boot_metadata[w->current_image];
}

static inline bool wim_current_image_is_modified(const WIMStruct *w)
{
	return w->boot_metadata[w->current_image].modified;
}

/* Inline utility functions for WIMDentries */
static inline bool dentry_is_root(const WIMDentry *dentry)
{
	return dentry->parent == dentry;
}

static inline bool dentry_is_first_sibling(const WIMDentry *dentry)
{
	return dentry_is_root(dentry) || dentry->parent->children == dentry;
}

static inline bool dentry_is_only_child(const WIMDentry *dentry)
{
	return dentry->next == dentry;
}

static inline bool dentry_is_directory(const WIMDentry *dentry)
{
	return (dentry->attributes & WIM_FILE_ATTRIBUTE_DIRECTORY) != 0;
}

static inline bool dentry_is_regular_file(const WIMDentry *dentry)
{
	//return (dentry->attributes & WIM_FILE_ATTRIBUTE_NORMAL) != 0;
	return !dentry_is_directory(dentry);
}

static inline void dentry_set_is_directory(WIMDentry *dentry)
{
	dentry->attributes |= WIM_FILE_ATTRIBUTE_DIRECTORY;
}

static inline void dentry_set_is_regular_file(WIMDentry *dentry)
{
	dentry->attributes |= WIM_FILE_ATTRIBUTE_NORMAL;
}

static inline bool dentry_is_empty_directory(const WIMDentry *dentry)
{
	return dentry_is_directory(dentry) && dentry->children == NULL;
}

/* Prints a hash code field. */
static inline void print_hash(const u8 hash[])
{
	print_byte_field(hash, WIM_HASH_SIZE);
}

/* dentry.c */
void stbuf_to_dentry(struct stat *stbuf, WIMDentry *dentry);
void dentry_to_stbuf(WIMDentry *dentry, struct stat *stbuf, 
					WIMLookupTable *table);
bool for_dentry_in_tree(WIMDentry *root, 
		bool (*visitor)(WIMDentry*, void*), void *args);
bool for_dentry_in_tree_depth(WIMDentry *root, 
		bool (*visitor)(WIMDentry*, void*), void *args);

void calculate_dentry_full_path(WIMDentry *dentry);
bool do_calculate_dentry_full_path(WIMDentry *dentry, void *ignore);
void calculate_subdir_offsets(WIMDentry *dentry, off_t *subdir_offset_p);
void change_dentry_name(WIMDentry *dentry, const char *new_name);
bool dentry_has_name(const WIMDentry *dentry, 
			const char *name, int name_len);

void unlink_dentry(WIMDentry *dentry);
void link_dentry(WIMDentry *dentry, WIMDentry *parent);

bool print_dentry(WIMDentry *dentry, void *lookup_table);
bool print_dentry_full_path(WIMDentry *entry, void *ignore);

WIMDentry *get_dentry(WIMStruct *w, const char *path);
WIMDentry *get_parent_dentry(WIMStruct *w, const char *path);
WIMDentry *get_dentry_child_with_name(const WIMDentry *dentry, 
							const char *name);
void dentry_update_all_timestamps(WIMDentry *dentry);
void init_dentry(WIMDentry *dentry, const char *name);
WIMDentry *create_dentry(const char *name);

void free_dentry(WIMDentry *dentry);
void free_dentry_tree(WIMDentry *root, WIMLookupTable *lookup_table, 
					bool decrement_refcnt);
void increment_dentry_tree_refcnt(WIMDentry *tree);
void calculate_dir_tree_statistics(WIMDentry *root, WIMLookupTable *table, 
			u64 *dir_count_ret, u64 *file_count_ret, 
			u64 *total_bytes_ret, u64 *hard_link_bytes_ret);
bool read_dentry(const u8 metadata_resource[], u64 metadata_resource_len, 
					u64 offset, WIMDentry *dentry);
bool read_dentry_tree(const u8 metadata_resource[], 
		u64 metadata_resource_len, WIMDentry *dentry);
bool write_dentry_tree(WIMDentry *tree, FILE *out);

/* header.c */
bool read_header(FILE *fp, WIMHeader *hdr);
bool write_header(const WIMHeader *hdr, FILE *out);
void init_header(WIMHeader *hdr, int hdr_flags);

/* integrity.c */
int check_integrity(WIMStruct *w);
bool write_integrity(FILE *out, u64 end_header_offset, 
				u64 end_lookup_table_offset);

/* lookup_table.c */
void lookup_table_init(WIMLookupTable *table, size_t capacity);
void lookup_table_insert(WIMLookupTable *lookup_table, 
					WIMLookupTableEntry *entry);
void lookup_table_remove(WIMLookupTable *table, WIMLookupTableEntry *entry);
void lookup_table_unlink(WIMLookupTable *table, 
						WIMLookupTableEntry *entry);
void lookup_table_zero_out_refcnts(WIMLookupTable *table);
WIMLookupTableEntry *new_lookup_table_entry();
bool for_lookup_table_entry(WIMLookupTable *table, 
		bool (*visitor)(WIMLookupTableEntry *, void *), void *arg);
WIMLookupTableEntry *lookup_file_resource(const WIMDentry *dentry, 
						const WIMLookupTable *lookup_table);
bool print_lookup_table_entry(WIMLookupTableEntry *entry, void *ignore);
void lookup_table_decrement_refcnt(WIMLookupTable* table, 
					const WIMDentry *dentry);
bool read_lookup_table(FILE *fp, u64 offset, u64 size, 
		WIMLookupTable *lookup_table_ret);
bool write_lookup_table(WIMLookupTable *table, FILE *out);
void free_lookup_table_entry(WIMLookupTableEntry *entry);
void lookup_table_destroy(WIMLookupTable *table);

/* resource.c */
const u8 *get_resource_entry(const u8 *p, WIMResourceEntry *entry);
u8 *put_resource_entry(u8 *p, const WIMResourceEntry *entry);
bool read_resource(FILE *fp, u64 resource_size, 
				u64 resource_original_size,
				u64 resource_offset, int resource_ctype, 
				u64 len, u64 offset, void *contents_ret);
bool read_full_resource(FILE *fp, u64 resource_size, 
				u64 resource_original_size,
				u64 resource_offset, 
				int resource_ctype, void *contents_ret);
bool extract_resource_to_fd(WIMStruct *w, 
				const WIMResourceEntry *entry, int fd, 
								u64 size);
bool extract_full_resource_to_fd(WIMStruct *w, const WIMResourceEntry *entry, 
								int fd);
void *get_wim_resource(WIMStruct *w, const WIMResourceEntry *entry);
bool read_boot_metadata(FILE *fp, const WIMResourceEntry *metadata,
		int wim_ctype, WIMSecurityData *sd, WIMDentry **root_dentry_p);
int resource_compression_type(int wim_ctype, int reshdr_flags);

/* security.c */
bool read_security_data(const u8 metadata_resource[], 
		u64 metadata_resource_len, WIMSecurityData *sd);

void print_security_data(const WIMSecurityData *sd);
bool write_security_data(const WIMSecurityData *sd, FILE *out);
void init_security_data(WIMSecurityData *sd);
void destroy_security_data(WIMSecurityData *sd);

/* wim.c */
void wimlib_error(const char *format, ...);
int wim_hdr_flags_compression_type(int wim_hdr_flags);
int wim_resource_compression_type(const WIMStruct *w, 
					const WIMResourceEntry *entry);

/* xml.c */
void xml_export_image(const struct WIMInfo *old_wim_info, int image, 
					struct WIMInfo **new_wim_info_p,
				const char *dest_image_name,
				const char *dest_image_description);

void xml_delete_image(struct WIMInfo **wim_info_p, int image);
void xml_add_image(WIMStruct *w, WIMDentry *root_dentry, const char *name,
			const char *description, const char *flags);
u64 xml_info_get_total_bytes(const struct WIMInfo *wim_info);
void print_image_info(const struct WIMInfo *wim_info, int image);
bool read_xml_data(FILE *fp, WIMResourceEntry *res, 
				u8 **xml_data_ret, struct WIMInfo **info_ret);
bool write_xml_data(const struct WIMInfo *wim_info, int image, FILE *out);
void free_xml_info(struct WIMInfo *info);

#include "wimlib.h"



#endif /* _WIMLIB_INTERNAL_H */

