/*
 * write.c
 *
 * Support for writing WIM files.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2010 Carl Thijssen
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */
#include "wimlib_internal.h"
#include "io.h"
#include <unistd.h>


/*
 * Copy @len bytes from @in to @out, at the current position in @out, and at an
 * offset of @in_offset in @in.   Returns true on success, false on failure.
 */
static bool copy_between_files(FILE *in, off_t in_offset, FILE *out, size_t len)
{
	if (fseeko(in, in_offset, SEEK_SET) != 0) {
		wimlib_error("Failed to seek to byte %"PRIu64" of input file: %m\n",
				in_offset);
		return false;
	}

	u8 buf[BUFFER_SIZE];
	while (len != 0) {
		u64 n = min(len, BUFFER_SIZE);
		if (fread(buf, 1, n, in) != n)
			goto copy_between_files_error;

		if (fwrite(buf, 1, n, out) != n)
			goto copy_between_files_error;
		len -= n;
	}

	return true;
copy_between_files_error:
	if (feof(in))
		wimlib_error("Unexpected EOF!\n");
	else
		wimlib_error("Error copying data between files: %m\n");
	return false;
}

/* 
 * Transfers a file resource from a FILE* opened for reading to a FILE* opened
 * for writing, possibly changing the compression type. 
 *
 * @in:			The FILE* that contains the file resource.
 * @size:		The (compressed) size of the file resource.
 * @original_size:	The uncompressed size of the file resource.
 * @offset:		The offset of the file resource in the input file.
 * @input_ctype:	The compression type of the file resource in the input
 * 				file.
 * @out:		The FILE* for the output file.  The file resource is 
 * 				written at the current position of @out.
 * @output_ctype:	The compression type to which the file resource will be
 * 				converted.
 * @output_res_entry:	A pointer to a resource entry that, upon successful
 * 				return of this function,  will have the size,
 * 				original size, offset, and flags fields filled
 * 				in for the file resource written to the output
 * 				file.
 * @return:	True on success, false on failure.
 */
static bool transfer_file_resource(FILE *in, u64 size, u64 original_size, 
			off_t offset, int input_ctype, FILE *out, 
			int output_ctype, WIMResourceEntry *output_res_entry)
{
	/*DEBUG("Transfering file resource between files; input compression = %s, "*/
			/*"output compression = %s\n", */
			/*wimlib_get_compression_type_str(input_ctype), */
			/*wimlib_get_compression_type_str(output_ctype));*/

	/* Get current offset in the output file. */
	output_res_entry->offset = ftello(out);
	if (output_res_entry->offset == -1) {
		wimlib_error("Failed to get output position: %m\n");
		return false;
	}

	/* Check for unsupported case */
	if (output_ctype != WIM_COMPRESSION_TYPE_NONE && 
			input_ctype != output_ctype) 
	{
		wimlib_error("Warning: Currently, compressing the output "
			"(if it isn't already compressed) is unsupported.\n");
		wimlib_error("Defaulting to uncompressed output.\n");
		output_ctype = WIM_COMPRESSION_TYPE_NONE;
	}

	if (output_ctype == input_ctype) {
		/* The same compression types. */

		if (!copy_between_files(in, offset, out, size))
			return false;
		output_res_entry->size = size;
		output_res_entry->original_size = original_size;
	} else {
		/* Different compression types. */

		/* We currently require the output ctype to be uncompressed. */
		wimlib_assert(output_ctype == WIM_COMPRESSION_TYPE_NONE);

		u8 *file_contents = xmalloc(original_size);
		if (!read_full_resource(in, size, original_size, 
					offset, input_ctype, file_contents))
			return false;
		size_t bytes_written = fwrite(file_contents, 1, original_size, out);
		free(file_contents);
		if (bytes_written != original_size) {
			wimlib_error("Failed to write file resource to "
					"output file: %m\n");
			return false;
		}

		output_res_entry->size = original_size;
		output_res_entry->original_size = original_size;
	}

	if (output_ctype == WIM_COMPRESSION_TYPE_NONE)
		output_res_entry->flags = 0;
	else
		output_res_entry->flags = WIM_RESHDR_FLAG_COMPRESSED;
	return true;
}

/* 
 * Writes a file resource to the output file. 
 *
 * @dentry:  The dentry for the file resource.
 * @wim_struct_p:  A pointer to the WIMStruct.  The fields of interest to this
 * 	function are the input and output file streams and the lookup table. 
 * @return true on success, false on failure. 
 */
static bool write_file_resource(WIMDentry *dentry, void *wim_struct_p)
{
	WIMStruct *w = (WIMStruct*)wim_struct_p;
	WIMLookupTable *lookup_table = &w->lookup_table;
	FILE *out = w->out_fp;
	FILE *in;

	/* Directories don't need file resources. */
	if (dentry_is_directory(dentry))
		return true;

	/* Get the lookup entry for the file resource. */
	WIMLookupTableEntry *lookup_table_entry = lookup_file_resource(dentry, 
								lookup_table);
	if (lookup_table_entry != NULL) {

		lookup_table_entry->out_refcnt++;

		/* No need to write file resources twice.  (This indicates file
		 * resources that are part of a hard link set.) */
		if (lookup_table_entry->out_refcnt != 1)
			return true;

		int in_wim_ctype, out_wim_ctype;

		out_wim_ctype = wimlib_get_compression_type(w);
		WIMResourceEntry *output_res_entry = 
			&lookup_table_entry->output_resource_entry;

		/* Figure out if we can read the resource from the WIM file, or
		 * if we have to read it from the filesystem outside. */
		if (lookup_table_entry->file_on_disk != NULL) {

			/* Read from disk (uncompressed) */

			u64 len = lookup_table_entry->resource_entry.original_size;

			in = fopen(lookup_table_entry->file_on_disk, "rb");
			if (in == NULL) {
				wimlib_error("Failed to open the file `%s': %m\n",
						lookup_table_entry->file_on_disk);
				return false;
			}

			bool ret = transfer_file_resource(in, len, len, 0,
					WIM_COMPRESSION_TYPE_NONE, out, 
					out_wim_ctype, output_res_entry);
			fclose(in);

			return ret;
		} else {

			/* Read from input WIM (possibly compressed) */

			/* It may be a different WIM file, in the case of
			 * exporting images from one WIM file to another */
			if (lookup_table_entry->other_wim_fp != NULL) {
				/* Different WIM file. */
				in = lookup_table_entry->other_wim_fp;
				in_wim_ctype = lookup_table_entry->other_wim_ctype;
			} else {
				/* Same WIM file. */
				in = w->fp;
				in_wim_ctype = out_wim_ctype;
			}

			WIMResourceEntry *input_res_entry = 
				&lookup_table_entry->resource_entry;

			int input_res_ctype = resource_compression_type(in_wim_ctype,
					input_res_entry->flags);

			return transfer_file_resource(in, 
						input_res_entry->size,
						input_res_entry->original_size, 
						input_res_entry->offset,
						input_res_ctype, 
						out, 
						out_wim_ctype,
						output_res_entry);
		}
	} else {
		return true;
	}
}

/* Reopens the FILE* for a WIM read-write. */
static bool reopen_rw(WIMStruct *w)
{
	wimlib_assert(w->fp != NULL && w->filename != NULL);
	if (fclose(w->fp) != 0) {
		wimlib_error("Failed to close the file `%s': %m\n", w->filename);
		return false;
	}
	FILE *fp = fopen(w->filename, "r+b");
	if (fp == NULL) {
		wimlib_error("Failed to open `%s' for reading and writing: "
				"%m\n", w->filename);
		return false;
	}
	w->fp = fp;
	return true;
}



/* 
 * Writes a WIM file to the original file that it was read from, overwriting it.
 *
 * After this function returns @w will no longer be valid and will have had all
 * its internal memory freed, and it must be either freed using free() or
 * wimlib_free(), or it must be associated with a new WIM file using
 * wimlib_begin_read(). 
 */
bool wimlib_overwrite(WIMStruct *w, bool check_integrity)
{
	const char *wimfile_name = w->filename;

	wimlib_assert(wimfile_name != NULL);

	DEBUG("Replacing WIM file `%s'\n", wimfile_name);

	/* Write the WIM to a temporary file. */
	/* XXX should the temporary file be somewhere else? */
	size_t wim_name_len = strlen(wimfile_name);
	char tmpfile[wim_name_len + 10];
	memcpy(tmpfile, wimfile_name, wim_name_len);
	randomize_byte_array_with_alnum((u8*)(tmpfile + wim_name_len), 9);
	tmpfile[wim_name_len + 9] = '\0';

	if (!wimlib_write(w, tmpfile, WIM_ALL_IMAGES, check_integrity)) {
		wimlib_error("Failed to write the WIM file `%s'!\n", tmpfile);
		return false;
	}

	DEBUG("Closing original WIM file.\n");
	/* Close the original WIM file that was opened for reading. */
	if (w->fp != NULL) {
		if (fclose(w->fp) != 0) {
			wimlib_error("Failed to close the file `%s'!\n",
					wimfile_name);
		}
		w->fp = NULL;
	}

	DEBUG("Renaming `%s' to `%s'\n", tmpfile, wimfile_name);

	/* Rename the new file to the old file .*/
	if (rename(tmpfile, wimfile_name) != 0) {
		wimlib_error("Failed to rename `%s' to `%s': %m\n", tmpfile, 
								wimfile_name);
		/* Remove temporary file. */
		if (unlink(tmpfile) != 0) {
			wimlib_error("Failed to remove `%s': %m\n", tmpfile);
		}
		return false;
	}

	DEBUG("Destroying WIMStruct.\n");
	/* @w is no longer valid. */
	wimlib_destroy(w);
	return true;
}


bool wimlib_overwrite_xml_and_header(WIMStruct *w, bool check)
{
	DEBUG("Overwriting XML and header of `%s', check = %d\n", 
				w->filename, check);
	if (!reopen_rw(w))
		return false;
	FILE *fp = w->fp;

	u8 *integrity_table = NULL;

	/* The old integrity table is still OK, as it doesn't include the header
	 * or XML data.  Save it for later if it exists and the @check parameter
	 * was specified. */
	if (check && w->hdr.integrity.offset != 0) {
		DEBUG("Reading existing integrity table.\n");
		integrity_table = xmalloc(w->hdr.integrity.size);
		if (fread(integrity_table, 1, w->hdr.integrity.size, fp) !=
						w->hdr.integrity.size) {
			if (feof(fp)) {
				wimlib_error("Unexpected EOF in existing "
					"integrity table!\n");
			} else {
				wimlib_error("Failed to read existing integrity "
					"table: %m\n");
			}
			free(integrity_table);
			return false;
		}
		DEBUG("Done reading existing integrity table.\n");
	}

	DEBUG("Overwriting XML data.\n");
	/* Overwrite the XML data. */
	if (fseeko(fp, w->hdr.xml_metadata.offset, SEEK_SET) != 0) {
		wimlib_error("Failed to seek to byte %"PRIu64" for XML data: "
				"%m\n", w->hdr.xml_metadata.offset);
		return false;
	}
	if (!write_xml_data(w->xml_info, WIM_ALL_IMAGES, fp))
		return false;

	DEBUG("Updating XML resource entry.\n");
	/* Update the XML resource entry in the WIM header. */
	off_t xml_end = ftello(fp);
	off_t xml_size = xml_end - w->hdr.xml_metadata.offset;
	w->hdr.xml_metadata.size = xml_size;
	w->hdr.xml_metadata.original_size = xml_size;
	/* Offset and flags are the same. */


	if (check) { /* need to write integrity table. */
		DEBUG("Writing integrity table.\n");
		if (integrity_table == NULL) {
			/* There was no existing integrity table, so a new one
			 * must be calculated. */
			if (!write_integrity(fp, w->hdr.hdr_size, 
					w->hdr.lookup_table_metadata.offset + 
					w->hdr.lookup_table_metadata.size))
				return false;
			off_t end_offset               = ftello(fp);
			off_t integrity_size           = end_offset - xml_end;
			w->hdr.integrity.offset        = xml_end;
			w->hdr.integrity.size          = integrity_size;
			w->hdr.integrity.original_size = integrity_size;
			w->hdr.integrity.flags         = 0;
		} else {
			/* The existing integrity table was saved. */
			size_t bytes_written = fwrite(integrity_table, 1, 
							w->hdr.integrity.size, fp);
			free(integrity_table);
			if (bytes_written != w->hdr.integrity.size) {
				wimlib_error("Failed to write integrity table: %m\n");
				return false;
			}
			w->hdr.integrity.offset = xml_end;
		}
	} else {
		DEBUG("Truncating file to end of XML data.\n");
		/* No integrity table to write.  The file should be truncated
		 * because it's possible that the old file was longer (due to it
		 * including an integrity table, or due to its XML data being
		 * longer) */
		if (fflush(fp) != 0) {
			wimlib_error("Failed to flush stream for file `%s': %m\n",
					w->filename);
			return false;
		}
		if (ftruncate(fileno(fp), xml_end) != 0) {
			wimlib_error("Failed to truncate `%s' to %"PRIu64" "
					"bytes: %m\n", w->filename);
			return false;
		}
		memset(&w->hdr.integrity, 0, sizeof(WIMResourceEntry));
	}

	DEBUG("Overwriting header.\n");
	/* Overwrite the header. */
	if (fseeko(fp, 0, SEEK_SET) != 0) {
		wimlib_error("Failed to seek to beginning of `%s': %m\n",
				w->filename);
		return false;
	}

	if (!write_header(&w->hdr, fp))
		return false;

	DEBUG("Closing file.\n");
	if (fclose(fp) != 0) {
		wimlib_error("Failed to close `%s': %m\n", w->filename);
		return false;
	}
	w->fp = NULL;
	DEBUG("Done.\n");
	return true;
}

/* 
 * Writes the WIM to a file.
 *
 * @w:  A WIMStruct that specifies a WIM file that has been opened for reading.
 * 	There may be in-memory changes made to it, which are then reflected in 
 * 	the output file.
 * @out_wim_path:  	The file to write the WIM to.
 * @image:		The image inside the WIM to write.
 * @check_integrity:	Whether to include integrity information or not.
 * @return:  True on success, false on failure.
 */
bool wimlib_write(WIMStruct *w, const char *out_wim_path, int image, 
							bool check)
{
	wimlib_assert(image == WIM_ALL_IMAGES || 
			(image >= 1 && image <= w->hdr.image_count));

	bool ret = true;

	if (image == WIM_ALL_IMAGES)
		DEBUG("Writing all images to `%s'\n", out_wim_path);
	else
		DEBUG("Writing image %d to `%s'\n", image, out_wim_path);

	const char *mode;
	/* checking the integrity requires going back over the file to read it.
	 * XXX 
	 * (It also would be possible to keep a running sha1sum as the file
	 * as written-- this would be faster, but a bit more complicated) */
	if (check) 
		mode = "w+b";
	else
		mode = "wb";

	FILE *out = fopen(out_wim_path, mode);
	if (out == NULL) {
		wimlib_error("Failed to open the file `%s' for writing!\n", 
								out_wim_path);
		return false;
	}

	w->out_fp = out;

	/* Create and write a preliminary header based on the current header. It
	 * will be overwritten later. */
	WIMHeader hdr;
	memcpy(&hdr, &w->hdr, sizeof(WIMHeader));
	if (!write_header(&hdr, out))
		goto write_error;

	/* Set all lookup table entries to not written. This flag indicates that
	 * their file resources have not yet been written to the WIM file. */
	lookup_table_zero_out_refcnts(&w->lookup_table);

	/* Write the file resources for each image, or only for the specified
	 * image if WIM_ALL_IMAGES was not given. */
	for (u32 i = 1; i <= w->hdr.image_count; i++) {

		if (image != WIM_ALL_IMAGES && i != image)
			continue;

		if (!wimlib_select_image(w, i))
			goto write_error;

		DEBUG("Writing file resources for image %u\n", i);
		if (!for_dentry_in_tree(wim_root_dentry(w), write_file_resource, w)) 
		{
			wimlib_error("Failed to write file resources "
					"to `%s'!\n", out_wim_path);
			goto write_error;
		}
	}

	off_t boot_metadata_offset = ftello(out);
	off_t cur_offset = boot_metadata_offset;

	/* Write the metadata resource for each image, or only for the specified
	 * image if WIM_ALL_IMAGES was not given. */
	for (u32 i = 1; i <= w->hdr.image_count; i++) {

		if (image != WIM_ALL_IMAGES && i != image)
			continue;

		if (!wimlib_select_image(w, i))
			goto write_error;

		DEBUG("Writing the boot metadata for image %u.\n", i);

		/* Write the security data. */
		if (!write_security_data(wim_security_data(w), out))
			goto write_error;

		/* Write the dentries. */
		off_t offset = wim_security_data(w)->total_length + 
					wim_root_dentry(w)->length + 8;
		DEBUG("Calculating subdir offsets for image %u.\n", i);
		calculate_subdir_offsets(wim_root_dentry(w), &offset);

		DEBUG("Writing dentry tree for image %u.\n", i);
		if (!write_dentry_tree(wim_root_dentry(w), out))
			goto write_error;

		DEBUG("Updating resource entry for image %u's boot metadata.\n",
				i);
		/* Update the resource entry for this image's boot
		 * metadata. */
		WIMResourceEntry *entry = wim_metadata_output_resource_entry(w);
		wim_metadata_lookup_table_entry(w)->out_refcnt++;
		entry->offset        = cur_offset;
		cur_offset           = ftello(out);
		entry->size          = cur_offset - entry->offset;
		entry->original_size = entry->size;
		entry->flags         = WIM_RESHDR_FLAG_METADATA;
	}

	off_t lookup_table_offset = ftello(out);

	DEBUG("Writing lookup table.\n");
	/* Write the lookup table. */
	if (!write_lookup_table(&w->lookup_table, out))
		goto write_error;

	DEBUG("Writing XML data.\n");

	off_t xml_data_offset = ftello(out);

	off_t lookup_table_size = xml_data_offset - lookup_table_offset;

	hdr.lookup_table_metadata.offset        = lookup_table_offset;
	hdr.lookup_table_metadata.size          = lookup_table_size;
	hdr.lookup_table_metadata.original_size = lookup_table_size;
	hdr.lookup_table_metadata.flags         = WIM_RESHDR_FLAG_METADATA;

	if (!write_xml_data(w->xml_info, image, out))
		goto write_error;

	off_t integrity_offset = ftello(out);

	off_t xml_data_size = integrity_offset - xml_data_offset;

	hdr.xml_metadata.offset                 = xml_data_offset;
	hdr.xml_metadata.size                   = xml_data_size;
	hdr.xml_metadata.original_size          = xml_data_size;
	hdr.xml_metadata.flags                  = 0;

	if (check) {
		if (!write_integrity(out, hdr.hdr_size, xml_data_offset))
			goto write_error;
		off_t end_offset = ftello(out);
		off_t integrity_size = end_offset - integrity_offset;
		hdr.integrity.offset = integrity_offset;
		hdr.integrity.size   = integrity_size;
		hdr.integrity.original_size = integrity_size;
	} else {
		hdr.integrity.offset        = 0;
		hdr.integrity.size          = 0;
		hdr.integrity.original_size = 0;
	}
	hdr.integrity.flags = 0;

	DEBUG("Updating WIM header.\n");


	/* 
	 * In the WIM header, here is room for one resource entry for boot
	 * metadata.  This entry was zeroed out on the 5-image install.wim from
	 * the Windows 7 DVD, and you have to search the lookup table to find
	 * the resource entries for the metadata resources.  However, Windows PE
	 * was unable to read a 2-image WIM with this entry zeroed out.  I'm not
	 * sure exactly what it's used for.  For now, just put the resource
	 * entry for the first metadata resource there.
	 */

	if (w->boot_metadata != NULL) {
		unsigned i;
		if (image == WIM_ALL_IMAGES)
			i = 0;
		else
			i = image - 1;
		memcpy(&hdr.boot_metametadata, 
				&w->boot_metadata[i].lookup_table_entry->output_resource_entry,
				sizeof(WIMResourceEntry));
	} else {
		memset(&hdr.boot_metametadata, 0, sizeof(WIMResourceEntry));
	}

	/* Set image count and boot index correctly for single image writes */
	if (image != WIM_ALL_IMAGES) {
		hdr.image_count = 1;
		if (hdr.boot_idx == image)
			hdr.boot_idx = 1;
		else
			hdr.boot_idx = 0;
	}

	fseeko(out, 0, SEEK_SET);
	if (!write_header(&hdr, out))
		goto write_error;


	ret = true;

	DEBUG("Done writing WIM file to `%s'.\n", out_wim_path);

write_error:
	DEBUG("Closing output file.\n");
	w->out_fp = NULL;
	if (fclose(out) != 0) {
		wimlib_error("Failed to close the file `%s': %m\n", out_wim_path);
		return false;
	}

	return ret;
}
