/*
 * xml.c
 *
 * Deals with the XML information in WIM files.  Uses the C library libxml2.
 *
 * wimlib - Library for working with WIM files 
 *
 * Copyright (C) 2012 Eric Biggers
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this library; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#include "wimlib_internal.h"
#include <string.h>

#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xmlwriter.h>

typedef struct WIMWindowsVersion {
	u64 major;
	u64 minor;
	u64 build;
	u64 sp_build;
	u64 sp_level;
} WIMWindowsVersion;

typedef struct WIMWindowsInfo {
	u64    arch;
	char  *product_name;
	char  *edition_id;
	char  *installation_type;
	char  *hal;
	char  *product_type;
	char  *product_suite;
	char **languages;
	char  *default_language;
	u64    num_languages;
	char  *system_root;
	bool   windows_version_exists;
	WIMWindowsVersion windows_version;
} WIMWindowsInfo;

typedef struct WIMImageInfo {
	u64   index;
	u64   dir_count;
	u64   file_count;
	u64   total_bytes;
	u64   hard_link_bytes;
	u64   creation_time;
	u64   last_modification_time;
	bool  windows_info_exists;
	WIMWindowsInfo windows_info;
	char *name;
	char *description;
	char  *display_name;
	char  *display_description;
	char  *flags;
} WIMImageInfo;

typedef struct WIMInfo {
	u64 total_bytes;
	u64 num_images;
	WIMImageInfo *images;
} WIMInfo;

/* Returns a statically allocated string that is a string representation of the
 * architecture number. */
static const char *get_arch(int arch)
{
	static char buf[20];
	switch (arch) {
	case 0:
		return "x86";
	case 9:
		return "x86_64";
	/* XXX Are there other arch values? */
	default:
		snprintf(buf, sizeof(buf), "%d (unknown)", arch);
		return buf;
	}
}


/* Iterate through the children of an xmlNode. */
#define for_node_child(parent, child) for (child = parent->children; \
				child != NULL; child = child->next)

/* Utility functions for xmlNodes */
static inline bool node_is_element(xmlNode *node)
{
	return node->type == XML_ELEMENT_NODE;
}

static inline bool node_is_text(xmlNode *node)
{
	return node->type == XML_TEXT_NODE;
}

static inline bool node_is_attribute(xmlNode *node)
{
	return node->type == XML_ATTRIBUTE_NODE;
}

static inline bool node_name_is(xmlNode *node, const char *name)
{
	/* For now, both upper case and lower case element names are accepted. */
	return strcasecmp(node->name, name) == 0;
}

/* Finds the text node that is a child of an element node and returns its
 * content converted to a 64-bit unsigned integer.  Returns 0 if no text node is
 * found. */
static u64 node_get_u64(const xmlNode *u64_node)
{
	xmlNode *child;
	for_node_child(u64_node, child)
		if (node_is_text(child))
			return strtoull(child->content, NULL, 10);
	return 0;
}

/* Like node_get_u64(), but expects a number in base 16. */
static u64 node_get_hex_u64(const xmlNode *u64_node)
{
	xmlNode *child;
	for_node_child(u64_node, child)
		if (node_is_text(child))
			return strtoull(child->content, NULL, 16);
	return 0;
}

/* Like node_get_u64(), but returns the content as a dynamically allocated
 * string. */
static char *node_get_string(const xmlNode *string_node)
{
	xmlNode *child;
	for_node_child(string_node, child)
		if (node_is_text(child))
			return xstrdup(child->content);
	return NULL;
}

/* Returns the timestamp from a time node.  It has child elements <HIGHPART> and
 * <LOWPART> that are then used to construct a 64-bit timestamp. */
static u64 node_get_timestamp(const xmlNode *time_node)
{
	u32 high_part = 0;
	u32 low_part = 0;
	xmlNode *child;
	for_node_child(time_node, child) {
		if (!node_is_element(child))
			continue;
		if (node_name_is(child, "HIGHPART"))
			high_part = node_get_hex_u64(child);
		else if (node_name_is(child, "LOWPART"))
			low_part = node_get_hex_u64(child);
	}
	return (u64)low_part | ((u64)high_part << 32);
}

/* Used to sort an array of WIMImageInfos by their image indices. */
static int sort_by_index(const void *p1, const void *p2)
{
	u64 index_1 = ((WIMImageInfo*)p1)->index;
	u64 index_2 = ((WIMImageInfo*)p1)->index;
	if (index_1 < index_2)
		return -1;
	else if (index_1 > index_2)
		return 1;
	else
		return 0;
}


/* Frees memory allocated inside a WIMWindowsInfo structure. */
static void destroy_windows_info(WIMWindowsInfo *windows_info)
{
	free(windows_info->product_name);
	free(windows_info->edition_id);
	free(windows_info->installation_type);
	free(windows_info->product_type);
	for (u32 i = 0; i < windows_info->num_languages; i++)
		free(windows_info->languages[i]);
	free(windows_info->languages);
	free(windows_info->system_root);
}

/* Frees memory allocated inside a WIMImageInfo structure. */
static void destroy_image_info(WIMImageInfo *image_info)
{
	free(image_info->name);
	free(image_info->description);
	destroy_windows_info(&image_info->windows_info);
}

/* Frees memory allocated inside a WIMInfo structure. */
static void destroy_wim_info(struct WIMInfo *info)
{
	/*DEBUG("Destroying WIMInfo with num_images = %d\n",*/
			/*info->num_images);*/
	for (u32 i = 0; i < info->num_images; i++)
		destroy_image_info(&info->images[i]);
	free(info->images);
}


/* Reads the information from a <VERSION> element inside the <WINDOWS> element.
 * */
static void xml_read_windows_version(const xmlNode *version_node, 
					WIMWindowsVersion* windows_version)
{
	xmlNode *child;
	for_node_child(version_node, child) {
		if (!node_is_element(child))
			continue;
		if (node_name_is(child, "MAJOR"))
			windows_version->major    = node_get_u64(child);
		else if (node_name_is(child, "MINOR"))
			windows_version->minor    = node_get_u64(child);
		else if (node_name_is(child, "BUILD"))
			windows_version->build    = node_get_u64(child);
		else if (node_name_is(child, "SPBUILD"))
			windows_version->sp_build = node_get_u64(child);
		else if (node_name_is(child, "SPLEVEL"))
			windows_version->sp_level = node_get_u64(child);
	}
}

/* Reads the information from a <LANGUAGE> element inside a <WINDOWS> element.
 * */
static void xml_read_languages(const xmlNode *languages_node, 
			char ***languages_ret, u64 *num_languages_ret,
			char **default_language_ret)
{
	xmlNode *child;
	*num_languages_ret = 0;
	for_node_child(languages_node, child)
		if (node_is_element(child) && node_name_is(child, "LANGUAGE"))
			(*num_languages_ret)++;
	*languages_ret = xmalloc(*num_languages_ret * sizeof(**languages_ret));
	u64 i = 0;
	for_node_child(languages_node, child) {
		if (!node_is_element(child))
			continue;
		if (node_name_is(child, "LANGUAGE"))
			(*languages_ret)[i] = node_get_string(child);
		else if (node_name_is(child, "DEFAULT"))
			*default_language_ret = node_get_string(child);
	}
}

/* Reads the information from a <WINDOWS> element inside an <IMAGE> element. */
static void xml_read_windows_info(const xmlNode *windows_node, 
					WIMWindowsInfo *windows_info)
{
	xmlNode *child;
	for_node_child(windows_node, child) {
		if (!node_is_element(child))
			continue;
		if (node_name_is(child, "ARCH"))
			windows_info->arch = node_get_u64(child);
		else if (node_name_is(child, "PRODUCTNAME"))
			windows_info->product_name = node_get_string(child);
		else if (node_name_is(child, "EDITIONID"))
			windows_info->edition_id = node_get_string(child);
		else if (node_name_is(child, "INSTALLATIONTYPE"))
			windows_info->installation_type = node_get_string(child);
		else if (node_name_is(child, "PRODUCTTYPE"))
			windows_info->product_type = node_get_string(child);
		else if (node_name_is(child, "PRODUCTSUITE"))
			windows_info->product_suite = node_get_string(child);
		else if (node_name_is(child, "LANGUAGES"))
			xml_read_languages(child, &windows_info->languages,
					&windows_info->num_languages,
					&windows_info->default_language);
		else if (node_name_is(child, "VERSION")) {
			xml_read_windows_version(child, &windows_info->windows_version);
			windows_info->windows_version_exists = true;
		}
		else if (node_name_is(child, "SYSTEMROOT"))
			windows_info->system_root = node_get_string(child);
		else if (node_name_is(child, "HAL"))
			windows_info->hal = node_get_string(child);
	}
}

/* Reads the information from an <IMAGE> element. */
static bool xml_read_image_info(xmlNode *image_node, WIMImageInfo *image_info)
{
	xmlNode *child;
	xmlChar *index_prop = xmlGetProp(image_node, "INDEX");
	if (index_prop == NULL)
		image_info->index = 0;
	else {
		image_info->index = strtoul(index_prop, NULL, 10);
		free(index_prop);
	}
	for_node_child(image_node, child) {
		if (!node_is_element(child))
			continue;

		if (node_name_is(child, "DIRCOUNT"))
			image_info->dir_count = node_get_u64(child);
		else if (node_name_is(child, "FILECOUNT"))
			image_info->file_count = node_get_u64(child);
		else if (node_name_is(child, "TOTALBYTES"))
			image_info->total_bytes = node_get_u64(child);
		else if (node_name_is(child, "HARDLINKBYTES"))
			image_info->hard_link_bytes = node_get_u64(child);
		else if (node_name_is(child, "CREATIONTIME"))
			image_info->creation_time = node_get_timestamp(child);
		else if (node_name_is(child, "LASTMODIFICATIONTIME"))
			image_info->last_modification_time = node_get_timestamp(child);
		else if (node_name_is(child, "WINDOWS")) {
			DEBUG("Found <WINDOWS> tag\n");
			xml_read_windows_info(child, &image_info->windows_info);
			image_info->windows_info_exists = true;
		}
		else if (node_name_is(child, "NAME"))
			image_info->name = node_get_string(child);
		else if (node_name_is(child, "DESCRIPTION"))
			image_info->description = node_get_string(child);
		else if (node_name_is(child, "FLAGS"))
			image_info->flags = node_get_string(child);
		else if (node_name_is(child, "DISPLAYNAME"))
			image_info->display_name = node_get_string(child);
		else if (node_name_is(child, "DISPLAYDESCRIPTION"))
			image_info->display_description = node_get_string(child);
	}
	return true;
}

/* Reads the information from a <WIM> element, which should be the root element
 * of the XML tree. */
static bool xml_read_wim_info(xmlNode *wim_node, WIMInfo *wim_info)
{
	/* Count how many images there are. */
	xmlNode *child;
	for_node_child(wim_node, child)
		if (node_is_element(child) && node_name_is(child, "IMAGE"))
			wim_info->num_images++;

	/* There must be at least one image. */
	if (wim_info->num_images == 0) {
		wimlib_error("WARNING: Could not find any <IMAGE> elements in "
				"the XML data!\n");
		return true;
	}

	/* Allocate the array of WIMImageInfos and fill them in. */
	wim_info->images = xzalloc(wim_info->num_images * sizeof(WIMImageInfo));
	WIMImageInfo *cur_image_info = wim_info->images;
	for_node_child(wim_node, child) {
		if (node_is_element(child) && node_name_is(child, "IMAGE")) {
			DEBUG("Found <IMAGE> tag\n");
			if (!xml_read_image_info(child, cur_image_info++))
				goto xml_read_wim_info_error;
		}
		if (node_is_element(child) && node_name_is(child, "TOTALBYTES"))
			wim_info->total_bytes = node_get_u64(child);
	}

	/* Sort the array of WIMImageInfos by image index. */
	qsort(wim_info->images, wim_info->num_images, sizeof(WIMImageInfo),
			sort_by_index);
	return true;
xml_read_wim_info_error:
	destroy_wim_info(wim_info);
	return false;
}

/* Prints the information contained in a WIMWindowsInfo structure. */
static void print_windows_info(const WIMWindowsInfo *windows_info)
{
	printf("Architecture:           %s\n", get_arch(windows_info->arch));
	printf("Product Name:           %s\n", windows_info->product_name);
	printf("Edition ID:             %s\n", windows_info->edition_id);
	printf("Installation Type:      %s\n", windows_info->installation_type);
	if (windows_info->hal != NULL)
		printf("HAL:                    %s\n", windows_info->hal);
	printf("Product Type:           %s\n", windows_info->product_type);
	if (windows_info->product_suite != NULL)
		printf("Product Suite:          %s\n", windows_info->product_suite);
	printf("Languages:              ");
	for (u64 i = 0; i < windows_info->num_languages; i++) {
		fputs(windows_info->languages[i], stdout);
		putchar(' ');
	}
	putchar('\n');
	printf("Default Language:       %s\n", windows_info->default_language);
	printf("System Root:            %s\n", windows_info->system_root);
	const WIMWindowsVersion *windows_version = &windows_info->windows_version;
	printf("Major Version:          %"PRIu64"\n", windows_version->major);
	printf("Minor Version:          %"PRIu64"\n", windows_version->minor);
	printf("Build:                  %"PRIu64"\n", windows_version->build);
	printf("Service Pack Build:     %"PRIu64"\n", windows_version->sp_build);
	printf("Service Pack Level:     %"PRIu64"\n", windows_version->sp_level);
}


/* Writes the information contained in a WIMWindowsVersion structure to the XML
 * document being constructed in memory.  This is the <VERSION> element inside
 * the <WINDOWS> element. */
static int xml_write_windows_version(xmlTextWriter *writer, 
					const WIMWindowsVersion *version)
{
	int rc;
	rc = xmlTextWriterStartElement(writer, "VERSION");
	if (rc < 0)
		return rc;

	rc = xmlTextWriterWriteFormatElement(writer, "MAJOR", "%"PRIu64, 
								version->major);
	if (rc < 0)
		return rc;

	rc = xmlTextWriterWriteFormatElement(writer, "MINOR", "%"PRIu64, 
								version->minor);
	if (rc < 0)
		return rc;

	rc = xmlTextWriterWriteFormatElement(writer, "BUILD", "%"PRIu64, 
								version->build);
	if (rc < 0)
		return rc;

	rc = xmlTextWriterWriteFormatElement(writer, "SPBUILD", "%"PRIu64, 
								version->sp_build);
	if (rc < 0)
		return rc;

	rc = xmlTextWriterWriteFormatElement(writer, "SPLEVEL", "%"PRIu64, 
								version->sp_level);
	if (rc < 0)
		return rc;

	return xmlTextWriterEndElement(writer); /* </VERSION> */
}

/* Writes the information contained in a WIMWindowsInfo structure to the XML
 * document being constructed in memory. This is the <WINDOWS> element. */
static int xml_write_windows_info(xmlTextWriter *writer, 
					const WIMWindowsInfo *windows_info)
{
	int rc;
	rc = xmlTextWriterStartElement(writer, "WINDOWS");
	if (rc < 0)
		return rc;


	rc = xmlTextWriterWriteFormatElement(writer, "ARCH", "%"PRIu64, 
							windows_info->arch);
	if (rc < 0)
		return rc;
	
	if (windows_info->product_name != NULL) {
		rc = xmlTextWriterWriteElement(writer, "PRODUCTNAME", 
							windows_info->product_name);
		if (rc < 0)
			return rc;
	}

	if (windows_info->edition_id != NULL) {
		rc = xmlTextWriterWriteElement(writer, "EDITIONID", 
							windows_info->edition_id);
		if (rc < 0)
			return rc;
	}

	if (windows_info->installation_type != NULL) {
		rc = xmlTextWriterWriteElement(writer, "INSTALLATIONTYPE", 
							windows_info->installation_type);
		if (rc < 0)
			return rc;
	}

	if (windows_info->hal != NULL) {
		rc = xmlTextWriterWriteElement(writer, "HAL", 
							windows_info->hal);
		if (rc < 0)
			return rc;
	}

	if (windows_info->system_root != NULL) {
		rc = xmlTextWriterWriteElement(writer, "SYSTEMROOT", 
						windows_info->system_root);
			if (rc < 0)
				return rc;
	}

	if (windows_info->product_type != NULL) {
		rc = xmlTextWriterWriteElement(writer, "PRODUCTTYPE", 
						windows_info->product_type);
		if (rc < 0)
			return rc;
	}

	if (windows_info->product_suite != NULL) {
		rc = xmlTextWriterWriteElement(writer, "PRODUCTSUITE", 
						windows_info->product_suite);
			if (rc < 0)
				return rc;
	}

	if (windows_info->num_languages != 0) {
		rc = xmlTextWriterStartElement(writer, "LANGUAGES");
		if (rc < 0)
			return rc;

		for (int i = 0; i < windows_info->num_languages; i++) {
			rc = xmlTextWriterWriteElement(writer, "LANGUAGE", 
							windows_info->languages[i]);
			if (rc < 0)
				return rc;
		}
		rc = xmlTextWriterWriteElement(writer, "DEFAULT", 
						windows_info->default_language);
		if (rc < 0)
			return rc;

		rc = xmlTextWriterEndElement(writer); /* </LANGUAGES> */
		if (rc < 0)
			return rc;
	}

	if (windows_info->windows_version_exists) {
		rc = xml_write_windows_version(writer, &windows_info->windows_version);
		if (rc < 0)
			return rc;
	}

	return xmlTextWriterEndElement(writer); /* </WINDOWS> */
}

/* Writes a time element to the XML document being constructed in memory. */
static int xml_write_time(xmlTextWriter *writer, const char *element_name, 
								u64 time) 
{
	int rc;
	rc = xmlTextWriterStartElement(writer, element_name);
	if (rc < 0)
		return rc;

	rc = xmlTextWriterWriteFormatElement(writer, "HIGHPART", 
					"0x%"PRIX32, (u32)(time >> 32));
	if (rc < 0)
		return rc;

	rc = xmlTextWriterWriteFormatElement(writer, "LOWPART",
						"0x%"PRIX32, (u32)time);
	if (rc < 0)
		return rc;

	rc = xmlTextWriterEndElement(writer); /* </@element_name> */
	if (rc < 0)
		return rc;
	return 0;
}


/* Writes an <IMAGE> element to the XML document. */
static int xml_write_image_info(xmlTextWriter *writer, 
					const WIMImageInfo *image_info)
{
	int rc;
	rc = xmlTextWriterStartElement(writer, "IMAGE");
	if (rc < 0)
		return rc;

	rc = xmlTextWriterWriteFormatAttribute(writer, "INDEX", "%"PRIu64, 
						image_info->index);
	if (rc < 0)
		return rc;

	rc = xmlTextWriterWriteFormatElement(writer, "DIRCOUNT", "%"PRIu64, 
						image_info->dir_count);
	if (rc < 0)
		return rc;

	rc = xmlTextWriterWriteFormatElement(writer, "FILECOUNT", "%"PRIu64, 
						image_info->file_count);
	if (rc < 0)
		return rc;

	rc = xmlTextWriterWriteFormatElement(writer, "TOTALBYTES", "%"PRIu64, 
						image_info->total_bytes);
	if (rc < 0)
		return rc;

	rc = xmlTextWriterWriteFormatElement(writer, "HARDLINKBYTES", "%"PRIu64, 
						image_info->hard_link_bytes);
	if (rc < 0)
		return rc;

	rc = xml_write_time(writer, "CREATIONTIME", 
						image_info->creation_time);
	if (rc < 0)
		return rc;

	rc = xml_write_time(writer, "LASTMODIFICATIONTIME", 
						image_info->last_modification_time);
	if (rc < 0)
		return rc;

	if (image_info->windows_info_exists) {
		rc = xml_write_windows_info(writer, &image_info->windows_info);
		if (rc < 0)
			return rc;
	} else {
		DEBUG("<WINDOWS> tag does not exist.\n");
	}

	if (image_info->name != NULL) {
		rc = xmlTextWriterWriteElement(writer, "NAME", image_info->name);
		if (rc < 0)
			return rc;
	}
	if (image_info->description != NULL) {
		rc = xmlTextWriterWriteElement(writer, "DESCRIPTION", 
							image_info->description);
		if (rc < 0)
			return rc;
	}
	if (image_info->display_name != NULL) {
		rc = xmlTextWriterWriteElement(writer, "DISPLAYNAME", 
						image_info->display_name);
		if (rc < 0)
			return rc;
	}
	if (image_info->display_description != NULL) {
		rc = xmlTextWriterWriteElement(writer, "DISPLAYDESCRIPTION", 
						image_info->display_description);
		if (rc < 0)
			return rc;
	}

	if (image_info->flags != NULL) {
		rc = xmlTextWriterWriteElement(writer, "FLAGS",
				image_info->flags);
	}

	return xmlTextWriterEndElement(writer); /* </IMAGE> */
}



/* Makes space for another image in the XML information. */
static WIMImageInfo *xml_add_image_struct(WIMInfo *wim_info)
{
	wim_info->num_images++;
	wim_info->images = xrealloc(wim_info->images, wim_info->num_images * 
							sizeof(WIMImageInfo));
	WIMImageInfo *image_info = &wim_info->images[wim_info->num_images - 1];
	memset(image_info, 0, sizeof(WIMImageInfo));
	return image_info;
}

static void clone_windows_info(const WIMWindowsInfo *old, WIMWindowsInfo *new)
{
	memcpy(new, old, sizeof(*old));
	new->product_name = xstrdup(old->product_name);
	new->edition_id = xstrdup(old->edition_id);
	new->installation_type = xstrdup(old->installation_type);
	new->hal = xstrdup(old->hal);
	new->product_type = xstrdup(old->product_type);
	new->product_suite = xstrdup(old->product_suite);
	new->languages = xmalloc(new->num_languages * sizeof(char*));
	for (unsigned i = 0; i < new->num_languages; i++)
		new->languages[i] = xstrdup(old->languages[i]);
	new->default_language = xstrdup(old->default_language);
	new->system_root = xstrdup(old->system_root);
}

static void clone_image_info(const WIMImageInfo *old, WIMImageInfo *new)
{
	memcpy(new, old, sizeof(*old));
	new->name                = xstrdup(old->name);
	new->description         = xstrdup(old->description);
	new->display_name        = xstrdup(old->display_name);
	new->display_description = xstrdup(old->display_description);
	new->flags               = xstrdup(old->flags);
	if (old->windows_info_exists)
		clone_windows_info(&old->windows_info, &new->windows_info);
}

/* Copies the XML information for an image between WIM files. 
 *
 * @dest_image_name and @dest_image_description are ignored if they are NULL;
 * otherwise, they are used to overwrite the image name and/or image description
 * from the XML data in the source WIM file. */
INTERNAL void xml_export_image(const WIMInfo *old_wim_info, int image, 
				WIMInfo **new_wim_info_p, 
				const char *dest_image_name, 
				const char *dest_image_description)
{
	DEBUG("Copying XML data between WIM files for source image %d",
			image);
	if (*new_wim_info_p == NULL)
		*new_wim_info_p = xzalloc(sizeof(WIMInfo));
	WIMInfo *new_wim_info = *new_wim_info_p;

	wimlib_assert(image >= 1 && image <= old_wim_info->num_images);

	WIMImageInfo *image_info = xml_add_image_struct(new_wim_info);

	clone_image_info(&old_wim_info->images[image - 1], image_info);

	image_info->index = new_wim_info->num_images;
	if (dest_image_name != NULL)
		image_info->name = xstrdup(dest_image_name);
	if (dest_image_description != NULL)
		image_info->description = xstrdup(dest_image_description);
}

/* Removes an image from the XML information. */
INTERNAL void xml_delete_image(WIMInfo **wim_info_p, int image)
{
	WIMInfo *wim_info = *wim_info_p;

	wimlib_assert(wim_info != NULL);
	wimlib_assert(image >= 1 && image <= wim_info->num_images);

	destroy_image_info(&wim_info->images[image - 1]);

	for (int i = image - 1; i < wim_info->num_images - 1; i++) {
		memcpy(&wim_info->images[i], &wim_info->images[i + 1],
					sizeof(WIMImageInfo));
		wim_info->images[i].index--;
	}

	if (--wim_info->num_images == 0) {
		free_xml_info(wim_info);
		*wim_info_p = NULL;
	}
}

/* Adds an image to the XML information. */
INTERNAL void xml_add_image(WIMStruct *w, WIMDentry *root_dentry, 
			const char *name, const char *description, 
						const char *flags)
{


	/*Calculate statistics for the image and update the WIM XML data. */
	DEBUG("Calculating statistics for the image.\n");
	u64 dir_count, file_count, total_bytes, hard_link_bytes;

	calculate_dir_tree_statistics(root_dentry, &w->lookup_table,
			&dir_count, &file_count, &total_bytes,
			&hard_link_bytes);

	DEBUG("dir_count = %"PRIu64", file_count = %"PRIu64", "
			"total_bytes = %"PRIu64", hard_link_bytes = %"PRIu64"\n",
			dir_count, file_count, total_bytes, hard_link_bytes);


	/* If this is the first image , allocate the WIMInfo. */
	if (w->xml_info == NULL)
		w->xml_info = xzalloc(sizeof(WIMInfo));

	WIMInfo *wim_info = w->xml_info;

	WIMImageInfo *image_info    = xml_add_image_struct(wim_info);
	image_info->index           = wim_info->num_images;
	image_info->dir_count       = dir_count;
	image_info->file_count      = file_count;
	image_info->total_bytes     = total_bytes;
	image_info->hard_link_bytes = hard_link_bytes;

	u64 now = get_timestamp();
	image_info->creation_time          = now;
	image_info->last_modification_time = now;
	image_info->name                   = xstrdup(name);
	image_info->description            = xstrdup(description);
	image_info->flags                  = xstrdup(flags);
}

/* Returns the value of the <TOTALBYTES> element inside the <WIM> element. */
INTERNAL u64 xml_info_get_total_bytes(const WIMInfo *info)
{
	return info->total_bytes;
}


/* Prints information about the specified image from WIMInfo structure. 
 * @image may be WIM_ALL_IMAGES. */
INTERNAL void print_image_info(const WIMInfo *wim_info, int image)
{
	if (image == WIM_ALL_IMAGES) {
		for (int i = 1; i <= wim_info->num_images; i++)
			print_image_info(wim_info, i);
	} else {
		wimlib_assert(image >= 1 && image <= wim_info->num_images);
		const WIMImageInfo *image_info = &wim_info->images[image - 1];
		printf("Index:                  %"PRIu64"\n", image_info->index);
		printf("Name:                   %s\n", image_info->name);

		/* Always print the Description: part even if there is no
		 * description. */
		const char *desc;
		if (image_info->description == NULL)
			desc = "";
		else
			desc = image_info->description;
		printf("Description:            %s\n", desc);

		if (image_info->display_name != NULL)
			printf("Display Name:           %s\n", 
						image_info->display_name);
		if (image_info->display_description != NULL)
			printf("Display Description:    %s\n", 
						image_info->display_description);
		printf("Directory Count:        %"PRIu64"\n", image_info->dir_count);
		printf("File Count:             %"PRIu64"\n", image_info->file_count);
		printf("Total Bytes:            %"PRIu64"\n", image_info->total_bytes);
		printf("Hard Link Bytes:        %"PRIu64"\n", image_info->hard_link_bytes);

		time_t ctime = ms_timestamp_to_unix(image_info->creation_time);
		time_t mtime = ms_timestamp_to_unix(image_info->last_modification_time);

		printf("Creation Time:          %s", asctime(localtime(&ctime)));
		printf("Last Modification Time: %s", asctime(localtime(&mtime)));
		if (image_info->windows_info_exists)
			print_windows_info(&image_info->windows_info);
		if (image_info->flags != NULL)
			printf("Flags:                  %s\n", image_info->flags);
		putchar('\n');
	}
}

/* 
 * Reads the XML data from the WIM file.
 *
 * @fp:		  The FILE* for the WIM file.
 * @res:	  A pointer to the resource entry for the XML data.
 * @xml_data_ret: A pointer to a location into which a pointer to the raw XML
 * 		  data will be allocated and returned.
 * @xml_tree_ret: A pointer to a location into which the parsed XML data will
 * 		  be allocated and returned.
 * @return:	  True on success, false on failure. It is failure if the XML
 * 		  data does not include a toplevel <WIM> element, or if there
 * 		  are syntax errors in it.
 */
INTERNAL bool read_xml_data(FILE *fp, WIMResourceEntry *res, u8 **xml_data_ret, 
							WIMInfo **info_ret)
{
	u8 *xml_data = NULL;
	xmlDoc *doc = NULL;
	xmlNode *root;
	WIMInfo *wim_info;

	DEBUG("XML data is %"PRIu64" bytes long.\n", (u64)res->size);

	if (resource_is_compressed(res)) {
		wimlib_error("XML data is supposed to be uncompressed!\n");
		goto read_xml_data_error;
	}
	if (res->size < 2) {
		wimlib_error("XML data must be at least 2 bytes!\n");
		goto read_xml_data_error;
	}

	if (*xml_data_ret == NULL) {
		/* Read the XML data from the WIM. */
		DEBUG("Reading XML data from the WIM.\n");
		xml_data = xmalloc(res->size + 1);
		if (!read_full_resource(fp, res->size, res->size, res->offset, 
					WIM_COMPRESSION_TYPE_NONE, xml_data))
			goto read_xml_data_error;
		xml_data[res->size] = '\0';
	} else {
		/* We have already read the XML data from the WIM. */
		DEBUG("Already read the XML data!\n");
		xml_data = *xml_data_ret;
	}

	LIBXML_TEST_VERSION

	DEBUG("Parsing XML using libxml2 to create XML tree.\n");

	doc = xmlReadMemory(xml_data, res->size, 
			"noname.xml", "UTF-16", 0);

	xmlCleanupParser();

	if (doc == NULL) {
		wimlib_error("Failed to parse XML data!\n");
		goto read_xml_data_error;
	}

	DEBUG("Constructing WIM information structure from XML tree.\n");

	root = xmlDocGetRootElement(doc);
	if (root == NULL) {
		wimlib_error("Empty XML document!\n");
		goto read_xml_data_error;
	}

	if (!(node_is_element(root) && node_name_is(root, "WIM"))) {
		wimlib_error("Expected <WIM> for the root XML element! "
				"(Found <%s>)\n", root->name);
		goto read_xml_data_error;
	}

	wim_info = xzalloc(sizeof(WIMInfo));
	if (!xml_read_wim_info(root, wim_info))
		goto read_xml_data_error;

	DEBUG("Freeing XML tree.\n");

	xmlFreeDoc(doc);
	*xml_data_ret = xml_data;
	*info_ret = wim_info;
	return true;
read_xml_data_error:
	free(xml_data);
	xmlFreeDoc(doc);
	return false;
}

#define CHECK_RC  ({ 	if (rc < 0)  { \
				wimlib_error("Error writing XML data!\n"); \
				goto xml_write_cleanup; \
			} })

/* 
 * Writes XML data to a WIM file.
 *
 * @xml_info:	The pointer to the WIMInfo structure that contains the XML
 * 		information for the WIM file. 
 * @out:	The FILE* for the output file.
 * @return:	True on success, false on failure.
 */
INTERNAL bool write_xml_data(const WIMInfo *wim_info, int image, FILE *out)
{
	xmlBuffer     *buf = NULL;
	xmlTextWriter *writer = NULL;
	char          *utf16_str = NULL;
	int rc;

	wimlib_assert(image == WIM_ALL_IMAGES || 
			(wim_info != NULL && image >= 1 && 
			 image <= wim_info->num_images));

	/* The contents of the <TOTALBYTES> element in the XML data, under the
	 * <WIM> element not the <IMAGE> element, is the size of the WIM file
	 * excluding the XML data.  Which is the current offset, since the XML
	 * data goes at the very end of the WIM file. */
	off_t total_bytes = ftello(out);

	DEBUG("Creating XML buffer and text writer\n");
	buf = xmlBufferCreate();
	if (buf == NULL) {
		wimlib_error("Failed to allocated XML buffer!\n");
		goto xml_write_cleanup;
	}
	writer = xmlNewTextWriterMemory(buf, 0);
	if (writer == NULL) {
		wimlib_error("Failed to allocate XML writer!\n");
		goto xml_write_cleanup;
	}

	DEBUG("Beginning document\n");
	/* XXX */
	/* M$'s WIM files do not have XML declarations, so do not write one.
	 * I'm not sure how we can force the document to be written in UTF-16
	 * without calling xmlTextWriterStartDocument(), though, so currently it
	 * is composed in a buffer UTF-8, then converted to UTF-16. */
#if 0
	rc = xmlTextWriterStartDocument(writer, NULL, "UTF-16", NULL);
	CHECK_RC;
#endif

	DEBUG("Writing <WIM> element\n");
	rc = xmlTextWriterStartElement(writer, "WIM");
	CHECK_RC;

	DEBUG("Writing <TOTALBYTES> element\n");

	rc = xmlTextWriterWriteFormatElement(writer, "TOTALBYTES", "%"PRIu64,
			total_bytes);
	CHECK_RC;

	DEBUG("Writing <IMAGE> elements\n");

	int num_images;
	if (wim_info == NULL)
		num_images = 0;
	else
		num_images = wim_info->num_images;

	for (int i = 1; i <= num_images; i++) {
		if (image != WIM_ALL_IMAGES && i != image)
			continue;
		DEBUG("Writing <IMAGE> element for image %d\n", i);
		rc = xml_write_image_info(writer, &wim_info->images[i - 1]);
		if (rc < 0) {
			wimlib_error("Error writing XML data for image %d!\n",
					i);
			goto xml_write_cleanup;

		}
	}

	rc = xmlTextWriterEndElement(writer);
	CHECK_RC;

	rc = xmlTextWriterEndDocument(writer);
	CHECK_RC;

	DEBUG("Done composing XML document. Now converting to UTF-16 and "
			"writing it to the output file.\n");

	const xmlChar *content = xmlBufferContent(buf);
	size_t len = xmlBufferLength(buf);

	size_t utf16_len;
	utf16_str = utf8_to_utf16(content, len, &utf16_len);
	wimlib_assert(utf16_str != NULL);

	size_t bytes_written = fwrite(utf16_str, 1, utf16_len, out);


	if (bytes_written != utf16_len) {
		wimlib_error("Error writing XML data: %m\n");
		goto xml_write_cleanup;
	}
	DEBUG("Cleaning up.\n");

	free(utf16_str);
	/*if (writer != NULL)*/
		/*xmlFreeTextWriter(writer);*/
	/*if (buf != NULL)*/
		/*xmlBufferFree(buf);*/
	DEBUG("Done writing XML data.\n");
	return true;
xml_write_cleanup:
	free(utf16_str);
	if (writer != NULL)
		xmlFreeTextWriter(writer);
	if (buf != NULL)
		xmlBufferFree(buf);
	return false;
}

INTERNAL void free_xml_info(WIMInfo *wim_info)
{
	if (wim_info == NULL)
		return;
	destroy_wim_info(wim_info);
	free(wim_info);
}

/* 
 * Prints the XML data for the WIM file.
 *
 * @xml_data:	The XML data.
 * @size:	The size of the XML data.
 */
static void print_xml_data(const char *xml_data, u64 size)
{
	const char *utf16_str = xml_data;
	size_t      utf16_len = size;
	const char *utf8_str;
	size_t      utf8_len;
	utf8_str = utf16_to_utf8(utf16_str, utf16_len, &utf8_len);
	puts(utf8_str);
}

/* Returns the name of the specified image. */
const char *wimlib_get_image_name(WIMStruct *w, int image)
{
	wimlib_assert(image >= 1 && image <= w->hdr.image_count);
	WIMImageInfo *image_info = &w->xml_info->images[image - 1];
	wimlib_assert(image_info->name != NULL);
	return image_info->name;
}

/* Returns true if the specified name is already used by some image in the WIM.
 * */
bool wimlib_image_name_in_use(WIMStruct *w, const char *name)
{
	wimlib_assert(name != NULL);
	DEBUG("Checking to see if `%s' is already in use\n", name);
	if (w->xml_info == NULL)
		return false;

	for (int i = 1; i <= w->xml_info->num_images; i++) {
		if (strcmp(w->xml_info->images[i - 1].name, name) == 0)
			return true;
	}
	return false;
}

/* Print the XML data from the WIM file associated with @w. 
 * Return true on success, false on failure. */
bool wimlib_print_xml_data(WIMStruct *w, int image)
{
	if (image == WIM_ALL_IMAGES)
		print_xml_data(w->xml_data, w->hdr.xml_metadata.size);
	else
		;	/* XXX */
	return true;
}

/* Extracts the XML data of the WIM to the file outside the WIM named by
 * @out_file. */
bool wimlib_extract_xml_data(WIMStruct *w, const char *out_file)
{
	FILE *fp = fopen(out_file, "wb");
	if (fp == NULL) {
		wimlib_error("Failed to open the file `%s' for writing XML data!\n",
				out_file);
		return false;
	}
	if (fwrite(w->xml_data, 1, w->hdr.xml_metadata.size, fp) != 
			w->hdr.xml_metadata.size) 
	{
		wimlib_error("Failed to write XML data to the file `%s'!",
				out_file);
		return false;
	}
	return true;
}

/* Sets the name of an image in the WIM. */
void wimlib_set_image_name(WIMStruct *w, int image, const char *name)
{
	wimlib_assert(image >= 1 && image <= w->hdr.image_count &&
			w->xml_info != NULL);
	free(w->xml_info->images[image - 1].name);
	w->xml_info->images[image - 1].name = xstrdup(name);
}

/* Sets the description of an image in the WIM. */
void wimlib_set_image_descripton(WIMStruct *w, int image, const char *description)
{
	wimlib_assert(image >= 1 && image <= w->hdr.image_count &&
			w->xml_info != NULL);
	free(w->xml_info->images[image - 1].description);
	w->xml_info->images[image - 1].description = xstrdup(description);
}
